/*
 * ocl_sha512.c
 *
 * hashkill - a hash cracking tool
 * Copyright (C) 2010 Milen Rangelov <gat3way@gat3way.eu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CL_CL_H

#include <CL/cl.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include "err.h"
#include "ocl-threads.h"
#include "plugins.h"
#include "hashinterface.h"
#include "sessions.h"
#include "ocl_support.h"


#define ROTL(p,q) ((p) << (q)) | ((p) >> (32-(q)));
#define ROTR(p,q) ((p) >> (q)) | ((p) << (32-(q)));
#define REV(p) { tmp=(p);tmp1=ROTL(tmp,8);tmp2=ROTL(tmp,24); (p)=(tmp1 & 0x00FF00FF)|(tmp2 & 0xFF00FF00); }
#define H0 0x6A09E667
#define H1 0xBB67AE85
#define H2 0x3C6EF372
#define H3 0xA54FF53A
#define H4 0x510E527F
#define H5 0x9B05688C
#define H6 0x1F83D9AB
#define H7 0x5BE0CD19


cl_platform_id platform;
cl_device_id device[16];
pthread_mutex_t biglock;

cl_context context[64];
int devicesnum;
pthread_t crack_threads[32];
cl_program program[64];


static void ocl_bruteforce_init_restore(int self, int len);
static void ocl_markov_init_restore(int self);

#define ALGO_THRESHOLD 1300
static void bfi_int_magic(unsigned char *kernel, int len)
{
    unsigned int dword0,dword1,dword2,j,i,k;
    
    for (k=4;k<12;k++)
    {
        j=0;
        for (i=4+k;i<(len-8);i+=8)
        {
            dword0 = *(unsigned int *)(kernel + i);
            dword1 = *(unsigned int *)(kernel + i-4);
            if  (
                ((((dword1)>>5)&31)==13) && // INST
                ((((dword1)>>4)&1)==0) &&   // NEG
                ((((dword1)>>23)&1)==0) &&    // CLAMP
                ((((dword1)>>20)&1)==0) &&    // DSTREL
                ((((dword0)>>4)&1)==0) &&    // SRC1NEG
                ((((dword0)>>17)&1)==0)     // SRC2NEG
            ) j++;
        }
        if (j>=ALGO_THRESHOLD)
        for (i=4+k;i<(len-8);i+=8)
        {
            dword0 = *(unsigned int *)(kernel + i);
            dword1 = *(unsigned int *)(kernel + i-4);
            if  (
                ((((dword1)>>5)&31)==13) && // INST
                ((((dword1)>>4)&1)==0) &&   // NEG
                ((((dword1)>>23)&1)==0) &&    // CLAMP
                ((((dword1)>>20)&1)==0) &&    // DSTREL
                ((((dword0)>>4)&1)==0) &&    // SRC1NEG
                ((((dword0)>>17)&1)==0)     // SRC2NEG
            )
            {
                *(unsigned int *)(kernel+i-4) = (dword1 & 0xFFFFFFFF) ^ (11 << 5);
            }
        }
    }

}


hash_stat ocl_bruteforce_sha512_thread(void *arg)
{
    int err;
    cl_command_queue queue;
    cl_mem buffer;
    size_t global_work_size;
    cl_ulong4 image;
    int self;
    cl_kernel kernel;
    unsigned char hex1[16];
    char plain[64];
    cl_mem bitmaps_buf;
    cl_uint *bitmaps;
    int a,b,c,d,e;
    int a1,a2,a3,a4,a5,a6,a7,a8,a9,a10;
    char *ptr = malloc(strlen(bruteforce_charset)*strlen(bruteforce_charset)*strlen(bruteforce_charset)*64*ocl_vector);
    int charset_size = (int)strlen(bruteforce_charset);
    cl_mem found_ind_buf;
    cl_uint *found_ind=malloc(strlen(bruteforce_charset)*strlen(bruteforce_charset)*strlen(bruteforce_charset)*sizeof(cl_uint));
    int found;
    cl_mem found_buf;
    struct  hash_list_s  *mylist, *addlist;
    unsigned int csize=1;
    cl_mem table_buf;
    unsigned int *table;
    cl_ulong4 cbase;
    cl_uint4 singlehash;
    int session_restore_lflag;


    pthread_mutex_lock(&biglock);
    memcpy(&self,arg,sizeof(int));
    bitmaps = malloc(256*256*32*4);
    session_restore_lflag = session_restore_flag;
    kernel = clCreateKernel(program[self], "sha512", &err );
    if (err != CL_SUCCESS)
    {
	elog("clCreateKernel error (%d)\n",err);
	return hash_err;
    }
    queue = clCreateCommandQueue(context[self], device[self/ocl_threads], 0, &err );
    if (err != CL_SUCCESS)
    {
        elog("clCreateCommandQueue error (%d)\n",err);
	return hash_err;
    }
    buffer = clCreateBuffer(context[self], CL_MEM_WRITE_ONLY, charset_size*charset_size*charset_size*64*ocl_vector, NULL, &err );
    if (err != CL_SUCCESS)
    {
	elog("clCreateBuffer error (%d)\n",err);
	return hash_err;
    }
    found_buf = clCreateBuffer(context[self], CL_MEM_WRITE_ONLY, 4, NULL, &err );
    if (err != CL_SUCCESS)
    {
	elog("clCreateBuffer error (%d)\n",err);
	return hash_err;
    }
    if (err != CL_SUCCESS)
    {
	elog("clCreateBuffer error (%d)\n",err);
	return hash_err;
    }
    table = malloc(sizeof(uint)*charset_size*charset_size*charset_size);

    for (a=0;a<charset_size*charset_size*charset_size;a++)
    {
	unsigned char x,y,z,t1;
	t1 = (a / (charset_size*charset_size));
	x = (bruteforce_charset[(a / (charset_size*charset_size))]);
	y = (bruteforce_charset[((a-t1*(charset_size*charset_size))/charset_size)]);
	z = (bruteforce_charset[(a % charset_size)]);
	table[a] = (x<<16)|(y<<8)|(z);
    }
    for (a=0;a<256*256*32;a++)
    {
	    bitmaps[a]=0;
    }

    mylist = hash_list;
    a=0;
    while (mylist)
    {
        memcpy(hex1,mylist->hash,16);
	a++;
	unsigned int *b_a = (unsigned int *)hex1;
	unsigned int *b_b = (unsigned int *)&hex1[4];
	unsigned int *b_c = (unsigned int *)&hex1[8];
	unsigned int *b_d = (unsigned int *)&hex1[12];
	unsigned int bind_a = (*b_a)>>13;
	unsigned int bval_a = (1<<((*b_a)&31));
	unsigned int bind_b = (*b_b)>>13;
	unsigned int bval_b = (1<<((*b_b)&31));
	unsigned int bind_c = (*b_c)>>13;
	unsigned int bval_c = (1<<((*b_c)&31));
	unsigned int bind_d = (*b_d)>>13;
	unsigned int bval_d = (1<<((*b_d)&31));
	bitmaps[bind_a] |=bval_a;
	bitmaps[bind_b+65535*8] |=bval_b;
	bitmaps[bind_c+65535*16] |=bval_c;
	bitmaps[bind_d+65535*24] |=bval_d;
	singlehash.x |= (1<<((*b_b)&31));
	singlehash.y |= (1<<((*b_c)&31));
	singlehash.z |= (1<<((*b_d)&31));
        if (mylist) mylist = mylist->next;
    }

    if (a==1)
    {
	mylist = hash_list;
	memcpy(hex1,mylist->hash,4);
	memcpy(&singlehash.x, hex1, 4);
	memcpy(hex1,mylist->hash+4,4);
	memcpy(&singlehash.y, hex1, 4);
	memcpy(hex1,mylist->hash+8,4);
	memcpy(&singlehash.z, hex1, 4);
	memcpy(hex1,mylist->hash+12,4);
	memcpy(&singlehash.w, hex1, 4);
	
	clSetKernelArg(kernel, 8, sizeof(cl_uint4), (void*) &singlehash);
    }
    else 
    {
	singlehash.w=0;
	clSetKernelArg(kernel, 8, sizeof(cl_uint4), (void*) &singlehash);
    }

    clSetKernelArg(kernel, 8, sizeof(cl_uint4), (void*) &singlehash);
    bitmaps_buf = clCreateBuffer(context[self], CL_MEM_READ_ONLY|CL_MEM_USE_HOST_PTR, 65535*32*4, bitmaps, &err );
    if (err != CL_SUCCESS)
    {
	elog("clCreateBuffer error (%d)\n",err);
	return hash_err;
    }
    table_buf = clCreateBuffer(context[self], CL_MEM_READ_ONLY|CL_MEM_USE_HOST_PTR, sizeof(uint)*charset_size*charset_size*charset_size, table, &err );


    cbase.s0 = bruteforce_charset[0];
    cbase.s1 = bruteforce_charset[1];
    cbase.s2 = bruteforce_charset[2];
    cbase.s3 = bruteforce_charset[3];


    err=clSetKernelArg(kernel, 3, sizeof(cl_ulong4), (void*) &cbase);
    if (err != CL_SUCCESS)
    {
	elog("clSetKernelArg error (%d)\n",err);
	return hash_err;
    }

    err=clSetKernelArg(kernel, 0, sizeof(cl_mem), (void*) &buffer);
    if (err != CL_SUCCESS)
    {
	elog("clSetKernelArg error (%d)\n",err);
	return hash_err;
    }

    err=clSetKernelArg(kernel, 1, sizeof(cl_ulong4), (void*) &image);
    if (err != CL_SUCCESS)
    {
	elog("clSetKernelArg error (%d)\n",err);
	return hash_err;
    }

    err=clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);
    if (err != CL_SUCCESS)
    {
	elog("clSetKernelArg error (%d)\n",err);
	return hash_err;
    }

    err=clSetKernelArg(kernel, 3, sizeof(cl_ulong4), (void*) &cbase);
    if (err != CL_SUCCESS)
    {
	elog("clSetKernelArg error (%d)\n",err);
	return hash_err;
    }

    bzero(found_ind,sizeof(uint)*charset_size*charset_size*charset_size);
    found_ind_buf = clCreateBuffer(context[self], CL_MEM_WRITE_ONLY, charset_size*charset_size*charset_size*sizeof(cl_uint), NULL, &err );
    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    if (err != CL_SUCCESS)
    {
	elog("clCreateBuffer error (%d)\n",err);
	return hash_err;
    }
    clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);

    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    err=clSetKernelArg(kernel, 4, sizeof(cl_mem), (void*) &found_ind_buf);
    if (err != CL_SUCCESS)
    {
	elog("clSetKernelArg error (%d)\n",err);
	return hash_err;
    }

    err=clSetKernelArg(kernel, 5, sizeof(cl_mem), (void*) &bitmaps_buf);

    if (err != CL_SUCCESS)
    {
	elog("clSetKernelArg error (%d)\n",err);
	return hash_err;
    }
    err=clSetKernelArg(kernel, 6, sizeof(cl_mem), (void*) &found_buf);
    if (err != CL_SUCCESS)
    {
	elog("clSetKernelArg error (%d)\n",err);
	return hash_err;
    }

    err=clSetKernelArg(kernel, 7, sizeof(cl_mem), (void*) &table_buf);
    if (err != CL_SUCCESS)
    {
	elog("clSetKernelArg error (%d)\n",err);
	return hash_err;
    }

    global_work_size = charset_size*charset_size*charset_size;
    while ((global_work_size %256)!=0) global_work_size++;
    if (err != CL_SUCCESS)
    {
	elog("clEnqueueBuffer error (%d)\n",err);
	return hash_err;
    }
    image.x=image.y=image.z=image.w=0;




    /* Bruteforce, len=4 */
    csize=1;
    err=clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);
    if (err != CL_SUCCESS)
    {
	elog("clSetKernelArg error (%d)\n",err);
	return hash_err;
    }

    err=clSetKernelArg(kernel, 1, sizeof(cl_long4), (void*) &image);
    if (err != CL_SUCCESS)
    {
	elog("clSetKernelArg error (%d)\n",err);
	return hash_err;
    }

    pthread_mutex_unlock(&biglock); // right now it should be safe to release the mutex
    if (session_restore_lflag==0) ocl_bruteforce_init_restore(self, 0);

    if ((bruteforce_start<=4) && (bruteforce_end>=4))
    if (thread_restore[self].cstart<=0)
    for (a1=thread_restore[self].cpos[0];a1<charset_size;a1+=(ocl_vector*ocl_threads*devicesnum))
    {
	if (attack_over!=0) return hash_ok;
	thread_save[self].cpos[0]=a1;
	thread_save[self].cstr[0]=bruteforce_charset[a1];
	thread_save[self].cstr[1]=0;
	thread_save[self].cstart=0;

	cbase.s0 = bruteforce_charset[a1];
	cbase.s1 = bruteforce_charset[a1+1];
	cbase.s2 = bruteforce_charset[a1+2];
	cbase.s3 = bruteforce_charset[a1+3];

	clSetKernelArg(kernel, 3, sizeof(cl_ulong4), (void*) &cbase);

	image.x=0;image.y=0;image.z=0;image.w=0;
        if (found==1)
        {
    	    found = 0;
    	    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    	}
        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
        //clFlush(queue);
	invocations++;
	if (err != CL_SUCCESS)
	{
	    elog("clEnqueueNDRangeKernel error (%d)\n",err);
	    return hash_err;
	}
	//err = clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
        int *fnd;
        fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
        found = *fnd;

	if (err != CL_SUCCESS)
	{
	    elog("clEnqueueReadBuffer error (%d)\n",err);
	    return hash_err;
	}
/*
    int i;
    clEnqueueReadBuffer(queue, buffer, CL_TRUE, 0, 8*16, ptr, 0, NULL, NULL);
    for(i=0; i < 64; i+=16)
         printf("aaa? =  %02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x\n",
              ptr[i]&255, ptr[i+1]&255,ptr[i+2]&255,ptr[i+3]&255,ptr[i+4]&255,ptr[i+5]&255,
                   ptr[i+6]&255,ptr[i+7]&255,ptr[i+8]&255,ptr[i+9]&255,ptr[i+10]&255,ptr[i+11]&255,
                        ptr[i+12]&255,ptr[i+13]&255,ptr[i+14]&255,ptr[i+15]&255
                             );
                             
*/
        if (session_restore_lflag==1) ocl_bruteforce_init_restore(self, 0);
	session_restore_lflag = 0;
	if (found==1) 
	{
	    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*sizeof(uint), found_ind, 0, NULL, NULL);
	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    for (a=0;a<charset_size;a++)
	    for (b=0;b<charset_size;b++)
	    for (c=0;c<charset_size;c++)
	    if (found_ind[a*charset_size*charset_size+b*charset_size+c]==1)
	    {
	    e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector;
    	    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, e*32, 64*ocl_vector, ptr+e*64, 0, NULL, NULL);
    	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    for (d=0;d<ocl_vector;d++)
	    {
		e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector+d;
    		if ( (hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count<640) && (hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count>0))
    		{
    		    int i;
    		    for (i=0;i<hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count;i++)
    		    if (memcmp(hash_index[ptr[e*64]&255][ptr[e*64+1]&255].nodes[i]->hash, (char *)ptr+(e)*64, hash_ret_len) == 0)
    		    {
                        mylist = hash_index[ptr[e*64]&255][ptr[e*64+1]&255].nodes[i];
                        int flag = 0;
                        plain[0]=bruteforce_charset[(a1+(d))%charset_size];
                        plain[1]=bruteforce_charset[c];
                        plain[2]=bruteforce_charset[b];
                        plain[3]=bruteforce_charset[a];
                        plain[4]=0;
                        pthread_mutex_lock(&crackedmutex);
                        addlist = cracked_list;
                        while (addlist)
                        {
                            if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                            addlist = addlist->next;
                        }
                        pthread_mutex_unlock(&crackedmutex);
                        if (flag == 0)
                        {
                            add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                        }
    		    }
    		}
    		else if (hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count>=320)
    		{
    		    mylist = hash_list;
    		    while (mylist)
    		    {
            		if (memcmp(mylist->hash, (char *)ptr+(e)*64, hash_ret_len) == 0)
            		{
                    	    int flag = 0;
                    	    plain[0]=bruteforce_charset[(a1+(d))%charset_size];
                    	    plain[1]=bruteforce_charset[c];
                    	    plain[2]=bruteforce_charset[b];
                    	    plain[3]=bruteforce_charset[a];
                    	    plain[4]=0;
                    	    pthread_mutex_lock(&crackedmutex);
                    	    addlist = cracked_list;
                    	    while (addlist)
                    	    {
                        	if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                        	addlist = addlist->next;
                    	    }
                    	    pthread_mutex_unlock(&crackedmutex);
                    	    if (flag == 0)
                    	    {
                        	add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    	    }
            		}
            		mylist = mylist->next;
        	    }
        	}
	    }
	    }
	clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
        bzero(found_ind,charset_size*charset_size*charset_size*sizeof(cl_uint));
        clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
	}
    }




    // Bruteforce, len=5 
    csize=2;
    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);
    if (session_restore_lflag==0) ocl_bruteforce_init_restore(self, 1);

    if ((bruteforce_start<=5) && (bruteforce_end>=5))
    if (thread_restore[self].cstart<=1)
    for (a1=thread_restore[self].cpos[0];a1<charset_size;a1++)
    for (a2=thread_restore[self].cpos[1];a2<charset_size;a2+=(ocl_vector*ocl_threads*devicesnum))
    {

	thread_save[self].cpos[0]=a1;
	thread_save[self].cpos[1]=a2;
	thread_save[self].cstr[0]=bruteforce_charset[a1];
	thread_save[self].cstr[1]=bruteforce_charset[a2];
	thread_save[self].cstr[2]=0;
	thread_save[self].cstart=1;


	cbase.s0 = bruteforce_charset[a2];
	cbase.s1 = bruteforce_charset[a2+1];
	cbase.s2 = bruteforce_charset[a2+2];
	cbase.s3 = bruteforce_charset[a2+3];

	clSetKernelArg(kernel, 3, sizeof(cl_ulong4), (void*) &cbase);

	if (attack_over!=0) return hash_ok;
	image.x=(bruteforce_charset[a1]);// |bruteforce_charset[a2]<<8, etc
	image.y=0;image.z=0;image.w=0;
	clSetKernelArg(kernel, 1, sizeof(cl_ulong4), (void*) &image);

	if (found==1)
	{
    	    found = 0;
    	    clEnqueueWriteBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
    	}

        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
        //clFinish(queue);
	invocations++;
	if (err != CL_SUCCESS)
	{
	    elog("clEnqueueNDRangeKernel error (%d)\n",err);
	    return hash_err;
	}
	//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
        int *fnd;
        fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
        found = *fnd;

        if (session_restore_lflag==1) ocl_bruteforce_init_restore(self, 1);
	session_restore_lflag = 0;
	if (found==1) 
	{
	    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*sizeof(uint), found_ind, 0, NULL, NULL);

	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    e=-1;
	    for (a=0;a<charset_size;a++)
	    for (b=0;b<charset_size;b++)
	    for (c=0;c<charset_size;c++)
	    if (found_ind[a*charset_size*charset_size+b*charset_size+c]==1)
	    {
    	    e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector;
    	    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, e*64, 64*ocl_vector, ptr+e*64, 0, NULL, NULL);
    	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);

	    for (d=0;d<ocl_vector;d++)
	    {
    		e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector+d;

    		if ( (hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count<640) && (hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count>0))
    		{
    		    int i;
    		    for (i=0;i<hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count;i++)
    		    if (memcmp(hash_index[ptr[e*64]&255][ptr[e*64+1]&255].nodes[i]->hash, (char *)ptr+(e)*64, hash_ret_len) == 0)
    		    {
                        mylist = hash_index[ptr[e*64]&255][ptr[e*64+1]&255].nodes[i];
                        int flag = 0;
                        plain[0]=bruteforce_charset[a1];
                        plain[1]=bruteforce_charset[a2+(d)];
                        plain[2]=bruteforce_charset[c];
                        plain[3]=bruteforce_charset[b];
                        plain[4]=bruteforce_charset[a];
                        plain[5]=0;
                        pthread_mutex_lock(&crackedmutex);
                        addlist = cracked_list;
                        while (addlist)
                        {
                            if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                            addlist = addlist->next;
                        }
                        pthread_mutex_unlock(&crackedmutex);
                        if (flag == 0)
                        {
                            add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                        }
    		    }
    		}
    		else if (hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count>=320)
		{
    		    mylist = hash_list;
    		    while (mylist)
    		    {
            		if ( (memcmp(mylist->hash, (char *)ptr+(e)*64, hash_ret_len) == 0) )
            		{
                    	    int flag = 0;
                    	    plain[0]=bruteforce_charset[a1];
                    	    plain[1]=bruteforce_charset[a2+(d)];
                    	    plain[2]=bruteforce_charset[c];
                    	    plain[3]=bruteforce_charset[b];
                    	    plain[4]=bruteforce_charset[a];
                    	    plain[5]=0;
                    	    pthread_mutex_lock(&crackedmutex);
                    	    addlist = cracked_list;
                    	    while (addlist)
                    	    {
                        	if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                        	addlist = addlist->next;
                    	    }
                    	    pthread_mutex_unlock(&crackedmutex);
                    	    if (flag == 0)
                    	    {
                        	add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    	    }
            		}
            		mylist = mylist->next;
        	    }
		}
	    }
    	    }
	clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
    	bzero(found_ind,charset_size*charset_size*charset_size*sizeof(cl_uint));
    	clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
	}
    }




    // Bruteforce, len=6 
    csize=3;
    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);
    if (session_restore_lflag==0) ocl_bruteforce_init_restore(self, 2);

    if ((bruteforce_start<=6) && (bruteforce_end>=6))
    if (thread_restore[self].cstart<=2)
    for (a1=thread_restore[self].cpos[0];a1<charset_size;a1++)
    for (a2=thread_restore[self].cpos[1];a2<charset_size;a2++)
    for (a3=thread_restore[self].cpos[2];a3<charset_size;a3+=(ocl_vector*ocl_threads*devicesnum))
    {
	cbase.s0 = bruteforce_charset[a3];
	cbase.s1 = bruteforce_charset[a3+1];
	cbase.s2 = bruteforce_charset[a3+2];
	cbase.s3 = bruteforce_charset[a3+3];

	clSetKernelArg(kernel, 3, sizeof(cl_ulong4), (void*) &cbase);

	thread_save[self].cpos[0]=a1;
	thread_save[self].cpos[1]=a2;
	thread_save[self].cpos[2]=a3;
	thread_save[self].cstr[0]=bruteforce_charset[a1];
	thread_save[self].cstr[1]=bruteforce_charset[a2];
	thread_save[self].cstr[2]=bruteforce_charset[a3];
	thread_save[self].cstr[3]=0;
	thread_save[self].cstart=2;


	if (attack_over!=0) return hash_ok;
	image.x=(bruteforce_charset[a1])|(bruteforce_charset[a2]<<8);// |bruteforce_charset[a2]<<8, etc
	image.y=0;image.z=0;image.w=0;
	clSetKernelArg(kernel, 1, sizeof(cl_ulong4), (void*) &image);

	if (found==1)
	{
    	    found = 0;
    	    clEnqueueWriteBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
    	}
        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
        //clFinish(queue);
	invocations++;
	if (err != CL_SUCCESS)
	{
	    elog("clEnqueueNDRangeKernel error (%d)\n",err);
	    return hash_err;
	}
	//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
        int *fnd;
        fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
        found = *fnd;

        if (session_restore_lflag==1) ocl_bruteforce_init_restore(self, 2);
	session_restore_lflag = 0;
	if (found==1) 
	{

	    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*sizeof(uint), found_ind, 0, NULL, NULL);
	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    e=-1;
	    for (a=0;a<charset_size;a++)
	    for (b=0;b<charset_size;b++)
	    for (c=0;c<charset_size;c++)
	    if (found_ind[a*charset_size*charset_size+b*charset_size+c]==1)
	    {
    	    e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector;
    	    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, e*64, 64*ocl_vector, ptr+e*64, 0, NULL, NULL);
    	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    for (d=0;d<ocl_vector;d++)
	    {
    		e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector+d;

    		if ( (hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count<320) && (hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count>0))
    		{
    		    int i;
    		    for (i=0;i<hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count;i++)
    		    if (memcmp(hash_index[ptr[e*64]&255][ptr[e*64+1]&255].nodes[i]->hash, (char *)ptr+(e)*64, hash_ret_len) == 0)
    		    {
                        mylist = hash_index[ptr[e*64]&255][ptr[e*64+1]&255].nodes[i];
                        int flag = 0;
                        plain[0]=bruteforce_charset[a1];
                        plain[1]=bruteforce_charset[a2];
                        plain[2]=bruteforce_charset[a3+(d)];
                        plain[3]=bruteforce_charset[c];
                        plain[4]=bruteforce_charset[b];
                        plain[5]=bruteforce_charset[a];
                        plain[6]=0;
                        pthread_mutex_lock(&crackedmutex);
                        addlist = cracked_list;
                        while (addlist)
                        {
                            if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                            addlist = addlist->next;
                        }
                        pthread_mutex_unlock(&crackedmutex);
                        if (flag == 0)
                        {
                            add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                        }
    		    }
    		}
    		else if (hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count>=320)
		{
    		    mylist = hash_list;
    		    while (mylist)
    		    {
            		if ( (memcmp(mylist->hash, (char *)ptr+(e)*64, hash_ret_len) == 0) )
            		{
                    	    int flag = 0;
                    	    plain[0]=bruteforce_charset[a1];
                    	    plain[1]=bruteforce_charset[a2];
                    	    plain[2]=bruteforce_charset[a3+(d)];
                    	    plain[3]=bruteforce_charset[c];
                    	    plain[4]=bruteforce_charset[b];
                    	    plain[5]=bruteforce_charset[a];
                    	    plain[6]=0;
                    	    pthread_mutex_lock(&crackedmutex);
                    	    addlist = cracked_list;
                    	    while (addlist)
                    	    {
                        	if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                        	addlist = addlist->next;
                    	    }
                    	    pthread_mutex_unlock(&crackedmutex);
                    	    if (flag == 0)
                    	    {
                        	add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    	    }
            		}
            		mylist = mylist->next;
        	    }
		}
	    }
    	    }
	clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
        bzero(found_ind,charset_size*charset_size*charset_size*sizeof(cl_uint));
        clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
	}
    }



    // Bruteforce, len=7 
    csize=4;
    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);
    if (session_restore_lflag==0) ocl_bruteforce_init_restore(self, 3);

    if ((bruteforce_start<=7) && (bruteforce_end>=7))
    if (thread_restore[self].cstart<=3)
    for (a1=thread_restore[self].cpos[0];a1<charset_size;a1++)
    for (a2=thread_restore[self].cpos[1];a2<charset_size;a2++)
    for (a3=thread_restore[self].cpos[2];a3<charset_size;a3++)
    for (a4=thread_restore[self].cpos[3];a4<charset_size;a4+=(ocl_vector*ocl_threads*devicesnum))
    {
	cbase.s0 = bruteforce_charset[a4];
	cbase.s1 = bruteforce_charset[a4+1];
	cbase.s2 = bruteforce_charset[a4+2];
	cbase.s3 = bruteforce_charset[a4+3];
	clSetKernelArg(kernel, 3, sizeof(cl_ulong4), (void*) &cbase);

	thread_save[self].cpos[0]=a1;
	thread_save[self].cpos[1]=a2;
	thread_save[self].cpos[2]=a3;
	thread_save[self].cpos[3]=a4;
	thread_save[self].cstr[0]=bruteforce_charset[a1];
	thread_save[self].cstr[1]=bruteforce_charset[a2];
	thread_save[self].cstr[2]=bruteforce_charset[a3];
	thread_save[self].cstr[3]=bruteforce_charset[a4];
	thread_save[self].cstr[4]=0;
	thread_save[self].cstart=3;

	if (attack_over!=0) return hash_ok;
	image.x=(bruteforce_charset[a1])|(bruteforce_charset[a2]<<8)|(bruteforce_charset[a3]<<16);// |bruteforce_charset[a2]<<8, etc
	image.y=0;image.z=0;image.w=0;
	clSetKernelArg(kernel, 1, sizeof(cl_ulong4), (void*) &image);

	if (found==1)
	{
    	    found = 0;
    	    clEnqueueWriteBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
    	}
        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
        //clFlush(queue);
	invocations++;
	if (err != CL_SUCCESS)
	{
	    elog("clEnqueueNDRangeKernel error (%d)\n",err);
	    return hash_err;
	}
	//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
        int *fnd;
        fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
        found = *fnd;

        if (session_restore_lflag==1) ocl_bruteforce_init_restore(self, 3);
	session_restore_lflag = 0;
	if (found==1) 
	{
	    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*sizeof(uint), found_ind, 0, NULL, NULL);
	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    e=-1;
	    for (a=0;a<charset_size;a++)
	    for (b=0;b<charset_size;b++)
	    for (c=0;c<charset_size;c++)
	    if (found_ind[a*charset_size*charset_size+b*charset_size+c]==1)
	    {
    	    e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector;
    	    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, e*64, 64*ocl_vector, ptr+e*64, 0, NULL, NULL);
    	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    for (d=0;d<ocl_vector;d++)
	    {
    		e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector+d;
    		if ( (hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count<640) && (hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count>0))
    		{
    		    int i;
    		    for (i=0;i<hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count;i++)
    		    if (memcmp(hash_index[ptr[e*64]&255][ptr[e*64+1]&255].nodes[i]->hash, (char *)ptr+(e)*64, hash_ret_len) == 0)
    		    {
                        mylist = hash_index[ptr[e*64]&255][ptr[e*64+1]&255].nodes[i];
                        int flag = 0;
                        plain[0]=bruteforce_charset[a1];
                        plain[1]=bruteforce_charset[a2];
                        plain[2]=bruteforce_charset[a3];
                        plain[3]=bruteforce_charset[a4+(d)];
                        plain[4]=bruteforce_charset[c];
                        plain[5]=bruteforce_charset[b];
                        plain[6]=bruteforce_charset[a];
                        plain[7]=0;
                        pthread_mutex_lock(&crackedmutex);
                        addlist = cracked_list;
                        while (addlist)
                        {
                            if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                            addlist = addlist->next;
                        }
                        pthread_mutex_unlock(&crackedmutex);
                        if (flag == 0)
                        {
                            add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                        }
    		    }
    		}
    		else if (hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count>=320)
		{
    		    mylist = hash_list;
    		    while (mylist)
    		    {
            		if ( (memcmp(mylist->hash, (char *)ptr+(e)*64, hash_ret_len) == 0) )
            		{
                    	    int flag = 0;
                    	    plain[0]=bruteforce_charset[a1];
                    	    plain[1]=bruteforce_charset[a2];
                    	    plain[2]=bruteforce_charset[a3];
                    	    plain[3]=bruteforce_charset[a4+(d)];
                    	    plain[4]=bruteforce_charset[c];
                    	    plain[5]=bruteforce_charset[b];
                    	    plain[6]=bruteforce_charset[a];
                    	    plain[7]=0;
                    	    pthread_mutex_lock(&crackedmutex);
                    	    addlist = cracked_list;
                    	    while (addlist)
                    	    {
                        	if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                        	addlist = addlist->next;
                    	    }
                    	    pthread_mutex_unlock(&crackedmutex);
                    	    if (flag == 0)
                    	    {
                        	add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    	    }
            		}
            		mylist = mylist->next;
        	    }
		}
	    }
    	    }
	clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
	bzero(found_ind,charset_size*charset_size*charset_size*sizeof(cl_uint));
	clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
	}
    }




    // Bruteforce, len=8
    csize=5;
    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);
    if (session_restore_lflag==0) ocl_bruteforce_init_restore(self, 4);

    if ((bruteforce_start<=8) && (bruteforce_end>=8))
    if (thread_restore[self].cstart<=4)
    for (a1=thread_restore[self].cpos[0];a1<charset_size;a1++)
    for (a2=thread_restore[self].cpos[1];a2<charset_size;a2++)
    for (a3=thread_restore[self].cpos[2];a3<charset_size;a3++)
    for (a4=thread_restore[self].cpos[3];a4<charset_size;a4++)
    for (a5=thread_restore[self].cpos[4];a5<charset_size;a5+=(ocl_vector*ocl_threads*devicesnum))
    {
	cbase.s0 = bruteforce_charset[a5];
	cbase.s1 = bruteforce_charset[a5+1];
	cbase.s2 = bruteforce_charset[a5+2];
	cbase.s3 = bruteforce_charset[a5+3];

	clSetKernelArg(kernel, 3, sizeof(cl_ulong4), (void*) &cbase);

	thread_save[self].cpos[0]=a1;
	thread_save[self].cpos[1]=a2;
	thread_save[self].cpos[2]=a3;
	thread_save[self].cpos[3]=a4;
	thread_save[self].cpos[4]=a5;
	thread_save[self].cstr[0]=bruteforce_charset[a1];
	thread_save[self].cstr[1]=bruteforce_charset[a2];
	thread_save[self].cstr[2]=bruteforce_charset[a3];
	thread_save[self].cstr[3]=bruteforce_charset[a4];
	thread_save[self].cstr[4]=bruteforce_charset[a5];
	thread_save[self].cstr[5]=0;
	thread_save[self].cstart=4;


	if (attack_over!=0) return hash_ok;
	image.x=(bruteforce_charset[a1])|(bruteforce_charset[a2]<<8)|(bruteforce_charset[a3]<<16)|(bruteforce_charset[a4]<<24);// |bruteforce_charset[a2]<<8, etc
	image.y=0;image.z=0;image.w=0;
	clSetKernelArg(kernel, 1, sizeof(cl_ulong4), (void*) &image);

	if (found==1)
	{
    	    found = 0;
    	    clEnqueueWriteBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
    	}
        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
        //clFlush(queue);
	invocations++;
	if (err != CL_SUCCESS)
	{
	    elog("clEnqueueNDRangeKernel error (%d)\n",err);
	    return hash_err;
	}
	//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
        int *fnd;
        fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
        found = *fnd;

        if (session_restore_lflag==1) ocl_bruteforce_init_restore(self, 4);
	session_restore_lflag = 0;
	if (found==1) 
	{
	    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*sizeof(uint), found_ind, 0, NULL, NULL);
	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    e=-1;
	    for (a=0;a<charset_size;a++)
	    for (b=0;b<charset_size;b++)
	    for (c=0;c<charset_size;c++)
	    if (found_ind[a*charset_size*charset_size+b*charset_size+c]==1)
	    {
    	    e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector;
    	    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, e*64, 64*ocl_vector, ptr+e*64, 0, NULL, NULL);
    	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    for (d=0;d<ocl_vector;d++)
	    {
    		e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector+d;
    		if ( (hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count<320) && (hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count>0))
    		{
    		    int i;
    		    for (i=0;i<hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count;i++)
    		    if (memcmp(hash_index[ptr[e*64]&255][ptr[e*64+1]&255].nodes[i]->hash, (char *)ptr+(e)*64, hash_ret_len) == 0)
    		    {
                        mylist = hash_index[ptr[e*64]&255][ptr[e*64+1]&255].nodes[i];
                        int flag = 0;
                        plain[0]=bruteforce_charset[a1];
                        plain[1]=bruteforce_charset[a2];
                        plain[2]=bruteforce_charset[a3];
                        plain[3]=bruteforce_charset[a4];
                        plain[4]=bruteforce_charset[a5+(d)];
                        plain[5]=bruteforce_charset[c];
                        plain[6]=bruteforce_charset[b];
                        plain[7]=bruteforce_charset[a];
                        plain[8]=0;
                        pthread_mutex_lock(&crackedmutex);
                        addlist = cracked_list;
                        while (addlist)
                        {
                            if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                            addlist = addlist->next;
                        }
                        pthread_mutex_unlock(&crackedmutex);
                        if (flag == 0)
                        {
                            add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                        }
    		    }
    		}
    		else if (hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count>=320)
		{
    		    mylist = hash_list;
    		    while (mylist)
    		    {
            		if ( (memcmp(mylist->hash, (char *)ptr+(e)*64, hash_ret_len) == 0) )
            		{
                    	    int flag = 0;
                    	    plain[0]=bruteforce_charset[a1];
                    	    plain[1]=bruteforce_charset[a2];
                    	    plain[2]=bruteforce_charset[a3];
                    	    plain[3]=bruteforce_charset[a4];
                    	    plain[4]=bruteforce_charset[a5+(d)];
                    	    plain[5]=bruteforce_charset[c];
                    	    plain[6]=bruteforce_charset[b];
                    	    plain[7]=bruteforce_charset[a];
                    	    plain[8]=0;
                    	    pthread_mutex_lock(&crackedmutex);
                    	    addlist = cracked_list;
                    	    while (addlist)
                    	    {
                        	if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                        	addlist = addlist->next;
                    	    }
                    	    pthread_mutex_unlock(&crackedmutex);
                    	    if (flag == 0)
                    	    {
                        	add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    	    }
            		}
            		mylist = mylist->next;
        	    }
		}
	    }
    	    }
	clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
        bzero(found_ind,charset_size*charset_size*charset_size*sizeof(cl_uint));
        clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
	}
    }




    // Bruteforce, len=9
    csize=6;
    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);
    if (session_restore_lflag==0) ocl_bruteforce_init_restore(self, 5);

    if ((bruteforce_start<=9) && (bruteforce_end>=9))
    if (thread_restore[self].cstart<=5)
    for (a1=thread_restore[self].cpos[0];a1<charset_size;a1++)
    for (a2=thread_restore[self].cpos[1];a2<charset_size;a2++)
    for (a3=thread_restore[self].cpos[2];a3<charset_size;a3++)
    for (a4=thread_restore[self].cpos[3];a4<charset_size;a4++)
    for (a5=thread_restore[self].cpos[4];a5<charset_size;a5++)
    for (a6=thread_restore[self].cpos[5];a6<charset_size;a6+=(ocl_vector*ocl_threads*devicesnum))
    {
	cbase.s0 = bruteforce_charset[a6];
	cbase.s1 = bruteforce_charset[a6+1];
	cbase.s2 = bruteforce_charset[a6+2];
	cbase.s3 = bruteforce_charset[a6+3];

	clSetKernelArg(kernel, 3, sizeof(cl_ulong8), (void*) &cbase);

	thread_save[self].cpos[0]=a1;
	thread_save[self].cpos[1]=a2;
	thread_save[self].cpos[2]=a3;
	thread_save[self].cpos[3]=a4;
	thread_save[self].cpos[4]=a5;
	thread_save[self].cpos[5]=a6;
	thread_save[self].cstr[0]=bruteforce_charset[a1];
	thread_save[self].cstr[1]=bruteforce_charset[a2];
	thread_save[self].cstr[2]=bruteforce_charset[a3];
	thread_save[self].cstr[3]=bruteforce_charset[a4];
	thread_save[self].cstr[4]=bruteforce_charset[a5];
	thread_save[self].cstr[5]=bruteforce_charset[a6];
	thread_save[self].cstr[6]=0;
	thread_save[self].cstart=5;

	if (attack_over!=0) return hash_ok;
	image.x=(bruteforce_charset[a1])|(bruteforce_charset[a2]<<8)|(bruteforce_charset[a3]<<16)|(bruteforce_charset[a4]<<24);// |bruteforce_charset[a2]<<8, etc
	image.y=(bruteforce_charset[a5]);image.z=0;image.w=0;
	clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);

	if (found==1)
	{
    	    found = 0;
    	    clEnqueueWriteBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
    	}
        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
        //clFlush(queue);
	invocations++;
	if (err != CL_SUCCESS)
	{
	    elog("clEnqueueNDRangeKernel error (%d)\n",err);
	    return hash_err;
	}
	//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
        int *fnd;
        fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
        found = *fnd;

        if (session_restore_lflag==1) ocl_bruteforce_init_restore(self, 5);
	session_restore_lflag = 0;
	if (found==1) 
	{
	    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*sizeof(uint), found_ind, 0, NULL, NULL);
	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    e=-1;
	    for (a=0;a<charset_size;a++)
	    for (b=0;b<charset_size;b++)
	    for (c=0;c<charset_size;c++)
	    if (found_ind[a*charset_size*charset_size+b*charset_size+c]==1)
	    {
    	    e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector;
    	    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, e*64, 64*ocl_vector, ptr+e*64, 0, NULL, NULL);
    	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    for (d=0;d<ocl_vector;d++)
	    {
    		e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector+d;
    		if ( (hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count<320) && (hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count>0))
    		{
    		    int i;
    		    for (i=0;i<hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count;i++)
    		    if (memcmp(hash_index[ptr[e*64]&255][ptr[e*64+1]&255].nodes[i]->hash, (char *)ptr+(e)*64, hash_ret_len) == 0)
    		    {
                        mylist = hash_index[ptr[e*64]&255][ptr[e*64+1]&255].nodes[i];
                        int flag = 0;
                        plain[0]=bruteforce_charset[a1];
                        plain[1]=bruteforce_charset[a2];
                        plain[2]=bruteforce_charset[a3];
                        plain[3]=bruteforce_charset[a4];
                        plain[4]=bruteforce_charset[a5];
                        plain[5]=bruteforce_charset[a6+(d)];
                        plain[6]=bruteforce_charset[c];
                        plain[7]=bruteforce_charset[b];
                        plain[8]=bruteforce_charset[a];
                        plain[9]=0;
                        pthread_mutex_lock(&crackedmutex);
                        addlist = cracked_list;
                        while (addlist)
                        {
                            if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                            addlist = addlist->next;
                        }
                        pthread_mutex_unlock(&crackedmutex);
                        if (flag == 0)
                        {
                            add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                        }
    		    }
    		}
    		else if (hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count>=320)
		{
    		    mylist = hash_list;
    		    while (mylist)
    		    {
            		if ( (memcmp(mylist->hash, (char *)ptr+(e)*64, hash_ret_len) == 0) )
            		{
                    	    int flag = 0;
                    	    plain[0]=bruteforce_charset[a1];
                    	    plain[1]=bruteforce_charset[a2];
                    	    plain[2]=bruteforce_charset[a3];
                    	    plain[3]=bruteforce_charset[a4];
                    	    plain[4]=bruteforce_charset[a5];
                    	    plain[5]=bruteforce_charset[a6+(d)];
                    	    plain[6]=bruteforce_charset[c];
                    	    plain[7]=bruteforce_charset[b];
                    	    plain[8]=bruteforce_charset[a];
                    	    plain[9]=0;
                    	    pthread_mutex_lock(&crackedmutex);
                    	    addlist = cracked_list;
                    	    while (addlist)
                    	    {
                        	if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                        	addlist = addlist->next;
                    	    }
                    	    pthread_mutex_unlock(&crackedmutex);
                    	    if (flag == 0)
                    	    {
                        	add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    	    }
            		}
            		mylist = mylist->next;
        	    }
		}
	    }
    	    }
	clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
        bzero(found_ind,charset_size*charset_size*charset_size*sizeof(cl_uint));
	clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
	}
    }


    // Bruteforce, len=10
    csize=7;
    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);
    if (session_restore_lflag==0) ocl_bruteforce_init_restore(self, 6);

    if ((bruteforce_start<=10) && (bruteforce_end>=10))
    if (thread_restore[self].cstart<=6)
    for (a1=thread_restore[self].cpos[0];a1<charset_size;a1++)
    for (a2=thread_restore[self].cpos[1];a2<charset_size;a2++)
    for (a3=thread_restore[self].cpos[2];a3<charset_size;a3++)
    for (a4=thread_restore[self].cpos[3];a4<charset_size;a4++)
    for (a5=thread_restore[self].cpos[4];a5<charset_size;a5++)
    for (a6=thread_restore[self].cpos[5];a6<charset_size;a6++)
    for (a7=thread_restore[self].cpos[6];a7<charset_size;a7+=(ocl_vector*ocl_threads*devicesnum))
    {
	cbase.s0 = bruteforce_charset[a7];
	cbase.s1 = bruteforce_charset[a7+1];
	cbase.s2 = bruteforce_charset[a7+2];
	cbase.s3 = bruteforce_charset[a7+3];

	clSetKernelArg(kernel, 3, sizeof(cl_ulong8), (void*) &cbase);

	thread_save[self].cpos[0]=a1;
	thread_save[self].cpos[1]=a2;
	thread_save[self].cpos[2]=a3;
	thread_save[self].cpos[3]=a4;
	thread_save[self].cpos[4]=a5;
	thread_save[self].cpos[5]=a6;
	thread_save[self].cpos[6]=a7;
	thread_save[self].cstr[0]=bruteforce_charset[a1];
	thread_save[self].cstr[1]=bruteforce_charset[a2];
	thread_save[self].cstr[2]=bruteforce_charset[a3];
	thread_save[self].cstr[3]=bruteforce_charset[a4];
	thread_save[self].cstr[4]=bruteforce_charset[a5];
	thread_save[self].cstr[5]=bruteforce_charset[a6];
	thread_save[self].cstr[6]=bruteforce_charset[a7];
	thread_save[self].cstr[7]=0;
	thread_save[self].cstart=6;

	if (attack_over!=0) return hash_ok;
	image.x=(bruteforce_charset[a1])|(bruteforce_charset[a2]<<8)|(bruteforce_charset[a3]<<16)|(bruteforce_charset[a4]<<24);
	image.y=(bruteforce_charset[a5])|(bruteforce_charset[a6]<<8);image.z=0;image.w=0;
	clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);

	if (found==1)
	{
    	    found = 0;
    	    clEnqueueWriteBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
    	}
        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
        //clFlush(queue);
	invocations++;
	if (err != CL_SUCCESS)
	{
	    elog("clEnqueueNDRangeKernel error (%d)\n",err);
	    return hash_err;
	}
	//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
        int *fnd;
        fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
        found = *fnd;

        if (session_restore_lflag==1) ocl_bruteforce_init_restore(self, 6);
	session_restore_lflag = 0;
	if (found==1) 
	{
	    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*sizeof(uint), found_ind, 0, NULL, NULL);
	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    e=-1;
	    for (a=0;a<charset_size;a++)
	    for (b=0;b<charset_size;b++)
	    for (c=0;c<charset_size;c++)
	    if (found_ind[a*charset_size*charset_size+b*charset_size+c]==1)
	    {
    	    e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector;
    	    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, e*64, 64*ocl_vector, ptr+e*64, 0, NULL, NULL);
    	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    for (d=0;d<ocl_vector;d++)
	    {
    		e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector+d;
    		if ( (hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count<640) && (hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count>0))
    		{
    		    int i;
    		    for (i=0;i<hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count;i++)
    		    if (memcmp(hash_index[ptr[e*64]&255][ptr[e*64+1]&255].nodes[i]->hash, (char *)ptr+(e)*64, hash_ret_len) == 0)
    		    {
                        mylist = hash_index[ptr[e*64]&255][ptr[e*64+1]&255].nodes[i];
                        int flag = 0;
                        plain[0]=bruteforce_charset[a1];
                        plain[1]=bruteforce_charset[a2];
                        plain[2]=bruteforce_charset[a3];
                        plain[3]=bruteforce_charset[a4];
                        plain[4]=bruteforce_charset[a5];
                        plain[5]=bruteforce_charset[a6];
                        plain[6]=bruteforce_charset[a7+(d)];
                        plain[7]=bruteforce_charset[c];
                        plain[8]=bruteforce_charset[b];
                        plain[9]=bruteforce_charset[a];
                        plain[10]=0;
                        pthread_mutex_lock(&crackedmutex);
                        addlist = cracked_list;
                        while (addlist)
                        {
                            if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                            addlist = addlist->next;
                        }
                        pthread_mutex_unlock(&crackedmutex);
                        if (flag == 0)
                        {
                            add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                        }
    		    }
    		}
    		else if (hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count>=320)
		{
    		    mylist = hash_list;
    		    while (mylist)
    		    {
            		if ( (memcmp(mylist->hash, (char *)ptr+(e)*64, hash_ret_len) == 0) )
            		{
                    	    int flag = 0;
                    	    plain[0]=bruteforce_charset[a1];
                    	    plain[1]=bruteforce_charset[a2];
                    	    plain[2]=bruteforce_charset[a3];
                    	    plain[3]=bruteforce_charset[a4];
                    	    plain[4]=bruteforce_charset[a5];
                    	    plain[5]=bruteforce_charset[a6];
                    	    plain[6]=bruteforce_charset[a7+(d)];
                    	    plain[7]=bruteforce_charset[c];
                    	    plain[8]=bruteforce_charset[b];
                    	    plain[9]=bruteforce_charset[a];
                    	    plain[10]=0;
                    	    pthread_mutex_lock(&crackedmutex);
                    	    addlist = cracked_list;
                    	    while (addlist)
                    	    {
                        	if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                        	addlist = addlist->next;
                    	    }
                    	    pthread_mutex_unlock(&crackedmutex);
                    	    if (flag == 0)
                    	    {
                        	add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    	    }
            		}
            		mylist = mylist->next;
        	    }
		}
	    }
    	    }
	clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
	bzero(found_ind,charset_size*charset_size*charset_size*sizeof(cl_uint));
    	clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
	}
    }


    // Bruteforce, len=11
    csize=8;
    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);
    if (session_restore_lflag==0) ocl_bruteforce_init_restore(self, 7);

    if ((bruteforce_start<=11) && (bruteforce_end>=11))
    if (thread_restore[self].cstart<=7)
    for (a1=thread_restore[self].cpos[0];a1<charset_size;a1++)
    for (a2=thread_restore[self].cpos[1];a2<charset_size;a2++)
    for (a3=thread_restore[self].cpos[2];a3<charset_size;a3++)
    for (a4=thread_restore[self].cpos[3];a4<charset_size;a4++)
    for (a5=thread_restore[self].cpos[4];a5<charset_size;a5++)
    for (a6=thread_restore[self].cpos[5];a6<charset_size;a6++)
    for (a7=thread_restore[self].cpos[6];a7<charset_size;a7++)
    for (a8=thread_restore[self].cpos[7];a8<charset_size;a8+=(ocl_vector*ocl_threads*devicesnum))
    {
	cbase.s0 = bruteforce_charset[a8];
	cbase.s1 = bruteforce_charset[a8+1];
	cbase.s2 = bruteforce_charset[a8+2];
	cbase.s3 = bruteforce_charset[a8+3];

	clSetKernelArg(kernel, 3, sizeof(cl_ulong8), (void*) &cbase);

	thread_save[self].cpos[0]=a1;
	thread_save[self].cpos[1]=a2;
	thread_save[self].cpos[2]=a3;
	thread_save[self].cpos[3]=a4;
	thread_save[self].cpos[4]=a5;
	thread_save[self].cpos[5]=a6;
	thread_save[self].cpos[6]=a7;
	thread_save[self].cpos[7]=a8;
	thread_save[self].cstr[0]=bruteforce_charset[a1];
	thread_save[self].cstr[1]=bruteforce_charset[a2];
	thread_save[self].cstr[2]=bruteforce_charset[a3];
	thread_save[self].cstr[3]=bruteforce_charset[a4];
	thread_save[self].cstr[4]=bruteforce_charset[a5];
	thread_save[self].cstr[5]=bruteforce_charset[a6];
	thread_save[self].cstr[6]=bruteforce_charset[a7];
	thread_save[self].cstr[7]=bruteforce_charset[a8];
	thread_save[self].cstr[8]=0;
	thread_save[self].cstart=7;

	if (attack_over!=0) return hash_ok;
	image.x=(bruteforce_charset[a1])|(bruteforce_charset[a2]<<8)|(bruteforce_charset[a3]<<16)|(bruteforce_charset[a4]<<24);
	image.y=(bruteforce_charset[a5])|(bruteforce_charset[a6]<<8)|(bruteforce_charset[a7]<<16);image.z=0;image.w=0;
	clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);

	if (found==1)
	{
    	    found = 0;
    	    clEnqueueWriteBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
    	}
        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
        //clFlush(queue);
	invocations++;
	if (err != CL_SUCCESS)
	{
	    elog("clEnqueueNDRangeKernel error (%d)\n",err);
	    return hash_err;
	}
	//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
        int *fnd;
        fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
        found = *fnd;

        if (session_restore_lflag==1) ocl_bruteforce_init_restore(self, 7);
	session_restore_lflag = 0;
	if (found==1) 
	{
	    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*sizeof(uint), found_ind, 0, NULL, NULL);
	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    e=-1;
	    for (a=0;a<charset_size;a++)
	    for (b=0;b<charset_size;b++)
	    for (c=0;c<charset_size;c++)
	    if (found_ind[a*charset_size*charset_size+b*charset_size+c]==1)
	    {
    	    e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector;
    	    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, e*64, 64*ocl_vector, ptr+e*64, 0, NULL, NULL);
    	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    for (d=0;d<ocl_vector;d++)
	    {
    		e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector+d;
    		if ( (hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count<320) && (hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count>0))
    		{
    		    int i;
    		    for (i=0;i<hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count;i++)
    		    if (memcmp(hash_index[ptr[e*64]&255][ptr[e*64+1]&255].nodes[i]->hash, (char *)ptr+(e)*64, hash_ret_len) == 0)
    		    {
                        mylist = hash_index[ptr[e*64]&255][ptr[e*64+1]&255].nodes[i];
                        int flag = 0;
                        plain[0]=bruteforce_charset[a1];
                        plain[1]=bruteforce_charset[a2];
                        plain[2]=bruteforce_charset[a3];
                        plain[3]=bruteforce_charset[a4];
                        plain[4]=bruteforce_charset[a5];
                        plain[5]=bruteforce_charset[a6];
                        plain[6]=bruteforce_charset[a7];
                        plain[7]=bruteforce_charset[a8+(d)];
                        plain[8]=bruteforce_charset[c];
                        plain[9]=bruteforce_charset[b];
                        plain[10]=bruteforce_charset[a];
                        plain[11]=0;
                        pthread_mutex_lock(&crackedmutex);
                        addlist = cracked_list;
                        while (addlist)
                        {
                            if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                            addlist = addlist->next;
                        }
                        pthread_mutex_unlock(&crackedmutex);
                        if (flag == 0)
                        {
                            add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                        }
    		    }
    		}
    		else if (hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count>=320)
		{
    		    mylist = hash_list;
    		    while (mylist)
    		    {
            		if ( (memcmp(mylist->hash, (char *)ptr+(e)*64, hash_ret_len) == 0) )
            		{
                    	    int flag = 0;
                    	    plain[0]=bruteforce_charset[a1];
                    	    plain[1]=bruteforce_charset[a2];
                    	    plain[2]=bruteforce_charset[a3];
                    	    plain[3]=bruteforce_charset[a4];
                    	    plain[4]=bruteforce_charset[a5];
                    	    plain[5]=bruteforce_charset[a6];
                    	    plain[6]=bruteforce_charset[a7];
                    	    plain[7]=bruteforce_charset[a8+(d)];
                    	    plain[8]=bruteforce_charset[c];
                    	    plain[9]=bruteforce_charset[b];
                    	    plain[10]=bruteforce_charset[a];
                    	    plain[11]=0;
                    	    pthread_mutex_lock(&crackedmutex);
                    	    addlist = cracked_list;
                    	    while (addlist)
                    	    {
                        	if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                        	addlist = addlist->next;
                    	    }
                    	    pthread_mutex_unlock(&crackedmutex);
                    	    if (flag == 0)
                    	    {
                        	add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    	    }
            		}
            		mylist = mylist->next;
        	    }
		}
	    }
    	    }
	clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
    	bzero(found_ind,charset_size*charset_size*charset_size*sizeof(cl_uint));
    	clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
	}
    }


    // Bruteforce, len=12
    csize=9;
    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);
    if (session_restore_lflag==0) ocl_bruteforce_init_restore(self, 8);

    if ((bruteforce_start<=12) && (bruteforce_end>=12))
    if (thread_restore[self].cstart<=8)
    for (a1=thread_restore[self].cpos[0];a1<charset_size;a1++)
    for (a2=thread_restore[self].cpos[1];a2<charset_size;a2++)
    for (a3=thread_restore[self].cpos[2];a3<charset_size;a3++)
    for (a4=thread_restore[self].cpos[3];a4<charset_size;a4++)
    for (a5=thread_restore[self].cpos[4];a5<charset_size;a5++)
    for (a6=thread_restore[self].cpos[5];a6<charset_size;a6++)
    for (a7=thread_restore[self].cpos[6];a7<charset_size;a7++)
    for (a8=thread_restore[self].cpos[7];a8<charset_size;a8++)
    for (a9=thread_restore[self].cpos[8];a9<charset_size;a9+=(ocl_vector*ocl_threads*devicesnum))
    {
	cbase.s0 = bruteforce_charset[a9];
	cbase.s1 = bruteforce_charset[a9+1];
	cbase.s2 = bruteforce_charset[a9+2];
	cbase.s3 = bruteforce_charset[a9+3];

	clSetKernelArg(kernel, 3, sizeof(cl_ulong8), (void*) &cbase);

	thread_save[self].cpos[0]=a1;
	thread_save[self].cpos[1]=a2;
	thread_save[self].cpos[2]=a3;
	thread_save[self].cpos[3]=a4;
	thread_save[self].cpos[4]=a5;
	thread_save[self].cpos[5]=a6;
	thread_save[self].cpos[6]=a7;
	thread_save[self].cpos[7]=a8;
	thread_save[self].cpos[8]=a9;
	thread_save[self].cstr[0]=bruteforce_charset[a1];
	thread_save[self].cstr[1]=bruteforce_charset[a2];
	thread_save[self].cstr[2]=bruteforce_charset[a3];
	thread_save[self].cstr[3]=bruteforce_charset[a4];
	thread_save[self].cstr[4]=bruteforce_charset[a5];
	thread_save[self].cstr[5]=bruteforce_charset[a6];
	thread_save[self].cstr[6]=bruteforce_charset[a7];
	thread_save[self].cstr[7]=bruteforce_charset[a8];
	thread_save[self].cstr[8]=bruteforce_charset[a9];
	thread_save[self].cstr[9]=0;
	thread_save[self].cstart=8;

	if (attack_over!=0) return hash_ok;
	image.x=(bruteforce_charset[a1])|(bruteforce_charset[a2]<<8)|(bruteforce_charset[a3]<<16)|(bruteforce_charset[a4]<<24);
	image.y=(bruteforce_charset[a5])|(bruteforce_charset[a6]<<8)|(bruteforce_charset[a7]<<16)|(bruteforce_charset[a8]<<24);image.z=0;image.w=0;
	clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);

	if (found==1)
	{
    	    found = 0;
    	    clEnqueueWriteBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
    	}
        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
        //clFlush(queue);
	invocations++;
	if (err != CL_SUCCESS)
	{
	    elog("clEnqueueNDRangeKernel error (%d)\n",err);
	    return hash_err;
	}
	//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
        int *fnd;
        fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
        found = *fnd;

        if (session_restore_lflag==1) ocl_bruteforce_init_restore(self, 8);
	session_restore_lflag = 0;
	if (found==1) 
	{
	    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, 0, charset_size*charset_size*charset_size*64*ocl_vector, ptr, 0, NULL, NULL);
	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    e=-1;
	    for (a=0;a<charset_size;a++)
	    for (b=0;b<charset_size;b++)
	    for (c=0;c<charset_size;c++)
	    if (found_ind[a*charset_size*charset_size+b*charset_size+c]==1)
	    {
    	    e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector;
	    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*sizeof(uint), found_ind, 0, NULL, NULL);
    	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    for (d=0;d<ocl_vector;d++)
	    {
    		e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector+d;
    		if ( (hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count<320) && (hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count>0))
    		{
    		    int i;
    		    for (i=0;i<hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count;i++)
    		    if (memcmp(hash_index[ptr[e*64]&255][ptr[e*64+1]&255].nodes[i]->hash, (char *)ptr+(e)*64, hash_ret_len) == 0)
    		    {
                        mylist = hash_index[ptr[e*64]&255][ptr[e*64+1]&255].nodes[i];
                        int flag = 0;
                        plain[0]=bruteforce_charset[a1];
                        plain[1]=bruteforce_charset[a2];
                        plain[2]=bruteforce_charset[a3];
                        plain[3]=bruteforce_charset[a4];
                        plain[4]=bruteforce_charset[a5];
                        plain[5]=bruteforce_charset[a6];
                        plain[6]=bruteforce_charset[a7];
                        plain[7]=bruteforce_charset[a8];
                        plain[8]=bruteforce_charset[a9+(d)];
                        plain[9]=bruteforce_charset[c];
                        plain[10]=bruteforce_charset[b];
                        plain[11]=bruteforce_charset[a];
                        plain[12]=0;
                        pthread_mutex_lock(&crackedmutex);
                        addlist = cracked_list;
                        while (addlist)
                        {
                            if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                            addlist = addlist->next;
                        }
                        pthread_mutex_unlock(&crackedmutex);
                        if (flag == 0)
                        {
                            add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                        }
    		    }
    		}
    		else if (hash_index[ptr[e*64]&255][ptr[e*64+1]&255].count>=320)
		{
    		    mylist = hash_list;
    		    while (mylist)
    		    {
            		if ( (memcmp(mylist->hash, (char *)ptr+(e)*64, hash_ret_len) == 0) )
            		{
                    	    int flag = 0;
                    	    plain[0]=bruteforce_charset[a1];
                    	    plain[1]=bruteforce_charset[a2];
                    	    plain[2]=bruteforce_charset[a3];
                    	    plain[3]=bruteforce_charset[a4];
                    	    plain[4]=bruteforce_charset[a5];
                    	    plain[5]=bruteforce_charset[a6];
                    	    plain[6]=bruteforce_charset[a7];
                    	    plain[7]=bruteforce_charset[a8];
                    	    plain[8]=bruteforce_charset[a9+(d)];
                    	    plain[9]=bruteforce_charset[c];
                    	    plain[10]=bruteforce_charset[b];
                    	    plain[11]=bruteforce_charset[a];
                    	    plain[12]=0;
                    	    pthread_mutex_lock(&crackedmutex);
                    	    addlist = cracked_list;
                    	    while (addlist)
                    	    {
                        	if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                        	addlist = addlist->next;
                    	    }
                    	    pthread_mutex_unlock(&crackedmutex);
                    	    if (flag == 0)
                    	    {
                        	add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    	    }
            		}
            		mylist = mylist->next;
        	    }
		}
	    }
    	    }
	clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
    	bzero(found_ind,charset_size*charset_size*charset_size*sizeof(cl_uint));
    	clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
	}
    }

    return hash_ok;
}




hash_stat ocl_bruteforce_sha512_short_thread(void *arg)
{
    int err;
    cl_command_queue queue;
    cl_mem buffer;
    size_t global_work_size;
    cl_uint4 image;
    int self;
    cl_kernel kernel;
    unsigned char hex1[16];
    char plain[64];
    cl_mem bitmaps_buf;
    cl_uint *bitmaps;
    int a,b,c,d,e,f;
    int a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11;
    char *ptr = malloc(strlen(bruteforce_charset)*strlen(bruteforce_charset)*strlen(bruteforce_charset)*strlen(bruteforce_charset)*32*ocl_vector);
    int charset_size = (int)strlen(bruteforce_charset);
    cl_mem found_ind_buf;
    cl_uint *found_ind=malloc(strlen(bruteforce_charset)*strlen(bruteforce_charset)*strlen(bruteforce_charset)*strlen(bruteforce_charset)*sizeof(cl_uint));
    int found;
    cl_mem found_buf;
    struct  hash_list_s  *mylist, *addlist;
    unsigned int csize=1;
    cl_mem table_buf;
    unsigned int *table;
    cl_uint8 cbase;
    cl_uint4 singlehash;
    int session_restore_lflag;

    pthread_mutex_lock(&biglock);
    memcpy(&self,arg,sizeof(int));
    bitmaps = malloc(256*256*32*sizeof(cl_uint));
    session_restore_lflag = session_restore_flag;
    kernel = clCreateKernel(program[self], "sha512_short", &err );
    if (err != CL_SUCCESS)
    {
	elog("clCreateKernel error (%d)\n",err);
	return hash_err;
    }
    queue = clCreateCommandQueue(context[self], device[self/ocl_threads], 0, &err );
    if (err != CL_SUCCESS)
    {
        elog("clCreateCommandQueue error (%d)\n",err);
	return hash_err;
    }
    buffer = clCreateBuffer(context[self], CL_MEM_WRITE_ONLY, charset_size*charset_size*charset_size*charset_size*32*ocl_vector, NULL, &err );
    if (err != CL_SUCCESS)
    {
	elog("clCreateBuffer error (%d)\n",err);
	return hash_err;
    }

    bzero(found_ind,sizeof(uint)*charset_size*charset_size*charset_size*charset_size);
    found_ind_buf = clCreateBuffer(context[self], CL_MEM_WRITE_ONLY, charset_size*charset_size*charset_size*charset_size*sizeof(cl_uint), NULL, &err );
    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    if (err != CL_SUCCESS)
    {
	elog("clCreateBuffer error (%d)\n",err);
	return hash_err;
    }
    clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);

    found_buf = clCreateBuffer(context[self], CL_MEM_WRITE_ONLY, 4, NULL, &err );
    if (err != CL_SUCCESS)
    {
	elog("clCreateBuffer error (%d)\n",err);
	return hash_err;
    }
    table = malloc(sizeof(uint)*charset_size*charset_size*charset_size*charset_size);
    for (a=0;a<charset_size*charset_size*charset_size*charset_size;a++)
    {
	unsigned char x,y,z,w,t1,t2;
	t1 = (a / (charset_size*charset_size*charset_size));
	x = (bruteforce_charset[t1]);
	t2 = ((a-t1*(charset_size*charset_size*charset_size))/(charset_size*charset_size));
	y = (bruteforce_charset[t2]);
	z = (bruteforce_charset[((a-t1*(charset_size*charset_size*charset_size)) - t2*charset_size*charset_size)/charset_size]);
	w = (bruteforce_charset[(a % charset_size)]);
	table[a] = (x<<24)|(y<<16)|(z<<8)|w;
    }
    for (a=0;a<256*256*32;a++)
    {
	    bitmaps[a]=0;
    }

    mylist = hash_list;
    a=0;
    while (mylist)
    {
        memcpy(hex1,mylist->hash,16);
	a++;
	unsigned int *b_a = (unsigned int *)hex1;
	unsigned int *b_b = (unsigned int *)&hex1[4];
	unsigned int *b_c = (unsigned int *)&hex1[8];
	unsigned int *b_d = (unsigned int *)&hex1[12];
	unsigned int bind_a = (*b_a)>>13;
	unsigned int bval_a = (1<<((*b_a)&31));
	unsigned int bind_b = (*b_b)>>13;
	unsigned int bval_b = (1<<((*b_b)&31));
	unsigned int bind_c = (*b_c)>>13;
	unsigned int bval_c = (1<<((*b_c)&31));
	unsigned int bind_d = (*b_d)>>13;
	unsigned int bval_d = (1<<((*b_d)&31));
	bitmaps[bind_a] |=bval_a;
	bitmaps[bind_b+65535*8] |=bval_b;
	bitmaps[bind_c+65535*16] |=bval_c;
	bitmaps[bind_d+65535*24] |=bval_d;
	singlehash.x |= (1<<((*b_b)&31));
	singlehash.y |= (1<<((*b_c)&31));
	singlehash.z |= (1<<((*b_d)&31));
        if (mylist) mylist = mylist->next;
    }

    if (a==1)
    {
	mylist = hash_list;
	memcpy(hex1,mylist->hash,4);
	memcpy(&singlehash.x, hex1, 4);
	memcpy(hex1,mylist->hash+4,4);
	memcpy(&singlehash.y, hex1, 4);
	memcpy(hex1,mylist->hash+8,4);
	memcpy(&singlehash.z, hex1, 4);
	memcpy(hex1,mylist->hash+12,4);
	memcpy(&singlehash.w, hex1, 4);
	clSetKernelArg(kernel, 8, sizeof(cl_uint4), (void*) &singlehash);
    }
    else 
    {
	singlehash.w=0;
	clSetKernelArg(kernel, 8, sizeof(cl_uint4), (void*) &singlehash);
    }

    cbase.s0 = bruteforce_charset[0];
    cbase.s1 = bruteforce_charset[1];
    cbase.s2 = bruteforce_charset[2];
    cbase.s3 = bruteforce_charset[3];
    cbase.s4 = bruteforce_charset[4];
    cbase.s5 = bruteforce_charset[5];
    cbase.s6 = bruteforce_charset[6];
    cbase.s7 = bruteforce_charset[7];

    table_buf = clCreateBuffer(context[self], CL_MEM_READ_ONLY|CL_MEM_USE_HOST_PTR, sizeof(uint)*charset_size*charset_size*charset_size*charset_size, table, &err );
    if (err != CL_SUCCESS)
    {
	elog("clCreateBuffer error (%d)\n",err);
	return hash_err;
    }
    bitmaps_buf = clCreateBuffer(context[self], CL_MEM_READ_ONLY|CL_MEM_USE_HOST_PTR, 256*256*32*4, bitmaps, &err );
    if (err != CL_SUCCESS)
    {
	elog("clCreateBuffer error (%d)\n",err);
	return hash_err;
    }



    clSetKernelArg(kernel, 0, sizeof(cl_mem), (void*) &buffer);
    clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);
    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);
    clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
    clSetKernelArg(kernel, 4, sizeof(cl_mem), (void*) &found_ind_buf);
    clSetKernelArg(kernel, 5, sizeof(cl_mem), (void*) &bitmaps_buf);
    clSetKernelArg(kernel, 6, sizeof(cl_mem), (void*) &found_buf);
    clSetKernelArg(kernel, 7, sizeof(cl_mem), (void*) &table_buf);
    global_work_size = charset_size*charset_size*charset_size*charset_size;
    while ((global_work_size %256)!=0) global_work_size++;

    if (err != CL_SUCCESS)
    {
	elog("clEnqueueBuffer error (%d)\n",err);
	return hash_err;
    }
    image.x=image.y=image.z=image.w=0;


    /* Bruteforce, len=5 */
    csize=1;
    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);
    clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);
    pthread_mutex_unlock(&biglock); // right now it should be safe to release the mutex
    if (session_restore_lflag==0) ocl_bruteforce_init_restore(self, 0);
    if ((bruteforce_start<=5) && (bruteforce_end>=5))
    if (thread_restore[self].cstart<=0)
    for (a1=thread_restore[self].cpos[0];a1<charset_size;a1+=(ocl_vector*ocl_threads*devicesnum))
    {
	if (attack_over!=0) return hash_ok;

	cbase.s0 = bruteforce_charset[a1];
	cbase.s1 = bruteforce_charset[a1+1];
	cbase.s2 = bruteforce_charset[a1+2];
	cbase.s3 = bruteforce_charset[a1+3];
	cbase.s4 = bruteforce_charset[a1+4];
	cbase.s5 = bruteforce_charset[a1+5];
	cbase.s6 = bruteforce_charset[a1+6];
	cbase.s7 = bruteforce_charset[a1+7];

	thread_save[self].cpos[0]=a1;
	thread_save[self].cstr[0]=bruteforce_charset[a1];
	thread_save[self].cstr[1]=0;
	thread_save[self].cstart=0;

	clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
	image.x=0;image.y=0;image.z=0;image.w=0;
        if (found==1)
        {
    	    found = 0;
    	    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    	}
        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
        //clFlush(queue);
	invocations++;
	if (err != CL_SUCCESS)
	{
	    elog("clEnqueueNDRangeKernel error (%d)\n",err);
	    return hash_err;
	}
	//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
        int *fnd;
        fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
        found = *fnd;

        if (session_restore_lflag==1) ocl_bruteforce_init_restore(self, 0);
	session_restore_lflag = 0;
	if (found==1) 
	{
	    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*charset_size*sizeof(uint), found_ind, 0, NULL, NULL);
	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    for (a=0;a<charset_size;a++)
	    for (b=0;b<charset_size;b++)
	    for (c=0;c<charset_size;c++)
	    for (d=0;d<charset_size;d++)
	    if (found_ind[a*charset_size*charset_size*charset_size+b*charset_size*charset_size+c*charset_size+d]==1)
	    {
	    f=(a*charset_size*charset_size*charset_size+b*charset_size*charset_size+c*charset_size+d)*ocl_vector;
    	    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, f*32, 32*ocl_vector, ptr+f*32, 0, NULL, NULL);
	    for (e=0;e<ocl_vector;e++)
	    {
    		f=(a*charset_size*charset_size*charset_size+b*charset_size*charset_size+c*charset_size+d)*ocl_vector+e;
    		if ( (hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count<320) && (hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count>0))
    		{
    		    int i;
    		    for (i=0;i<hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count;i++)
    		    if (memcmp(hash_index[ptr[f*32]&255][ptr[f*32+1]&255].nodes[i]->hash, (char *)ptr+(f)*32, hash_ret_len) == 0)
    		    {
                        mylist = hash_index[ptr[f*32]&255][ptr[f*32+1]&255].nodes[i];
                        int flag = 0;
                        plain[0]=bruteforce_charset[(a1+(e))%charset_size];
                        plain[1]=bruteforce_charset[d];
                        plain[2]=bruteforce_charset[c];
                        plain[3]=bruteforce_charset[b];
                        plain[4]=bruteforce_charset[a];
                        plain[5]=0;
                        pthread_mutex_lock(&crackedmutex);
                        addlist = cracked_list;
                        while (addlist)
                        {
                            if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                            addlist = addlist->next;
                        }
                        pthread_mutex_unlock(&crackedmutex);
                        if (flag == 0)
                        {
                            add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                        }
    		    }
    		}
    		else if (hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count>=320)
    		{
    		    mylist = hash_list;
    		    while (mylist)
    		    {
            		if (memcmp(mylist->hash, (char *)ptr+(f)*32, hash_ret_len) == 0)
            		{
                    	    int flag = 0;
                    	    plain[0]=bruteforce_charset[(a1+(e))%charset_size];
                    	    plain[1]=bruteforce_charset[d];
                    	    plain[2]=bruteforce_charset[c];
                    	    plain[3]=bruteforce_charset[b];
                    	    plain[4]=bruteforce_charset[a];
                    	    plain[5]=0;
                    	    pthread_mutex_lock(&crackedmutex);
                    	    addlist = cracked_list;
                    	    while (addlist)
                    	    {
                        	if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                        	addlist = addlist->next;
                    	    }
                    	    pthread_mutex_unlock(&crackedmutex);
                    	    if (flag == 0)
                    	    {
                        	add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    	    }
            		}
            		if (mylist) mylist = mylist->next;
        	    }
        	}
	    }
	    }
	clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
	bzero(found_ind,charset_size*charset_size*charset_size*charset_size*sizeof(cl_uint));
	clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
	}
    }


    /* Bruteforce, len=6 */
    csize=2;
    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);
    if (session_restore_lflag==0) ocl_bruteforce_init_restore(self, 1);
    if ((bruteforce_start<=6) && (bruteforce_end>=6))
    if (thread_restore[self].cstart<=1)
    for (a1=thread_restore[self].cpos[0];a1<charset_size;a1++)
    for (a2=thread_restore[self].cpos[1];a2<charset_size+ocl_vector;a2+=(ocl_vector*ocl_threads*devicesnum))
    {
	if (attack_over!=0) return hash_ok;
	cbase.s0 = bruteforce_charset[a2];
	cbase.s1 = bruteforce_charset[a2+1];
	cbase.s2 = bruteforce_charset[a2+2];
	cbase.s3 = bruteforce_charset[a2+3];
	cbase.s4 = bruteforce_charset[a2+4];
	cbase.s5 = bruteforce_charset[a2+5];
	cbase.s6 = bruteforce_charset[a2+6];
	cbase.s7 = bruteforce_charset[a2+7];
	thread_save[self].cpos[0]=a1;
	thread_save[self].cpos[1]=a2;
	thread_save[self].cstr[0]=bruteforce_charset[a1];
	thread_save[self].cstr[1]=bruteforce_charset[a2];
	thread_save[self].cstr[2]=0;
	thread_save[self].cstart=1;

	clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
	image.x=bruteforce_charset[a1];image.y=0;image.z=0;image.w=0;
	clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);
        if (found==1)
        {
    	    found = 0;
    	    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    	}
        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
        //clFlush(queue);
	invocations++;
	if (err != CL_SUCCESS)
	{
	    elog("clEnqueueNDRangeKernel error (%d)\n",err);
	    return hash_err;
	}
	//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
        int *fnd;
        fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
        found = *fnd;

        if (session_restore_lflag==1) ocl_bruteforce_init_restore(self, 1);
	session_restore_lflag = 0;

	if (found==1) 
	{
	    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*charset_size*sizeof(uint), found_ind, 0, NULL, NULL);
	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    for (a=0;a<charset_size;a++)
	    for (b=0;b<charset_size;b++)
	    for (c=0;c<charset_size;c++)
	    for (d=0;d<charset_size;d++)
	    if (found_ind[a*charset_size*charset_size*charset_size+b*charset_size*charset_size+c*charset_size+d]==1)
	    {
	    f=(a*charset_size*charset_size*charset_size+b*charset_size*charset_size+c*charset_size+d)*ocl_vector;
    	    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, f*32, 32*ocl_vector, ptr+f*32, 0, NULL, NULL);
	    for (e=0;e<ocl_vector;e++)
	    {
    		f=(a*charset_size*charset_size*charset_size+b*charset_size*charset_size+c*charset_size+d)*ocl_vector+e;
    		if ( (hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count<320) && (hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count>0))
    		{
    		    int i;
    		    for (i=0;i<hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count;i++)
    		    if (memcmp(hash_index[ptr[f*32]&255][ptr[f*32+1]&255].nodes[i]->hash, (char *)ptr+(f)*32, hash_ret_len) == 0)
    		    {
                        mylist = hash_index[ptr[f*32]&255][ptr[f*32+1]&255].nodes[i];
                        int flag = 0;
                        plain[0]=bruteforce_charset[a1];
                        plain[1]=bruteforce_charset[(a2+(e))%charset_size];
                        plain[2]=bruteforce_charset[d];
                        plain[3]=bruteforce_charset[c];
                        plain[4]=bruteforce_charset[b];
                        plain[5]=bruteforce_charset[a];
                        plain[6]=0;
                        pthread_mutex_lock(&crackedmutex);
                        addlist = cracked_list;
                        while (addlist)
                        {
                            if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                            addlist = addlist->next;
                        }
                        pthread_mutex_unlock(&crackedmutex);
                        if (flag == 0)
                        {
                            add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                        }
    		    }
    		}
    		else if (hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count>=320)
    		{
    		    mylist = hash_list;
    		    while (mylist)
    		    {
            		if (memcmp(mylist->hash, (char *)ptr+(f)*32, hash_ret_len) == 0)
            		{
                    	    int flag = 0;
                    	    plain[0]=bruteforce_charset[a1];
                    	    plain[1]=bruteforce_charset[(a2+(e))%charset_size];
                    	    plain[2]=bruteforce_charset[d];
                    	    plain[3]=bruteforce_charset[c];
                    	    plain[4]=bruteforce_charset[b];
                    	    plain[5]=bruteforce_charset[a];
                    	    plain[6]=0;
                    	    pthread_mutex_lock(&crackedmutex);
                    	    addlist = cracked_list;
                    	    while (addlist)
                    	    {
                        	if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                        	addlist = addlist->next;
                    	    }
                    	    pthread_mutex_unlock(&crackedmutex);
                    	    if (flag == 0)
                    	    {
                        	add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    	    }
            		}
            		if (mylist) mylist = mylist->next;
        	    }
        	}
	    }
	    }
	clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
	bzero(found_ind,charset_size*charset_size*charset_size*charset_size*sizeof(cl_uint));
	clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
	}
    }


    /* Bruteforce, len=7 */
    csize=3;
    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);
    if (session_restore_lflag==0) ocl_bruteforce_init_restore(self, 2);
    if ((bruteforce_start<=7) && (bruteforce_end>=7))
    if (thread_restore[self].cstart<=2)
    for (a1=thread_restore[self].cpos[0];a1<charset_size;a1++)
    for (a2=thread_restore[self].cpos[1];a2<charset_size;a2++)
    for (a3=thread_restore[self].cpos[2];a3<charset_size+ocl_vector;a3+=(ocl_vector*ocl_threads*devicesnum))
    {
	if (attack_over!=0) return hash_ok;
	cbase.s0 = bruteforce_charset[a3];
	cbase.s1 = bruteforce_charset[a3+1];
	cbase.s2 = bruteforce_charset[a3+2];
	cbase.s3 = bruteforce_charset[a3+3];
	cbase.s4 = bruteforce_charset[a3+4];
	cbase.s5 = bruteforce_charset[a3+5];
	cbase.s6 = bruteforce_charset[a3+6];
	cbase.s7 = bruteforce_charset[a3+7];
	thread_save[self].cpos[0]=a1;
	thread_save[self].cpos[1]=a2;
	thread_save[self].cpos[2]=a3;
	thread_save[self].cstr[0]=bruteforce_charset[a1];
	thread_save[self].cstr[1]=bruteforce_charset[a2];
	thread_save[self].cstr[2]=bruteforce_charset[a3];
	thread_save[self].cstr[3]=0;
	thread_save[self].cstart=2;

	clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
	image.x=bruteforce_charset[a1]|(bruteforce_charset[a2]<<8);image.y=0;image.z=0;image.w=0;
	clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);
        if (found==1)
        {
    	    found = 0;
    	    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    	}
        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
        //clFlush(queue);
	invocations++;
	if (err != CL_SUCCESS)
	{
	    elog("clEnqueueNDRangeKernel error (%d)\n",err);
	    return hash_err;
	}
	//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
        int *fnd;
        fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
        found = *fnd;

        if (session_restore_lflag==1) ocl_bruteforce_init_restore(self, 2);
	session_restore_lflag = 0;
	if (found==1) 
	{
	    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*charset_size*sizeof(uint), found_ind, 0, NULL, NULL);
	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    f=-1;
	    for (a=0;a<charset_size;a++)
	    for (b=0;b<charset_size;b++)
	    for (c=0;c<charset_size;c++)
	    for (d=0;d<charset_size;d++)
	    if (found_ind[a*charset_size*charset_size*charset_size+b*charset_size*charset_size+c*charset_size+d]==1)
	    {
	    f=(a*charset_size*charset_size*charset_size+b*charset_size*charset_size+c*charset_size+d)*ocl_vector;
    	    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, f*32, 32*ocl_vector, ptr+f*32, 0, NULL, NULL);
	    for (e=0;e<ocl_vector;e++)
	    {
    		f=(a*charset_size*charset_size*charset_size+b*charset_size*charset_size+c*charset_size+d)*ocl_vector+e;
    		if ( (hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count<320) && (hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count>0))
    		{
    		    int i;
    		    for (i=0;i<hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count;i++)
    		    if (memcmp(hash_index[ptr[f*32]&255][ptr[f*32+1]&255].nodes[i]->hash, (char *)ptr+(f)*32, hash_ret_len) == 0)
    		    {
                        mylist = hash_index[ptr[f*32]&255][ptr[f*32+1]&255].nodes[i];
                        int flag = 0;
                        plain[0]=bruteforce_charset[a1];
                        plain[1]=bruteforce_charset[a2];
                        plain[2]=bruteforce_charset[(a3+(e))%charset_size];
                        plain[3]=bruteforce_charset[d];
                        plain[4]=bruteforce_charset[c];
                        plain[5]=bruteforce_charset[b];
                        plain[6]=bruteforce_charset[a];
                        plain[7]=0;
                        pthread_mutex_lock(&crackedmutex);
                        addlist = cracked_list;
                        while (addlist)
                        {
                            if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                            addlist = addlist->next;
                        }
                        pthread_mutex_unlock(&crackedmutex);
                        if (flag == 0)
                        {
                            add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                        }
    		    }
    		}
    		else if (hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count>=320)
    		{
    		    mylist = hash_list;
    		    while (mylist)
    		    {
            		if (memcmp(mylist->hash, (char *)ptr+(f)*32, hash_ret_len) == 0)
            		{
                    	    int flag = 0;
                    	    plain[0]=bruteforce_charset[a1];
                    	    plain[1]=bruteforce_charset[a2];
                    	    plain[2]=bruteforce_charset[(a3+(e))%charset_size];
                    	    plain[3]=bruteforce_charset[d];
                    	    plain[4]=bruteforce_charset[c];
                    	    plain[5]=bruteforce_charset[b];
                    	    plain[6]=bruteforce_charset[a];
                    	    plain[7]=0;
                    	    pthread_mutex_lock(&crackedmutex);
                    	    addlist = cracked_list;
                    	    while (addlist)
                    	    {
                        	if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                        	addlist = addlist->next;
                    	    }
                    	    pthread_mutex_unlock(&crackedmutex);
                    	    if (flag == 0)
                    	    {
                        	add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    	    }
            		}
            		if (mylist) mylist = mylist->next;
        	    }
        	}
	    }
	    }
	clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
	bzero(found_ind,charset_size*charset_size*charset_size*charset_size*sizeof(cl_uint));
	clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
	}
    }


    /* Bruteforce, len=8 */
    csize=4;
    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);
    if (session_restore_lflag==0) ocl_bruteforce_init_restore(self, 3);
    if ((bruteforce_start<=8) && (bruteforce_end>=8))
    if (thread_restore[self].cstart<=3)
    for (a1=thread_restore[self].cpos[0];a1<charset_size;a1++)
    for (a2=thread_restore[self].cpos[1];a2<charset_size;a2++)
    for (a3=thread_restore[self].cpos[2];a3<charset_size;a3++)
    for (a4=thread_restore[self].cpos[3];a4<charset_size+ocl_vector;a4+=(ocl_vector*ocl_threads*devicesnum))
    {
	if (attack_over!=0) return hash_ok;
	cbase.s0 = bruteforce_charset[a4];
	cbase.s1 = bruteforce_charset[a4+1];
	cbase.s2 = bruteforce_charset[a4+2];
	cbase.s3 = bruteforce_charset[a4+3];
	cbase.s4 = bruteforce_charset[a4+4];
	cbase.s5 = bruteforce_charset[a4+5];
	cbase.s6 = bruteforce_charset[a4+6];
	cbase.s7 = bruteforce_charset[a4+7];
	thread_save[self].cpos[0]=a1;
	thread_save[self].cpos[1]=a2;
	thread_save[self].cpos[2]=a3;
	thread_save[self].cpos[3]=a4;
	thread_save[self].cstr[0]=bruteforce_charset[a1];
	thread_save[self].cstr[1]=bruteforce_charset[a2];
	thread_save[self].cstr[2]=bruteforce_charset[a3];
	thread_save[self].cstr[3]=bruteforce_charset[a4];
	thread_save[self].cstr[4]=0;
	thread_save[self].cstart=3;

	clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
	image.x=bruteforce_charset[a1]|(bruteforce_charset[a2]<<8)|(bruteforce_charset[a3]<<16);image.y=0;image.z=0;image.w=0;
	clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);
        if (found==1)
        {
    	    found = 0;
    	    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    	}
        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
        //clFlush(queue);
	invocations++;
	if (err != CL_SUCCESS)
	{
	    elog("clEnqueueNDRangeKernel error (%d)\n",err);
	    return hash_err;
	}
	//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
        int *fnd;
        fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
        found = *fnd;

        if (session_restore_lflag==1) ocl_bruteforce_init_restore(self, 3);
	session_restore_lflag = 0;
	if (found==1) 
	{
	    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*charset_size*sizeof(uint), found_ind, 0, NULL, NULL);
	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    f=-1;
	    for (a=0;a<charset_size;a++)
	    for (b=0;b<charset_size;b++)
	    for (c=0;c<charset_size;c++)
	    for (d=0;d<charset_size;d++)
	    if (found_ind[a*charset_size*charset_size*charset_size+b*charset_size*charset_size+c*charset_size+d]==1)
	    {
	    f=(a*charset_size*charset_size*charset_size+b*charset_size*charset_size+c*charset_size+d)*ocl_vector;
    	    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, f*32, 32*ocl_vector, ptr+f*32, 0, NULL, NULL);
	    for (e=0;e<ocl_vector;e++)
	    {
    		f=(a*charset_size*charset_size*charset_size+b*charset_size*charset_size+c*charset_size+d)*ocl_vector+e;
    		if ( (hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count<320) && (hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count>0))
    		{
    		    int i;
    		    for (i=0;i<hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count;i++)
    		    if (memcmp(hash_index[ptr[f*32]&255][ptr[f*32+1]&255].nodes[i]->hash, (char *)ptr+(f)*32, hash_ret_len) == 0)
    		    {
                        mylist = hash_index[ptr[f*32]&255][ptr[f*32+1]&255].nodes[i];
                        int flag = 0;
                        plain[0]=bruteforce_charset[a1];
                        plain[1]=bruteforce_charset[a2];
                        plain[2]=bruteforce_charset[a3];
                        plain[3]=bruteforce_charset[(a4+(e))%charset_size];
                        plain[4]=bruteforce_charset[d];
                        plain[5]=bruteforce_charset[c];
                        plain[6]=bruteforce_charset[b];
                        plain[7]=bruteforce_charset[a];
                        plain[8]=0;
                        pthread_mutex_lock(&crackedmutex);
                        addlist = cracked_list;
                        while (addlist)
                        {
                            if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
                                && (strcmp(addlist->salt, mylist->salt) == 0))
                                    flag = 1;
                            addlist = addlist->next;
                        }
                        pthread_mutex_unlock(&crackedmutex);
                        if (flag == 0)
                        {
                            add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                        }
    		    }
    		}
    		else if (hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count>=320)
    		{
    		    mylist = hash_list;
    		    while (mylist)
    		    {
            		if (memcmp(mylist->hash, (char *)ptr+(f)*32, hash_ret_len) == 0)
            		{
                    	    int flag = 0;
                    	    plain[0]=bruteforce_charset[a1];
                    	    plain[1]=bruteforce_charset[a2];
                    	    plain[2]=bruteforce_charset[a3];
                    	    plain[3]=bruteforce_charset[(a4+(e))%charset_size];
                    	    plain[4]=bruteforce_charset[d];
                    	    plain[5]=bruteforce_charset[c];
                    	    plain[6]=bruteforce_charset[b];
                    	    plain[7]=bruteforce_charset[a];
                    	    plain[8]=0;
                    	    pthread_mutex_lock(&crackedmutex);
                    	    addlist = cracked_list;
                    	    while (addlist)
                    	    {
                        	if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
                                && (strcmp(addlist->salt, mylist->salt) == 0))
                                    flag = 1;
                        	addlist = addlist->next;
                    	    }
                    	    pthread_mutex_unlock(&crackedmutex);
                    	    if (flag == 0)
                    	    {
                        	add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    	    }
            		}
            		if (mylist) mylist = mylist->next;
        	    }
        	}
	    }
	    }
	clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
	bzero(found_ind,charset_size*charset_size*charset_size*charset_size*sizeof(cl_uint));
	clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
	}
    }


    /* Bruteforce, len=9 */
    csize=5;
    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);
    if (session_restore_lflag==0) ocl_bruteforce_init_restore(self, 4);
    if ((bruteforce_start<=9) && (bruteforce_end>=9))
    if (thread_restore[self].cstart<=4)
    for (a1=thread_restore[self].cpos[0];a1<charset_size;a1++)
    for (a2=thread_restore[self].cpos[1];a2<charset_size;a2++)
    for (a3=thread_restore[self].cpos[2];a3<charset_size;a3++)
    for (a4=thread_restore[self].cpos[3];a4<charset_size;a4++)
    for (a5=thread_restore[self].cpos[4];a5<charset_size;a5+=(ocl_vector*ocl_threads*devicesnum))
    {
	if (attack_over!=0) return hash_ok;
	cbase.s0 = bruteforce_charset[a5];
	cbase.s1 = bruteforce_charset[a5+1];
	cbase.s2 = bruteforce_charset[a5+2];
	cbase.s3 = bruteforce_charset[a5+3];
	cbase.s4 = bruteforce_charset[a5+4];
	cbase.s5 = bruteforce_charset[a5+5];
	cbase.s6 = bruteforce_charset[a5+6];
	cbase.s7 = bruteforce_charset[a5+7];
	thread_save[self].cpos[0]=a1;
	thread_save[self].cpos[1]=a2;
	thread_save[self].cpos[2]=a3;
	thread_save[self].cpos[3]=a4;
	thread_save[self].cpos[4]=a5;
	thread_save[self].cstr[0]=bruteforce_charset[a1];
	thread_save[self].cstr[1]=bruteforce_charset[a2];
	thread_save[self].cstr[2]=bruteforce_charset[a3];
	thread_save[self].cstr[3]=bruteforce_charset[a4];
	thread_save[self].cstr[4]=bruteforce_charset[a5];
	thread_save[self].cstr[5]=0;
	thread_save[self].cstart=4;
	
	
	clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
	image.x=bruteforce_charset[a1]|(bruteforce_charset[a2]<<8)|(bruteforce_charset[a3]<<16)|(bruteforce_charset[a4]<<24);image.y=image.z=0;image.w=0;
	clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);
        if (found==1)
        {
    	    found = 0;
    	    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    	}
        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
        //clFlush(queue);
	invocations++;
	if (err != CL_SUCCESS)
	{
	    elog("clEnqueueNDRangeKernel error (%d)\n",err);
	    return hash_err;
	}
	//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
        int *fnd;
        fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
        found = *fnd;

        if (session_restore_lflag==1) ocl_bruteforce_init_restore(self, 4);
	session_restore_lflag = 0;
	if (found==1) 
	{
	    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*charset_size*sizeof(uint), found_ind, 0, NULL, NULL);
	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    f=-1;
	    for (a=0;a<charset_size;a++)
	    for (b=0;b<charset_size;b++)
	    for (c=0;c<charset_size;c++)
	    for (d=0;d<charset_size;d++)
	    if (found_ind[a*charset_size*charset_size*charset_size+b*charset_size*charset_size+c*charset_size+d]==1)
	    {
	    f=(a*charset_size*charset_size*charset_size+b*charset_size*charset_size+c*charset_size+d)*ocl_vector;
    	    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, f*32, 32*ocl_vector, ptr+f*32, 0, NULL, NULL);
	    for (e=0;e<ocl_vector;e++)
	    {
    		f=(a*charset_size*charset_size*charset_size+b*charset_size*charset_size+c*charset_size+d)*ocl_vector+e;
    		if ( (hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count<320) && (hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count>0))
    		{
    		    int i;
    		    for (i=0;i<hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count;i++)
    		    if (memcmp(hash_index[ptr[f*32]&255][ptr[f*32+1]&255].nodes[i]->hash, (char *)ptr+(f)*32, hash_ret_len) == 0)
    		    {
                        mylist = hash_index[ptr[f*32]&255][ptr[f*32+1]&255].nodes[i];
                        int flag = 0;
                        plain[0]=bruteforce_charset[a1];
                        plain[1]=bruteforce_charset[a2];
                        plain[2]=bruteforce_charset[a3];
                        plain[3]=bruteforce_charset[a4];
                        plain[4]=bruteforce_charset[(a5+(e))%charset_size];
                        plain[5]=bruteforce_charset[d];
                        plain[6]=bruteforce_charset[c];
                        plain[7]=bruteforce_charset[b];
                        plain[8]=bruteforce_charset[a];
                        plain[9]=0;
                        pthread_mutex_lock(&crackedmutex);
                        addlist = cracked_list;
                        while (addlist)
                        {
                            if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
                                && (strcmp(addlist->salt, mylist->salt) == 0))
                                    flag = 1;
                            addlist = addlist->next;
                        }
                        pthread_mutex_unlock(&crackedmutex);
                        if (flag == 0)
                        {
                            add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                        }
    		    }
    		}
    		else if (hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count>=320)
    		{
    		    mylist = hash_list;
    		    while (mylist)
    		    {
            		if (memcmp(mylist->hash, (char *)ptr+(f)*32, hash_ret_len) == 0)
            		{
                    	    int flag = 0;
                    	    plain[0]=bruteforce_charset[a1];
                    	    plain[1]=bruteforce_charset[a2];
                    	    plain[2]=bruteforce_charset[a3];
                    	    plain[3]=bruteforce_charset[a4];
                    	    plain[4]=bruteforce_charset[(a5+(e))%charset_size];
                    	    plain[5]=bruteforce_charset[d];
                    	    plain[6]=bruteforce_charset[c];
                    	    plain[7]=bruteforce_charset[b];
                    	    plain[8]=bruteforce_charset[a];
                    	    plain[9]=0;
                    	    pthread_mutex_lock(&crackedmutex);
                    	    addlist = cracked_list;
                    	    while (addlist)
                    	    {
                        	if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
                                && (strcmp(addlist->salt, mylist->salt) == 0))
                                    flag = 1;
                        	addlist = addlist->next;
                    	    }
                    	    pthread_mutex_unlock(&crackedmutex);
                    	    if (flag == 0)
                    	    {
                        	add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    	    }
            		}
            		if (mylist) mylist = mylist->next;
        	    }
        	}
	    }
	    }
	clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
	bzero(found_ind,charset_size*charset_size*charset_size*charset_size*sizeof(cl_uint));
	clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
	}
    }


    /* Bruteforce, len=10 */
    csize=6;
    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);
    if (session_restore_lflag==0) ocl_bruteforce_init_restore(self, 5);
    if ((bruteforce_start<=10) && (bruteforce_end>=10))
    if (thread_restore[self].cstart<=5)
    for (a1=thread_restore[self].cpos[0];a1<charset_size;a1++)
    for (a2=thread_restore[self].cpos[1];a2<charset_size;a2++)
    for (a3=thread_restore[self].cpos[2];a3<charset_size;a3++)
    for (a4=thread_restore[self].cpos[3];a4<charset_size;a4++)
    for (a5=thread_restore[self].cpos[4];a5<charset_size;a5++)
    for (a6=thread_restore[self].cpos[5];a6<charset_size;a6+=(ocl_vector*ocl_threads*devicesnum))
    {
	if (attack_over!=0) return hash_ok;
	cbase.s0 = bruteforce_charset[a6];
	cbase.s1 = bruteforce_charset[a6+1];
	cbase.s2 = bruteforce_charset[a6+2];
	cbase.s3 = bruteforce_charset[a6+3];
	cbase.s4 = bruteforce_charset[a6+4];
	cbase.s5 = bruteforce_charset[a6+5];
	cbase.s6 = bruteforce_charset[a6+6];
	cbase.s7 = bruteforce_charset[a6+7];

	thread_save[self].cpos[0]=a1;
	thread_save[self].cpos[1]=a2;
	thread_save[self].cpos[2]=a3;
	thread_save[self].cpos[3]=a4;
	thread_save[self].cpos[4]=a5;
	thread_save[self].cpos[5]=a6;
	thread_save[self].cstr[0]=bruteforce_charset[a1];
	thread_save[self].cstr[1]=bruteforce_charset[a2];
	thread_save[self].cstr[2]=bruteforce_charset[a3];
	thread_save[self].cstr[3]=bruteforce_charset[a4];
	thread_save[self].cstr[4]=bruteforce_charset[a5];
	thread_save[self].cstr[5]=bruteforce_charset[a6];
	thread_save[self].cstr[6]=0;
	thread_save[self].cstart=5;
	
	
	clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
	image.x=bruteforce_charset[a1]|(bruteforce_charset[a2]<<8)|(bruteforce_charset[a3]<<16)|(bruteforce_charset[a4]<<24);image.y=bruteforce_charset[a5];image.z=0;image.w=0;
	clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);
        if (found==1)
        {
    	    found = 0;
    	    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    	}
        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
        //clFlush(queue);
	invocations++;
	if (err != CL_SUCCESS)
	{
	    elog("clEnqueueNDRangeKernel error (%d)\n",err);
	    return hash_err;
	}
	//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
        int *fnd;
        fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
        found = *fnd;

        if (session_restore_lflag==1) ocl_bruteforce_init_restore(self, 4);
	session_restore_lflag = 0;
	if (found==1) 
	{
	    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*charset_size*sizeof(uint), found_ind, 0, NULL, NULL);
	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    f=-1;
	    for (a=0;a<charset_size;a++)
	    for (b=0;b<charset_size;b++)
	    for (c=0;c<charset_size;c++)
	    for (d=0;d<charset_size;d++)
	    if (found_ind[a*charset_size*charset_size*charset_size+b*charset_size*charset_size+c*charset_size+d]==1)
	    {
	    f=(a*charset_size*charset_size*charset_size+b*charset_size*charset_size+c*charset_size+d)*ocl_vector;
    	    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, f*32, 32*ocl_vector, ptr+f*32, 0, NULL, NULL);
	    for (e=0;e<ocl_vector;e++)
	    {
    		f=(a*charset_size*charset_size*charset_size+b*charset_size*charset_size+c*charset_size+d)*ocl_vector+e;
    		if ( (hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count<320) && (hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count>0))
    		{
    		    int i;
    		    for (i=0;i<hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count;i++)
    		    if (memcmp(hash_index[ptr[f*32]&255][ptr[f*32+1]&255].nodes[i]->hash, (char *)ptr+(f)*32, hash_ret_len) == 0)
    		    {
                        mylist = hash_index[ptr[f*32]&255][ptr[f*32+1]&255].nodes[i];
                        int flag = 0;
                        plain[0]=bruteforce_charset[a1];
                        plain[1]=bruteforce_charset[a2];
                        plain[2]=bruteforce_charset[a3];
                        plain[3]=bruteforce_charset[a4];
                        plain[4]=bruteforce_charset[a5];
                        plain[5]=bruteforce_charset[(a6+(e))%charset_size];
                        plain[6]=bruteforce_charset[d];
                        plain[7]=bruteforce_charset[c];
                        plain[8]=bruteforce_charset[b];
                        plain[9]=bruteforce_charset[a];
                        plain[10]=0;
                        pthread_mutex_lock(&crackedmutex);
                        addlist = cracked_list;
                        while (addlist)
                        {
                            if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
                                && (strcmp(addlist->salt, mylist->salt) == 0))
                                    flag = 1;
                            addlist = addlist->next;
                        }
                        pthread_mutex_unlock(&crackedmutex);
                        if (flag == 0)
                        {
                            add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                        }
    		    }
    		}
    		else if (hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count>=320)
    		{
    		    mylist = hash_list;
    		    while (mylist)
    		    {
            		if (memcmp(mylist->hash, (char *)ptr+(f)*32, hash_ret_len) == 0)
            		{
                    	    int flag = 0;
                    	    plain[0]=bruteforce_charset[a1];
                    	    plain[1]=bruteforce_charset[a2];
                    	    plain[2]=bruteforce_charset[a3];
                    	    plain[3]=bruteforce_charset[a4];
                    	    plain[4]=bruteforce_charset[a5];
                    	    plain[5]=bruteforce_charset[(a6+(e))%charset_size];
                    	    plain[6]=bruteforce_charset[d];
                    	    plain[7]=bruteforce_charset[c];
                    	    plain[8]=bruteforce_charset[b];
                    	    plain[9]=bruteforce_charset[a];
                    	    plain[10]=0;
                    	    pthread_mutex_lock(&crackedmutex);
                    	    addlist = cracked_list;
                    	    while (addlist)
                    	    {
                        	if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
                                && (strcmp(addlist->salt, mylist->salt) == 0))
                                    flag = 1;
                        	addlist = addlist->next;
                    	    }
                    	    pthread_mutex_unlock(&crackedmutex);
                    	    if (flag == 0)
                    	    {
                        	add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    	    }
            		}
            		if (mylist) mylist = mylist->next;
        	    }
        	}
	    }
	    }
	clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
	bzero(found_ind,charset_size*charset_size*charset_size*charset_size*sizeof(cl_uint));
	clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
	}
    }


    /* Bruteforce, len=11 */
    csize=7;
    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);
    if (session_restore_lflag==0) ocl_bruteforce_init_restore(self, 6);
    if ((bruteforce_start<=11) && (bruteforce_end>=11))
    if (thread_restore[self].cstart<=6)
    for (a1=thread_restore[self].cpos[0];a1<charset_size;a1++)
    for (a2=thread_restore[self].cpos[1];a2<charset_size;a2++)
    for (a3=thread_restore[self].cpos[2];a3<charset_size;a3++)
    for (a4=thread_restore[self].cpos[3];a4<charset_size;a4++)
    for (a5=thread_restore[self].cpos[4];a5<charset_size;a5++)
    for (a6=thread_restore[self].cpos[5];a6<charset_size;a6++)
    for (a7=thread_restore[self].cpos[6];a7<charset_size;a7+=(ocl_vector*ocl_threads*devicesnum))
    {
	if (attack_over!=0) return hash_ok;
	cbase.s0 = bruteforce_charset[a7];
	cbase.s1 = bruteforce_charset[a7+1];
	cbase.s2 = bruteforce_charset[a7+2];
	cbase.s3 = bruteforce_charset[a7+3];
	cbase.s4 = bruteforce_charset[a7+4];
	cbase.s5 = bruteforce_charset[a7+5];
	cbase.s6 = bruteforce_charset[a7+6];
	cbase.s7 = bruteforce_charset[a7+7];

	thread_save[self].cpos[0]=a1;
	thread_save[self].cpos[1]=a2;
	thread_save[self].cpos[2]=a3;
	thread_save[self].cpos[3]=a4;
	thread_save[self].cpos[4]=a5;
	thread_save[self].cpos[5]=a6;
	thread_save[self].cpos[6]=a7;
	thread_save[self].cstr[0]=bruteforce_charset[a1];
	thread_save[self].cstr[1]=bruteforce_charset[a2];
	thread_save[self].cstr[2]=bruteforce_charset[a3];
	thread_save[self].cstr[3]=bruteforce_charset[a4];
	thread_save[self].cstr[4]=bruteforce_charset[a5];
	thread_save[self].cstr[5]=bruteforce_charset[a6];
	thread_save[self].cstr[6]=bruteforce_charset[a7];
	thread_save[self].cstr[7]=0;
	thread_save[self].cstart=6;
	
	
	clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
	image.x=bruteforce_charset[a1]|(bruteforce_charset[a2]<<8)|(bruteforce_charset[a3]<<16)|(bruteforce_charset[a4]<<24);image.y=bruteforce_charset[a5]|(bruteforce_charset[a6]<<8);image.z=0;image.w=0;
	clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);
        if (found==1)
        {
    	    found = 0;
    	    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    	}
        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
        //clFlush(queue);
	invocations++;
	if (err != CL_SUCCESS)
	{
	    elog("clEnqueueNDRangeKernel error (%d)\n",err);
	    return hash_err;
	}
	//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
        int *fnd;
        fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
        found = *fnd;

        if (session_restore_lflag==1) ocl_bruteforce_init_restore(self, 4);
	session_restore_lflag = 0;
	if (found==1) 
	{
	    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*charset_size*sizeof(uint), found_ind, 0, NULL, NULL);
	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    f=-1;
	    for (a=0;a<charset_size;a++)
	    for (b=0;b<charset_size;b++)
	    for (c=0;c<charset_size;c++)
	    for (d=0;d<charset_size;d++)
	    if (found_ind[a*charset_size*charset_size*charset_size+b*charset_size*charset_size+c*charset_size+d]==1)
	    {
	    f=(a*charset_size*charset_size*charset_size+b*charset_size*charset_size+c*charset_size+d)*ocl_vector;
    	    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, f*32, 32*ocl_vector, ptr+f*32, 0, NULL, NULL);
	    for (e=0;e<ocl_vector;e++)
	    {
    		f=(a*charset_size*charset_size*charset_size+b*charset_size*charset_size+c*charset_size+d)*ocl_vector+e;
    		if ( (hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count<320) && (hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count>0))
    		{
    		    int i;
    		    for (i=0;i<hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count;i++)
    		    if (memcmp(hash_index[ptr[f*32]&255][ptr[f*32+1]&255].nodes[i]->hash, (char *)ptr+(f)*32, hash_ret_len) == 0)
    		    {
                        mylist = hash_index[ptr[f*32]&255][ptr[f*32+1]&255].nodes[i];
                        int flag = 0;
                        plain[0]=bruteforce_charset[a1];
                        plain[1]=bruteforce_charset[a2];
                        plain[2]=bruteforce_charset[a3];
                        plain[3]=bruteforce_charset[a4];
                        plain[4]=bruteforce_charset[a5];
                        plain[5]=bruteforce_charset[a6];
                        plain[6]=bruteforce_charset[(a7+(e))%charset_size];
                        plain[7]=bruteforce_charset[d];
                        plain[8]=bruteforce_charset[c];
                        plain[9]=bruteforce_charset[b];
                        plain[10]=bruteforce_charset[a];
                        plain[11]=0;
                        pthread_mutex_lock(&crackedmutex);
                        addlist = cracked_list;
                        while (addlist)
                        {
                            if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
                                && (strcmp(addlist->salt, mylist->salt) == 0))
                                    flag = 1;
                            addlist = addlist->next;
                        }
                        pthread_mutex_unlock(&crackedmutex);
                        if (flag == 0)
                        {
                            add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                        }
    		    }
    		}
    		else if (hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count>=320)
    		{
    		    mylist = hash_list;
    		    while (mylist)
    		    {
            		if (memcmp(mylist->hash, (char *)ptr+(f)*32, hash_ret_len) == 0)
            		{
                    	    int flag = 0;
                    	    plain[0]=bruteforce_charset[a1];
                    	    plain[1]=bruteforce_charset[a2];
                    	    plain[2]=bruteforce_charset[a3];
                    	    plain[3]=bruteforce_charset[a4];
                    	    plain[4]=bruteforce_charset[a5];
                    	    plain[5]=bruteforce_charset[a6];
                    	    plain[6]=bruteforce_charset[(a7+(e))%charset_size];
                    	    plain[7]=bruteforce_charset[d];
                    	    plain[8]=bruteforce_charset[c];
                    	    plain[9]=bruteforce_charset[b];
                    	    plain[10]=bruteforce_charset[a];
                    	    plain[11]=0;
                    	    pthread_mutex_lock(&crackedmutex);
                    	    addlist = cracked_list;
                    	    while (addlist)
                    	    {
                        	if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
                                && (strcmp(addlist->salt, mylist->salt) == 0))
                                    flag = 1;
                        	addlist = addlist->next;
                    	    }
                    	    pthread_mutex_unlock(&crackedmutex);
                    	    if (flag == 0)
                    	    {
                        	add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    	    }
            		}
            		if (mylist) mylist = mylist->next;
        	    }
        	}
	    }
	    }
	clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
	bzero(found_ind,charset_size*charset_size*charset_size*charset_size*sizeof(cl_uint));
	clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
	}
    }


    /* Bruteforce, len=12 */
    csize=8;
    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);
    if (session_restore_lflag==0) ocl_bruteforce_init_restore(self, 7);
    if ((bruteforce_start<=12) && (bruteforce_end>=12))
    if (thread_restore[self].cstart<=7)
    for (a1=thread_restore[self].cpos[0];a1<charset_size;a1++)
    for (a2=thread_restore[self].cpos[1];a2<charset_size;a2++)
    for (a3=thread_restore[self].cpos[2];a3<charset_size;a3++)
    for (a4=thread_restore[self].cpos[3];a4<charset_size;a4++)
    for (a5=thread_restore[self].cpos[4];a5<charset_size;a5++)
    for (a6=thread_restore[self].cpos[5];a6<charset_size;a6++)
    for (a7=thread_restore[self].cpos[6];a7<charset_size;a7++)
    for (a8=thread_restore[self].cpos[7];a8<charset_size;a8+=(ocl_vector*ocl_threads*devicesnum))
    {
	if (attack_over!=0) return hash_ok;
	cbase.s0 = bruteforce_charset[a8];
	cbase.s1 = bruteforce_charset[a8+1];
	cbase.s2 = bruteforce_charset[a8+2];
	cbase.s3 = bruteforce_charset[a8+3];
	cbase.s4 = bruteforce_charset[a8+4];
	cbase.s5 = bruteforce_charset[a8+5];
	cbase.s6 = bruteforce_charset[a8+6];
	cbase.s7 = bruteforce_charset[a8+7];

	thread_save[self].cpos[0]=a1;
	thread_save[self].cpos[1]=a2;
	thread_save[self].cpos[2]=a3;
	thread_save[self].cpos[3]=a4;
	thread_save[self].cpos[4]=a5;
	thread_save[self].cpos[5]=a6;
	thread_save[self].cpos[6]=a7;
	thread_save[self].cpos[7]=a8;
	thread_save[self].cstr[0]=bruteforce_charset[a1];
	thread_save[self].cstr[1]=bruteforce_charset[a2];
	thread_save[self].cstr[2]=bruteforce_charset[a3];
	thread_save[self].cstr[3]=bruteforce_charset[a4];
	thread_save[self].cstr[4]=bruteforce_charset[a5];
	thread_save[self].cstr[5]=bruteforce_charset[a6];
	thread_save[self].cstr[6]=bruteforce_charset[a7];
	thread_save[self].cstr[7]=bruteforce_charset[a8];
	thread_save[self].cstr[8]=0;
	thread_save[self].cstart=7;
	
	clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
	image.x=bruteforce_charset[a1]|(bruteforce_charset[a2]<<8)|(bruteforce_charset[a3]<<16)|(bruteforce_charset[a4]<<24);image.y=bruteforce_charset[a5]|(bruteforce_charset[a6]<<8)|(bruteforce_charset[a7]<<16);image.z=0;image.w=0;
	clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);
        if (found==1)
        {
    	    found = 0;
    	    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    	}
        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
        //clFlush(queue);
	invocations++;
	if (err != CL_SUCCESS)
	{
	    elog("clEnqueueNDRangeKernel error (%d)\n",err);
	    return hash_err;
	}
	//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
        int *fnd;
        fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
        found = *fnd;

        if (session_restore_lflag==1) ocl_bruteforce_init_restore(self, 4);
	session_restore_lflag = 0;
	if (found==1) 
	{
	    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*charset_size*sizeof(uint), found_ind, 0, NULL, NULL);
	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    f=-1;
	    for (a=0;a<charset_size;a++)
	    for (b=0;b<charset_size;b++)
	    for (c=0;c<charset_size;c++)
	    for (d=0;d<charset_size;d++)
	    if (found_ind[a*charset_size*charset_size*charset_size+b*charset_size*charset_size+c*charset_size+d]==1)
	    {
	    f=(a*charset_size*charset_size*charset_size+b*charset_size*charset_size+c*charset_size+d)*ocl_vector;
    	    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, f*32, 32*ocl_vector, ptr+f*32, 0, NULL, NULL);
	    for (e=0;e<ocl_vector;e++)
	    {
    		f=(a*charset_size*charset_size*charset_size+b*charset_size*charset_size+c*charset_size+d)*ocl_vector+e;
    		if ( (hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count<320) && (hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count>0))
    		{
    		    int i;
    		    for (i=0;i<hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count;i++)
    		    if (memcmp(hash_index[ptr[f*32]&255][ptr[f*32+1]&255].nodes[i]->hash, (char *)ptr+(f)*32, hash_ret_len) == 0)
    		    {
                        mylist = hash_index[ptr[f*32]&255][ptr[f*32+1]&255].nodes[i];
                        int flag = 0;
                        plain[0]=bruteforce_charset[a1];
                        plain[1]=bruteforce_charset[a2];
                        plain[2]=bruteforce_charset[a3];
                        plain[3]=bruteforce_charset[a4];
                        plain[4]=bruteforce_charset[a5];
                        plain[5]=bruteforce_charset[a6];
                        plain[6]=bruteforce_charset[a7];
                        plain[7]=bruteforce_charset[(a8+(e))%charset_size];
                        plain[8]=bruteforce_charset[d];
                        plain[9]=bruteforce_charset[c];
                        plain[10]=bruteforce_charset[b];
                        plain[11]=bruteforce_charset[a];
                        plain[12]=0;
                        pthread_mutex_lock(&crackedmutex);
                        addlist = cracked_list;
                        while (addlist)
                        {
                            if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
                                && (strcmp(addlist->salt, mylist->salt) == 0))
                                    flag = 1;
                            addlist = addlist->next;
                        }
                        pthread_mutex_unlock(&crackedmutex);
                        if (flag == 0)
                        {
                            add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                        }
    		    }
    		}
    		else if (hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count>=320)
    		{
    		    mylist = hash_list;
    		    while (mylist)
    		    {
            		if (memcmp(mylist->hash, (char *)ptr+(f)*32, hash_ret_len) == 0)
            		{
                    	    int flag = 0;
                    	    plain[0]=bruteforce_charset[a1];
                    	    plain[1]=bruteforce_charset[a2];
                    	    plain[2]=bruteforce_charset[a3];
                    	    plain[3]=bruteforce_charset[a4];
                    	    plain[4]=bruteforce_charset[a5];
                    	    plain[5]=bruteforce_charset[a6];
                    	    plain[6]=bruteforce_charset[a7];
                    	    plain[7]=bruteforce_charset[(a8+(e))%charset_size];
                    	    plain[8]=bruteforce_charset[d];
                    	    plain[9]=bruteforce_charset[c];
                    	    plain[10]=bruteforce_charset[b];
                    	    plain[11]=bruteforce_charset[a];
                    	    plain[12]=0;
                    	    pthread_mutex_lock(&crackedmutex);
                    	    addlist = cracked_list;
                    	    while (addlist)
                    	    {
                        	if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
                                && (strcmp(addlist->salt, mylist->salt) == 0))
                                    flag = 1;
                        	addlist = addlist->next;
                    	    }
                    	    pthread_mutex_unlock(&crackedmutex);
                    	    if (flag == 0)
                    	    {
                        	add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    	    }
            		}
            		if (mylist) mylist = mylist->next;
        	    }
        	}
	    }
	    }
	clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
	bzero(found_ind,charset_size*charset_size*charset_size*charset_size*sizeof(cl_uint));
	clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
	}
    }


    /* Bruteforce, len=13 */
    csize=9;
    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);
    if (session_restore_lflag==0) ocl_bruteforce_init_restore(self, 8);
    if ((bruteforce_start<=13) && (bruteforce_end>=13))
    if (thread_restore[self].cstart<=8)
    for (a1=thread_restore[self].cpos[0];a1<charset_size;a1++)
    for (a2=thread_restore[self].cpos[1];a2<charset_size;a2++)
    for (a3=thread_restore[self].cpos[2];a3<charset_size;a3++)
    for (a4=thread_restore[self].cpos[3];a4<charset_size;a4++)
    for (a5=thread_restore[self].cpos[4];a5<charset_size;a5++)
    for (a6=thread_restore[self].cpos[5];a6<charset_size;a6++)
    for (a7=thread_restore[self].cpos[6];a7<charset_size;a7++)
    for (a8=thread_restore[self].cpos[7];a8<charset_size;a8++)
    for (a9=thread_restore[self].cpos[8];a9<charset_size;a9+=(ocl_vector*ocl_threads*devicesnum))
    {
	if (attack_over!=0) return hash_ok;
	cbase.s0 = bruteforce_charset[a9];
	cbase.s1 = bruteforce_charset[a9+1];
	cbase.s2 = bruteforce_charset[a9+2];
	cbase.s3 = bruteforce_charset[a9+3];
	cbase.s4 = bruteforce_charset[a9+4];
	cbase.s5 = bruteforce_charset[a9+5];
	cbase.s6 = bruteforce_charset[a9+6];
	cbase.s7 = bruteforce_charset[a9+7];
	thread_save[self].cpos[0]=a1;
	thread_save[self].cpos[1]=a2;
	thread_save[self].cpos[2]=a3;
	thread_save[self].cpos[3]=a4;
	thread_save[self].cpos[4]=a5;
	thread_save[self].cpos[5]=a6;
	thread_save[self].cpos[6]=a7;
	thread_save[self].cpos[7]=a8;
	thread_save[self].cpos[8]=a9;
	thread_save[self].cstr[0]=bruteforce_charset[a1];
	thread_save[self].cstr[1]=bruteforce_charset[a2];
	thread_save[self].cstr[2]=bruteforce_charset[a3];
	thread_save[self].cstr[3]=bruteforce_charset[a4];
	thread_save[self].cstr[4]=bruteforce_charset[a5];
	thread_save[self].cstr[5]=bruteforce_charset[a6];
	thread_save[self].cstr[6]=bruteforce_charset[a7];
	thread_save[self].cstr[7]=bruteforce_charset[a8];
	thread_save[self].cstr[8]=bruteforce_charset[a9];
	thread_save[self].cstr[9]=0;
	thread_save[self].cstart=8;
	
	clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
	image.x=bruteforce_charset[a1]|(bruteforce_charset[a2]<<8)|(bruteforce_charset[a3]<<16)|(bruteforce_charset[a4]<<24);image.y=bruteforce_charset[a5]|(bruteforce_charset[a6]<<8)|(bruteforce_charset[a7]<<16)|(bruteforce_charset[a8]<<24);image.z=image.w=0;
	clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);
        if (found==1)
        {
    	    found = 0;
    	    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    	}
        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
        //clFlush(queue);
	invocations++;
	if (err != CL_SUCCESS)
	{
	    elog("clEnqueueNDRangeKernel error (%d)\n",err);
	    return hash_err;
	}
	//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
        int *fnd;
        fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
        found = *fnd;

        if (session_restore_lflag==1) ocl_bruteforce_init_restore(self, 4);
	session_restore_lflag = 0;
	if (found==1) 
	{
	    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*charset_size*sizeof(uint), found_ind, 0, NULL, NULL);
	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    f=-1;
	    for (a=0;a<charset_size;a++)
	    for (b=0;b<charset_size;b++)
	    for (c=0;c<charset_size;c++)
	    for (d=0;d<charset_size;d++)
	    if (found_ind[a*charset_size*charset_size*charset_size+b*charset_size*charset_size+c*charset_size+d]==1)
	    {
	    f=(a*charset_size*charset_size*charset_size+b*charset_size*charset_size+c*charset_size+d)*ocl_vector;
    	    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, f*32, 32*ocl_vector, ptr+f*32, 0, NULL, NULL);
	    for (e=0;e<ocl_vector;e++)
	    {
    		f=(a*charset_size*charset_size*charset_size+b*charset_size*charset_size+c*charset_size+d)*ocl_vector+e;
    		if ( (hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count<320) && (hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count>0))
    		{
    		    int i;
    		    for (i=0;i<hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count;i++)
    		    if (memcmp(hash_index[ptr[f*32]&255][ptr[f*32+1]&255].nodes[i]->hash, (char *)ptr+(f)*32, hash_ret_len) == 0)
    		    {
                        mylist = hash_index[ptr[f*32]&255][ptr[f*32+1]&255].nodes[i];
                        int flag = 0;
                        plain[0]=bruteforce_charset[a1];
                        plain[1]=bruteforce_charset[a2];
                        plain[2]=bruteforce_charset[a3];
                        plain[3]=bruteforce_charset[a4];
                        plain[4]=bruteforce_charset[a5];
                        plain[5]=bruteforce_charset[a6];
                        plain[6]=bruteforce_charset[a7];
                        plain[7]=bruteforce_charset[a8];
                        plain[8]=bruteforce_charset[(a9+(e))%charset_size];
                        plain[9]=bruteforce_charset[d];
                        plain[10]=bruteforce_charset[c];
                        plain[11]=bruteforce_charset[b];
                        plain[12]=bruteforce_charset[a];
                        plain[13]=0;
                        pthread_mutex_lock(&crackedmutex);
                        addlist = cracked_list;
                        while (addlist)
                        {
                            if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
                                && (strcmp(addlist->salt, mylist->salt) == 0))
                                    flag = 1;
                            addlist = addlist->next;
                        }
                        pthread_mutex_unlock(&crackedmutex);
                        if (flag == 0)
                        {
                            add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                        }
    		    }
    		}
    		else if (hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count>=320)
    		{
    		    mylist = hash_list;
    		    while (mylist)
    		    {
            		if (memcmp(mylist->hash, (char *)ptr+(f)*32, hash_ret_len) == 0)
            		{
                    	    int flag = 0;
                    	    plain[0]=bruteforce_charset[a1];
                    	    plain[1]=bruteforce_charset[a2];
                    	    plain[2]=bruteforce_charset[a3];
                    	    plain[3]=bruteforce_charset[a4];
                    	    plain[4]=bruteforce_charset[a5];
                    	    plain[5]=bruteforce_charset[a6];
                    	    plain[6]=bruteforce_charset[a7];
                    	    plain[7]=bruteforce_charset[a8];
                    	    plain[8]=bruteforce_charset[(a9+(e))%charset_size];
                    	    plain[9]=bruteforce_charset[d];
                    	    plain[10]=bruteforce_charset[c];
                    	    plain[11]=bruteforce_charset[b];
                    	    plain[12]=bruteforce_charset[a];
                    	    plain[13]=0;
                    	    pthread_mutex_lock(&crackedmutex);
                    	    addlist = cracked_list;
                    	    while (addlist)
                    	    {
                        	if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
                                && (strcmp(addlist->salt, mylist->salt) == 0))
                                    flag = 1;
                        	addlist = addlist->next;
                    	    }
                    	    pthread_mutex_unlock(&crackedmutex);
                    	    if (flag == 0)
                    	    {
                        	add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    	    }
            		}
            		if (mylist) mylist = mylist->next;
        	    }
        	}
	    }
	    }
	clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
	bzero(found_ind,charset_size*charset_size*charset_size*charset_size*sizeof(cl_uint));
	clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
	}
    }



    /* Bruteforce, len=14 */
    csize=10;
    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);
    if (session_restore_lflag==0) ocl_bruteforce_init_restore(self, 9);
    if ((bruteforce_start<=14) && (bruteforce_end>=14))
    if (thread_restore[self].cstart<=9)
    for (a1=thread_restore[self].cpos[0];a1<charset_size;a1++)
    for (a2=thread_restore[self].cpos[1];a2<charset_size;a2++)
    for (a3=thread_restore[self].cpos[2];a3<charset_size;a3++)
    for (a4=thread_restore[self].cpos[3];a4<charset_size;a4++)
    for (a5=thread_restore[self].cpos[4];a5<charset_size;a5++)
    for (a6=thread_restore[self].cpos[5];a6<charset_size;a6++)
    for (a7=thread_restore[self].cpos[6];a7<charset_size;a7++)
    for (a8=thread_restore[self].cpos[7];a8<charset_size;a8++)
    for (a9=thread_restore[self].cpos[8];a9<charset_size;a9++)
    for (a10=thread_restore[self].cpos[9];a10<charset_size;a10+=(ocl_vector*ocl_threads*devicesnum))
    {
	if (attack_over!=0) return hash_ok;
	cbase.s0 = bruteforce_charset[a9];
	cbase.s1 = bruteforce_charset[a9+1];
	cbase.s2 = bruteforce_charset[a9+2];
	cbase.s3 = bruteforce_charset[a9+3];
	cbase.s4 = bruteforce_charset[a9+4];
	cbase.s5 = bruteforce_charset[a9+5];
	cbase.s6 = bruteforce_charset[a9+6];
	cbase.s7 = bruteforce_charset[a9+7];
	thread_save[self].cpos[0]=a1;
	thread_save[self].cpos[1]=a2;
	thread_save[self].cpos[2]=a3;
	thread_save[self].cpos[3]=a4;
	thread_save[self].cpos[4]=a5;
	thread_save[self].cpos[5]=a6;
	thread_save[self].cpos[6]=a7;
	thread_save[self].cpos[7]=a8;
	thread_save[self].cpos[8]=a9;
	thread_save[self].cpos[9]=a10;
	thread_save[self].cstr[0]=bruteforce_charset[a1];
	thread_save[self].cstr[1]=bruteforce_charset[a2];
	thread_save[self].cstr[2]=bruteforce_charset[a3];
	thread_save[self].cstr[3]=bruteforce_charset[a4];
	thread_save[self].cstr[4]=bruteforce_charset[a5];
	thread_save[self].cstr[5]=bruteforce_charset[a6];
	thread_save[self].cstr[6]=bruteforce_charset[a7];
	thread_save[self].cstr[7]=bruteforce_charset[a8];
	thread_save[self].cstr[8]=bruteforce_charset[a9];
	thread_save[self].cstr[9]=bruteforce_charset[a10];
	thread_save[self].cstr[10]=0;
	thread_save[self].cstart=9;
	
	clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
	image.x=bruteforce_charset[a1]|(bruteforce_charset[a2]<<8)|(bruteforce_charset[a3]<<16)|(bruteforce_charset[a4]<<24);image.y=bruteforce_charset[a5]|(bruteforce_charset[a6]<<8)|(bruteforce_charset[a7]<<16)|(bruteforce_charset[a8]<<24);image.z=bruteforce_charset[a9];image.w=0;
	clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);
        if (found==1)
        {
    	    found = 0;
    	    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    	}
        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
        //clFlush(queue);
	invocations++;
	if (err != CL_SUCCESS)
	{
	    elog("clEnqueueNDRangeKernel error (%d)\n",err);
	    return hash_err;
	}
	//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
        int *fnd;
        fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
        found = *fnd;

        if (session_restore_lflag==1) ocl_bruteforce_init_restore(self, 4);
	session_restore_lflag = 0;
	if (found==1) 
	{
	    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*charset_size*sizeof(uint), found_ind, 0, NULL, NULL);
	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    f=-1;
	    for (a=0;a<charset_size;a++)
	    for (b=0;b<charset_size;b++)
	    for (c=0;c<charset_size;c++)
	    for (d=0;d<charset_size;d++)
	    if (found_ind[a*charset_size*charset_size*charset_size+b*charset_size*charset_size+c*charset_size+d]==1)
	    {
	    f=(a*charset_size*charset_size*charset_size+b*charset_size*charset_size+c*charset_size+d)*ocl_vector;
    	    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, f*32, 32*ocl_vector, ptr+f*32, 0, NULL, NULL);
	    for (e=0;e<ocl_vector;e++)
	    {
    		f=(a*charset_size*charset_size*charset_size+b*charset_size*charset_size+c*charset_size+d)*ocl_vector+e;
    		if ( (hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count<320) && (hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count>0))
    		{
    		    int i;
    		    for (i=0;i<hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count;i++)
    		    if (memcmp(hash_index[ptr[f*32]&255][ptr[f*32+1]&255].nodes[i]->hash, (char *)ptr+(f)*32, hash_ret_len) == 0)
    		    {
                        mylist = hash_index[ptr[f*32]&255][ptr[f*32+1]&255].nodes[i];
                        int flag = 0;
                        plain[0]=bruteforce_charset[a1];
                        plain[1]=bruteforce_charset[a2];
                        plain[2]=bruteforce_charset[a3];
                        plain[3]=bruteforce_charset[a4];
                        plain[4]=bruteforce_charset[a5];
                        plain[5]=bruteforce_charset[a6];
                        plain[6]=bruteforce_charset[a7];
                        plain[7]=bruteforce_charset[a8];
                        plain[8]=bruteforce_charset[a9];
                        plain[9]=bruteforce_charset[(a10+(e))%charset_size];
                        plain[10]=bruteforce_charset[d];
                        plain[11]=bruteforce_charset[c];
                        plain[12]=bruteforce_charset[b];
                        plain[13]=bruteforce_charset[a];
                        plain[14]=0;
                        pthread_mutex_lock(&crackedmutex);
                        addlist = cracked_list;
                        while (addlist)
                        {
                            if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
                                && (strcmp(addlist->salt, mylist->salt) == 0))
                                    flag = 1;
                            addlist = addlist->next;
                        }
                        pthread_mutex_unlock(&crackedmutex);
                        if (flag == 0)
                        {
                            add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                        }
    		    }
    		}
    		else if (hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count>=320)
    		{
    		    mylist = hash_list;
    		    while (mylist)
    		    {
            		if (memcmp(mylist->hash, (char *)ptr+(f)*32, hash_ret_len) == 0)
            		{
                    	    int flag = 0;
                    	    plain[0]=bruteforce_charset[a1];
                    	    plain[1]=bruteforce_charset[a2];
                    	    plain[2]=bruteforce_charset[a3];
                    	    plain[3]=bruteforce_charset[a4];
                    	    plain[4]=bruteforce_charset[a5];
                    	    plain[5]=bruteforce_charset[a6];
                    	    plain[6]=bruteforce_charset[a7];
                    	    plain[7]=bruteforce_charset[a8];
                    	    plain[8]=bruteforce_charset[a9];
                    	    plain[9]=bruteforce_charset[(a10+(e))%charset_size];
                    	    plain[10]=bruteforce_charset[d];
                    	    plain[11]=bruteforce_charset[c];
                    	    plain[12]=bruteforce_charset[b];
                    	    plain[13]=bruteforce_charset[a];
                    	    plain[14]=0;
                    	    pthread_mutex_lock(&crackedmutex);
                    	    addlist = cracked_list;
                    	    while (addlist)
                    	    {
                        	if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
                                && (strcmp(addlist->salt, mylist->salt) == 0))
                                    flag = 1;
                        	addlist = addlist->next;
                    	    }
                    	    pthread_mutex_unlock(&crackedmutex);
                    	    if (flag == 0)
                    	    {
                        	add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    	    }
            		}
            		if (mylist) mylist = mylist->next;
        	    }
        	}
	    }
	    }
	clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
	bzero(found_ind,charset_size*charset_size*charset_size*charset_size*sizeof(cl_uint));
	clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
	}
    }



    /* Bruteforce, len=15 */
    csize=11;
    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);

    if ((bruteforce_start<=13) && (bruteforce_end>=13))
    if (thread_restore[self].cstart<=10)
    for (a1=thread_restore[self].cpos[0];a1<charset_size;a1++)
    for (a2=thread_restore[self].cpos[1];a2<charset_size;a2++)
    for (a3=thread_restore[self].cpos[2];a3<charset_size;a3++)
    for (a4=thread_restore[self].cpos[3];a4<charset_size;a4++)
    for (a5=thread_restore[self].cpos[4];a5<charset_size;a5++)
    for (a6=thread_restore[self].cpos[5];a6<charset_size;a6++)
    for (a7=thread_restore[self].cpos[6];a7<charset_size;a7++)
    for (a8=thread_restore[self].cpos[7];a8<charset_size;a8++)
    for (a9=thread_restore[self].cpos[8];a9<charset_size;a9++)
    for (a10=thread_restore[self].cpos[9];a10<charset_size;a10++)
    for (a11=thread_restore[self].cpos[11];a11<charset_size;a11+=(ocl_vector*ocl_threads*devicesnum))
    {
	if (attack_over!=0) return hash_ok;
	cbase.s0 = bruteforce_charset[a10];
	cbase.s1 = bruteforce_charset[a10+1];
	cbase.s2 = bruteforce_charset[a10+2];
	cbase.s3 = bruteforce_charset[a10+3];
	cbase.s4 = bruteforce_charset[a10+4];
	cbase.s5 = bruteforce_charset[a10+5];
	cbase.s6 = bruteforce_charset[a10+6];
	cbase.s7 = bruteforce_charset[a10+7];
	thread_save[self].cpos[0]=a1;
	thread_save[self].cpos[1]=a2;
	thread_save[self].cpos[2]=a3;
	thread_save[self].cpos[3]=a4;
	thread_save[self].cpos[4]=a5;
	thread_save[self].cpos[5]=a6;
	thread_save[self].cpos[6]=a7;
	thread_save[self].cpos[7]=a8;
	thread_save[self].cpos[8]=a9;
	thread_save[self].cpos[9]=a10;
	thread_save[self].cpos[10]=a11;
	thread_save[self].cstr[0]=bruteforce_charset[a1];
	thread_save[self].cstr[1]=bruteforce_charset[a2];
	thread_save[self].cstr[2]=bruteforce_charset[a3];
	thread_save[self].cstr[3]=bruteforce_charset[a4];
	thread_save[self].cstr[4]=bruteforce_charset[a5];
	thread_save[self].cstr[5]=bruteforce_charset[a6];
	thread_save[self].cstr[6]=bruteforce_charset[a7];
	thread_save[self].cstr[7]=bruteforce_charset[a8];
	thread_save[self].cstr[8]=bruteforce_charset[a9];
	thread_save[self].cstr[9]=bruteforce_charset[a10];
	thread_save[self].cstr[10]=bruteforce_charset[a11];
	thread_save[self].cstr[11]=0;
	thread_save[self].cstart=10;
	
	clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
	image.x=bruteforce_charset[a1]|(bruteforce_charset[a2]<<8)|(bruteforce_charset[a3]<<16)|(bruteforce_charset[a4]<<24);image.y=bruteforce_charset[a5]|(bruteforce_charset[a6]<<8)|(bruteforce_charset[a7]<<16)|(bruteforce_charset[a8]<<24);image.z=bruteforce_charset[a9]|(bruteforce_charset[a10]<<8)|(bruteforce_charset[a10]<<8);image.w=0;
	clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);
        if (found==1)
        {
    	    found = 0;
    	    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    	}
        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
        //clFlush(queue);
	invocations++;
	if (err != CL_SUCCESS)
	{
	    elog("clEnqueueNDRangeKernel error (%d)\n",err);
	    return hash_err;
	}
	//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
        int *fnd;
        fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
        found = *fnd;

        if (session_restore_lflag==1) ocl_bruteforce_init_restore(self, 4);
	session_restore_lflag = 0;
	if (found==1) 
	{
	    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*charset_size*sizeof(uint), found_ind, 0, NULL, NULL);
	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    f=-1;
	    for (a=0;a<charset_size;a++)
	    for (b=0;b<charset_size;b++)
	    for (c=0;c<charset_size;c++)
	    for (d=0;d<charset_size;d++)
	    if (found_ind[a*charset_size*charset_size*charset_size+b*charset_size*charset_size+c*charset_size+d]==1)
	    {
	    f=(a*charset_size*charset_size*charset_size+b*charset_size*charset_size+c*charset_size+d)*ocl_vector;
    	    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, f*32, 32*ocl_vector, ptr+f*32, 0, NULL, NULL);
	    for (e=0;e<ocl_vector;e++)
	    {
    		f=(a*charset_size*charset_size*charset_size+b*charset_size*charset_size+c*charset_size+d)*ocl_vector+e;
    		if ( (hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count<320) && (hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count>0))
    		{
    		    int i;
    		    for (i=0;i<hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count;i++)
    		    if (memcmp(hash_index[ptr[f*32]&255][ptr[f*32+1]&255].nodes[i]->hash, (char *)ptr+(f)*32, hash_ret_len) == 0)
    		    {
                        mylist = hash_index[ptr[f*32]&255][ptr[f*32+1]&255].nodes[i];
                        int flag = 0;
                        plain[0]=bruteforce_charset[a1];
                        plain[1]=bruteforce_charset[a2];
                        plain[2]=bruteforce_charset[a3];
                        plain[3]=bruteforce_charset[a4];
                        plain[4]=bruteforce_charset[a5];
                        plain[5]=bruteforce_charset[a6];
                        plain[6]=bruteforce_charset[a7];
                        plain[7]=bruteforce_charset[a8];
                        plain[8]=bruteforce_charset[a9];
                        plain[9]=bruteforce_charset[a10];
                        plain[10]=bruteforce_charset[(a11+(e))%charset_size];
                        plain[11]=bruteforce_charset[d];
                        plain[12]=bruteforce_charset[c];
                        plain[13]=bruteforce_charset[b];
                        plain[14]=bruteforce_charset[a];
                        plain[15]=0;
                        pthread_mutex_lock(&crackedmutex);
                        addlist = cracked_list;
                        while (addlist)
                        {
                            if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
                                && (strcmp(addlist->salt, mylist->salt) == 0))
                                    flag = 1;
                            addlist = addlist->next;
                        }
                        pthread_mutex_unlock(&crackedmutex);
                        if (flag == 0)
                        {
                            add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                        }
    		    }
    		}
    		else if (hash_index[ptr[f*32]&255][ptr[f*32+1]&255].count>=320)
    		{
    		    mylist = hash_list;
    		    while (mylist)
    		    {
            		if (memcmp(mylist->hash, (char *)ptr+(f)*32, hash_ret_len) == 0)
    			{
                    	    int flag = 0;
                    	    plain[0]=bruteforce_charset[a1];
                    	    plain[1]=bruteforce_charset[a2];
                    	    plain[2]=bruteforce_charset[a3];
                    	    plain[3]=bruteforce_charset[a4];
                    	    plain[4]=bruteforce_charset[a5];
                    	    plain[5]=bruteforce_charset[a6];
                    	    plain[6]=bruteforce_charset[a7];
                    	    plain[7]=bruteforce_charset[a8];
                    	    plain[8]=bruteforce_charset[a9];
                    	    plain[9]=bruteforce_charset[a10];
                    	    plain[10]=bruteforce_charset[(a11+(e))%charset_size];
                    	    plain[11]=bruteforce_charset[d];
                    	    plain[12]=bruteforce_charset[c];
                    	    plain[13]=bruteforce_charset[b];
                    	    plain[14]=bruteforce_charset[a];
                    	    plain[15]=0;
                    	    pthread_mutex_lock(&crackedmutex);
                    	    addlist = cracked_list;
                    	    while (addlist)
                    	    {
                        	if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
                                && (strcmp(addlist->salt, mylist->salt) == 0))
                                    flag = 1;
                        	addlist = addlist->next;
                    	    }
                    	    pthread_mutex_unlock(&crackedmutex);
                    	    if (flag == 0)
                    	    {
                        	add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    	    }
            		}
            		if (mylist) mylist = mylist->next;
        	    }
        	}
	    }
	    }
	clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
	bzero(found_ind,charset_size*charset_size*charset_size*charset_size*sizeof(cl_uint));
	clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
	}
    }



    return hash_ok;

}






hash_stat ocl_markov_sha512_thread(void *arg)
{
    int err;
    int markov2[88][88];
    cl_command_queue queue;
    cl_mem buffer;
    size_t global_work_size;
    cl_uint4 image;
    int self;
    cl_kernel kernel;
    unsigned char hex1[16];
    char plain[64];
    cl_mem bitmaps_buf;
    cl_uint *bitmaps;
    int a,b,c,d,e;
    int a1,a2,a3,a4,a5,a6,a7,a8,a9,a10;
    char *ptr;
    int charset_size = (int)strlen(markov_charset);
    cl_mem found_ind_buf;
    cl_uint *found_ind;
    int found;
    cl_mem found_buf;
    struct  hash_list_s  *mylist, *addlist;
    unsigned int csize=1;
    cl_mem table_buf;
    unsigned int *table;
    cl_uint8 cbase;
    const int mymarkov_threshold = markov_threshold;
    char reduced_charset[88];
    int reduced_size;
    cl_uint4 singlehash;
    int session_restore_lflag;
    unsigned int zero=0;
    // byteorder reversal temps
    unsigned int tmp1,tmp2,tmp,l; 

    pthread_mutex_lock(&biglock);
    memcpy(&self,arg,sizeof(int));
    session_restore_lflag = session_restore_flag;
    bitmaps = malloc(256*256*32*4);
    kernel = clCreateKernel(program[self], "sha512_markov", &err );
    if (err != CL_SUCCESS)
    {
	elog("clCreateKernel error (%d)\n",err);
	return hash_err;
    }

    queue = clCreateCommandQueue(context[self], device[self/ocl_threads], 0, &err );
    if (err != CL_SUCCESS)
    {
        elog("clCreateCommandQueue error (%d)\n",err);
	return hash_err;
    }


    if (fast_markov == 1)
    {
	charset_size = charset_size - 23;
	if ((self==0)&&(session_restore_lflag==0)) markov_threshold = (markov_threshold*3)/2;
    }


    found_buf = clCreateBuffer(context[self], CL_MEM_WRITE_ONLY, 4, NULL, &err );
    if (err != CL_SUCCESS)
    {
	elog("clCreateBuffer error (%d)\n",err);
	return hash_err;
    }


    reduced_size=0;
    for (a=0;a<charset_size;a++) if (markov0[a]>markov_threshold)
    {
	reduced_charset[reduced_size]=markov_charset[a];
	// Create markov2 table
	for (b=0;b<strlen(markov_charset);b++) markov2[reduced_size][b] = markov1[a][b];
	reduced_size++;
	reduced_charset[reduced_size]=0;
    }
    

    ptr  = malloc(charset_size*charset_size*charset_size*ocl_vector*32);
    table = malloc(charset_size*charset_size*charset_size*4);
    table_buf = clCreateBuffer(context[self], CL_MEM_READ_ONLY, charset_size*charset_size*charset_size*4, NULL, &err );
    if (err != CL_SUCCESS)
    {
	elog("clCreateBuffer error (%d)\n",err);
	return hash_err;
    }


    for (a=0;a<charset_size*charset_size*charset_size;a++)
    {
	unsigned char x,y,z,t1;
	t1 = (a / (charset_size*charset_size));
	x = (markov_charset[(a / (charset_size*charset_size))]);
	y = (markov_charset[((a-t1*(charset_size*charset_size))/charset_size)]);
	z = (markov_charset[(a % charset_size)]);
	table[a] = (x<<16)|(y<<8)|(z);
    }


    buffer = clCreateBuffer(context[self], CL_MEM_WRITE_ONLY, charset_size*charset_size*charset_size*ocl_vector*32, NULL, &err );
    if (err != CL_SUCCESS)
    {
	elog("clCreateBuffer error (%d)\n",err);
	return hash_err;
    }

    for (a=0;a<256*256*32;a++)
    {
	    bitmaps[a]=0;
    }

    mylist = hash_list;
    a=0;
    while (mylist)
    {
        memcpy(hex1,mylist->hash,16);
	a++;
	unsigned int *b_a = (unsigned int *)hex1;
	unsigned int *b_b = (unsigned int *)&hex1[4];
	unsigned int *b_c = (unsigned int *)&hex1[8];
	unsigned int *b_d = (unsigned int *)&hex1[12];
	unsigned int bind_a = (*b_a)>>13;
	unsigned int bval_a = (1<<((*b_a)&31));
	unsigned int bind_b = (*b_b)>>13;
	unsigned int bval_b = (1<<((*b_b)&31));
	unsigned int bind_c = (*b_c)>>13;
	unsigned int bval_c = (1<<((*b_c)&31));
	unsigned int bind_d = (*b_d)>>13;
	unsigned int bval_d = (1<<((*b_d)&31));
	bitmaps[bind_a] |=bval_a;
	bitmaps[bind_b+65535*8] |=bval_b;
	bitmaps[bind_c+65535*16] |=bval_c;
	bitmaps[bind_d+65535*24] |=bval_d;
	singlehash.x |= (1<<((*b_b)&31));
	singlehash.y |= (1<<((*b_c)&31));
	singlehash.z |= (1<<((*b_d)&31));
        if (mylist) mylist = mylist->next;
    }

    if (a==1)
    {
	mylist = hash_list;
	memcpy(hex1,mylist->hash,4);
	memcpy(&singlehash.x, hex1, 4);
	memcpy(hex1,mylist->hash+4,4);
	memcpy(&singlehash.y, hex1, 4);
	memcpy(hex1,mylist->hash+8,4);
	memcpy(&singlehash.z, hex1, 4);
	memcpy(hex1,mylist->hash+12,4);
	memcpy(&singlehash.w, hex1, 4);
	REV(singlehash.x);
	REV(singlehash.y);
	REV(singlehash.z);
	REV(singlehash.w);
	singlehash.x -= H0;
	singlehash.y -= H1;
	singlehash.z -= H2;
	singlehash.w -= H3;

	clSetKernelArg(kernel, 8, sizeof(cl_uint4), (void*) &singlehash);
    }
    else 
    {
	singlehash.w=0;
	clSetKernelArg(kernel, 8, sizeof(cl_uint4), (void*) &singlehash);
    }

    found_ind=malloc(charset_size*charset_size*charset_size*sizeof(cl_uint));
    bzero(found_ind,sizeof(uint)*charset_size*charset_size*charset_size);
    found_ind_buf = clCreateBuffer(context[self], CL_MEM_WRITE_ONLY, charset_size*charset_size*charset_size*sizeof(cl_uint), NULL, &err );
    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);

    bitmaps_buf = clCreateBuffer(context[self], CL_MEM_READ_ONLY|CL_MEM_USE_HOST_PTR, 256*256*32*4, bitmaps, &err );
    if (err != CL_SUCCESS)
    {
	elog("clCreateBuffer error (%d)\n",err);
	return hash_err;
    }
    table_buf = clCreateBuffer(context[self], CL_MEM_READ_ONLY|CL_MEM_USE_HOST_PTR, charset_size*charset_size*charset_size*4, table, &err );
    if (err != CL_SUCCESS)
    {
	elog("clCreateBuffer error (%d)\n",err);
	return hash_err;
    }

    clSetKernelArg(kernel, 0, sizeof(cl_mem), (void*) &buffer);
    clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);
    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);
    clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
    clSetKernelArg(kernel, 4, sizeof(cl_mem), (void*) &found_ind_buf);
    clSetKernelArg(kernel, 5, sizeof(cl_mem), (void*) &bitmaps_buf);
    clSetKernelArg(kernel, 6, sizeof(cl_mem), (void*) &found_buf);
    clSetKernelArg(kernel, 7, sizeof(cl_mem), (void*) &table_buf);
    global_work_size = charset_size*charset_size*charset_size;
    while ((global_work_size %256)!=0) global_work_size++;

    if (err != CL_SUCCESS)
    {
	elog("clEnqueueBuffer error (%d)\n",err);
	return hash_err;
    }
    image.x=image.y=image.z=image.w=0;
    if (session_restore_lflag==0) ocl_markov_init_restore(self);


    /* Markov, len=4 */

    csize=1;
    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);
    clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);
    pthread_mutex_unlock(&biglock); // right now it should be safe to release the mutex

    if (markov_max_len>=4)
    if (thread_restore[self].cstart<=0)
    for (a1=thread_restore[self].cpos[0];a1<reduced_size;a1+=(ocl_vector*ocl_threads*devicesnum))
    {
	if (attack_over!=0) return hash_ok;
	cbase.s0 = reduced_charset[a1];
	cbase.s1 = reduced_charset[(a1+1)%reduced_size];
	cbase.s2 = reduced_charset[(a1+2)%reduced_size];
	cbase.s3 = reduced_charset[(a1+3)%reduced_size];
	cbase.s4 = reduced_charset[(a1+4)%reduced_size];
	cbase.s5 = reduced_charset[(a1+5)%reduced_size];
	cbase.s6 = reduced_charset[(a1+6)%reduced_size];
	cbase.s7 = reduced_charset[(a1+7)%reduced_size];

	thread_save[self].cpos[0]=a1;
	thread_save[self].cstr[0]=bruteforce_charset[a1];
	thread_save[self].cstr[1]=0;
	thread_save[self].cstart=0;

	clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
	image.x=0;image.y=0x80;image.z=0;image.w=0;
	REV(image.y);
	
	
	clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);
        if (found==1)
        {
    	    found = 0;
    	    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    	}
        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
	if (err != CL_SUCCESS)
	{
	    elog("clEnqueueNDRangeKernel error (%d)\n",err);
	    return hash_err;
	}

	invocations++;
	//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
        int *fnd;
        fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
        found = *fnd;
                     
        if (session_restore_lflag==1) ocl_markov_init_restore(self);
	session_restore_lflag = 0;

	if (found==1) 
	{
	    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*sizeof(uint), found_ind, 0, NULL, NULL);
	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    e=-1;
	    b=0;
	    for (a=0;a<charset_size;a++)
	    for (b=0;b<charset_size;b++)
	    for (c=0;c<charset_size;c++)
	    if (found_ind[a*charset_size*charset_size+b*charset_size+c]==1)
	    {
	    e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector;
    	    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, e*32, 32*ocl_vector, ptr+e*32, 0, NULL, NULL);
    	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    for (d=0;d<ocl_vector;d++)
	    {
		e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector+d;
    		if ( (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count<320) && (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>0))
    		{
    		    int i;
    		    for (i=0;i<hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count;i++)
    		    if (memcmp(hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i]->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
    		    {
                        mylist = hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i];
                        int flag = 0;
                        plain[0]=markov_charset[c];
                        plain[1]=markov_charset[b];
                        plain[2]=markov_charset[a];
                        plain[3]=reduced_charset[(a1+(d))%charset_size];
                        plain[4]=0;
                        pthread_mutex_lock(&crackedmutex);
                        addlist = cracked_list;
                        while (addlist)
                        {
                            if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                            addlist = addlist->next;
                        }
                        pthread_mutex_unlock(&crackedmutex);
                        if (flag == 0)
                        {
                            add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                        }
    		    }
    		}
    		else if (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>=320)
    		{
    		    mylist = hash_list;
    		    while (mylist)
    		    {
            		if (memcmp(mylist->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
            		{
                    	    int flag = 0;
                    	    plain[0]=markov_charset[c];
                    	    plain[1]=markov_charset[b];
                    	    plain[2]=markov_charset[a];
                    	    plain[3]=reduced_charset[(a1+(d))%charset_size];
                    	    plain[4]=0;
                    	    pthread_mutex_lock(&crackedmutex);
                    	    addlist = cracked_list;
                    	    while (addlist)
                    	    {
                        	if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                        	addlist = addlist->next;
                    	    }
                    	    pthread_mutex_unlock(&crackedmutex);
                    	    if (flag == 0)
                    	    {
                        	add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    	    }
            		}
            		if (mylist) mylist = mylist->next;
        	    }
        	}
	    }
	    }
	clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
        bzero(found_ind,charset_size*charset_size*charset_size*sizeof(cl_uint));
	clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
	}
    }


    /* Markov, len=5 */

    csize=2;
    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);

    if (markov_max_len>=5)
    if (thread_restore[self].cstart<=1)
    for (a1=thread_restore[self].cpos[0];a1<reduced_size;a1+=(ocl_vector*ocl_threads*devicesnum))
    for (a2=thread_restore[self].cpos[1];a2<charset_size;a2++) 
    if ((markov2[a1][a2] > markov_threshold)||(markov2[a1+1][a2] > markov_threshold)
    ||(markov2[a1+2][a2] > markov_threshold)||(markov2[a1+3][a2] > markov_threshold)
    ||(markov2[a1+4][a2] > markov_threshold)||(markov2[a1+5][a2] > markov_threshold)
    ||(markov2[a1+6][a2] > markov_threshold)||(markov2[a1+7][a2] > markov_threshold)
    )
    {
	if (attack_over!=0) return hash_ok;
	cbase.s0 = reduced_charset[a1];
	cbase.s1 = reduced_charset[(a1+1)];
	cbase.s2 = reduced_charset[(a1+2)];
	cbase.s3 = reduced_charset[(a1+3)];
	cbase.s4 = reduced_charset[(a1+4)];
	cbase.s5 = reduced_charset[(a1+5)];
	cbase.s6 = reduced_charset[(a1+6)];
	cbase.s7 = reduced_charset[(a1+7)];
	thread_save[self].cpos[0]=a1;
	thread_save[self].cpos[1]=a2;
	thread_save[self].cstr[0]=bruteforce_charset[a1];
	thread_save[self].cstr[1]=bruteforce_charset[a2];
	thread_save[self].cstr[2]=0;
	thread_save[self].cstart=1;

	clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
	image.x=0;
	image.y=markov_charset[a2]|(0x80<<8);
	image.z=0;image.w=0;
	REV(image.y);
	clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);
        if (found==1)
        {
    	    found = 0;
    	    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    	}
        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
        //clFlush(queue);
	invocations++;
	if (err != CL_SUCCESS)
	{
	    elog("clEnqueueNDRangeKernel error (%d)\n",err);
	    return hash_err;
	}
	//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
        int *fnd;
        fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
        found = *fnd;

        if (session_restore_lflag==1) ocl_markov_init_restore(self);
	session_restore_lflag = 0;

	if (found==1) 
	{
	    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*sizeof(uint), found_ind, 0, NULL, NULL);
	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    e=-1;
	    b=0;
	    for (a=0;a<charset_size;a++)
	    for (b=0;b<charset_size;b++)
	    for (c=0;c<charset_size;c++)
	    if (found_ind[a*charset_size*charset_size+b*charset_size+c]==1)
	    {
	    e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector;
    	    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, e*32, 32*ocl_vector, ptr+e*32, 0, NULL, NULL);
    	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    for (d=0;d<ocl_vector;d++)
	    {
		e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector+d;
    		if ( (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count<320) && (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>0))
    		{
    		    int i;
    		    for (i=0;i<hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count;i++)
    		    if (memcmp(hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i]->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
    		    {
                        mylist = hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i];
                        int flag = 0;
                        plain[0]=markov_charset[c];
                        plain[1]=markov_charset[b];
                        plain[2]=markov_charset[a];
                        plain[3]=reduced_charset[(a1+(d))%charset_size];
                        plain[4]=markov_charset[a2];
			plain[5]=0;
                        pthread_mutex_lock(&crackedmutex);
                        addlist = cracked_list;
                        while (addlist)
                        {
                            if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                            addlist = addlist->next;
                        }
                        pthread_mutex_unlock(&crackedmutex);
                        if (flag == 0)
                        {
                            add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                        }
    		    }
    		}
    		else if (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>=320)
    		{
    		    mylist = hash_list;
    		    while (mylist)
    		    {
            		if (memcmp(mylist->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
            		{
                    	    int flag = 0;
                    	    plain[0]=markov_charset[c];
                    	    plain[1]=markov_charset[b];
                    	    plain[2]=markov_charset[a];
                    	    plain[3]=reduced_charset[(a1+(d))%charset_size];
                    	    plain[4]=markov_charset[a2];
			    plain[5]=0;
                    	    pthread_mutex_lock(&crackedmutex);
                    	    addlist = cracked_list;
                    	    while (addlist)
                    	    {
                        	if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                        	addlist = addlist->next;
                    	    }
                    	    pthread_mutex_unlock(&crackedmutex);
                    	    if (flag == 0)
                    	    {
                        	add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    	    }
            		}
            		if (mylist) mylist = mylist->next;
        	    }
        	}
	    }
	    }
	clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
        bzero(found_ind,charset_size*charset_size*charset_size*sizeof(cl_uint));
	clEnqueueWriteBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
	}
    }


    /* Markov, len=6 */
    csize=3;
    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);

    if (markov_max_len>=6)
    if (thread_restore[self].cstart<=2)
    for (a1=thread_restore[self].cpos[0];a1<reduced_size;a1+=(ocl_vector*ocl_threads*devicesnum))
    for (a2=thread_restore[self].cpos[1];a2<charset_size;a2++) 
    if ((markov2[a1][a2] > markov_threshold)||(markov2[a1+1][a2] > markov_threshold)
    ||(markov2[a1+2][a2] > markov_threshold)||(markov2[a1+3][a2] > markov_threshold)
    ||(markov2[a1+4][a2] > markov_threshold)||(markov2[a1+5][a2] > markov_threshold)
    ||(markov2[a1+6][a2] > markov_threshold)||(markov2[a1+7][a2] > markov_threshold)
    )
    for (a3=thread_restore[self].cpos[2];a3<charset_size;a3++) if (markov1[a2][a3] > markov_threshold)
    {
	if (attack_over!=0) return hash_ok;
	cbase.s0 = reduced_charset[a1%reduced_size];
	cbase.s1 = reduced_charset[(a1+1)];
	cbase.s2 = reduced_charset[(a1+2)];
	cbase.s3 = reduced_charset[(a1+3)];
	cbase.s4 = reduced_charset[(a1+4)];
	cbase.s5 = reduced_charset[(a1+5)];
	cbase.s6 = reduced_charset[(a1+6)];
	cbase.s7 = reduced_charset[(a1+7)];
	thread_save[self].cpos[0]=a1;
	thread_save[self].cpos[1]=a2;
	thread_save[self].cpos[2]=a3;
	thread_save[self].cstr[0]=bruteforce_charset[a1];
	thread_save[self].cstr[1]=bruteforce_charset[a2];
	thread_save[self].cstr[2]=bruteforce_charset[a3];
	thread_save[self].cstr[3]=0;
	thread_save[self].cstart=2;

	clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
	image.x=0;
	image.y=markov_charset[a2]|(markov_charset[a3]<<8)|(0x80<<16);
	image.z=0;image.w=0;
	REV(image.y);
	clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);
        if (found==1)
        {
    	    found = 0;
    	    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    	}
        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
        //clFlush(queue);
	invocations++;
	if (err != CL_SUCCESS)
	{
	    elog("clEnqueueNDRangeKernel error (%d)\n",err);
	    return hash_err;
	}
	//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
        int *fnd;
        fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
        found = *fnd;

        if (session_restore_lflag==1) ocl_markov_init_restore(self);
	session_restore_lflag = 0;

	if (found==1) 
	{
	    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*sizeof(uint), found_ind, 0, NULL, NULL);
	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    e=-1;
	    for (a=0;a<charset_size;a++)
	    for (b=0;b<charset_size;b++)
	    for (c=0;c<charset_size;c++)
	    if (found_ind[a*charset_size*charset_size+b*charset_size+c]==1)
	    {
	    e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector;
    	    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, e*32, 32*ocl_vector, ptr+e*32, 0, NULL, NULL);
	    for (d=0;d<ocl_vector;d++)
	    {
		e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector+d;
    		if ( (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count<320) && (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>0))
    		{
    		    int i;
    		    for (i=0;i<hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count;i++)
    		    if (memcmp(hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i]->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
    		    {
                        mylist = hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i];
                        int flag = 0;
                        plain[0]=markov_charset[c];
                        plain[1]=markov_charset[b];
                        plain[2]=markov_charset[a];
                        plain[3]=reduced_charset[(a1+(d))%charset_size];
                        plain[4]=markov_charset[a2];
                        plain[5]=markov_charset[a3];
			plain[6]=0;
                        pthread_mutex_lock(&crackedmutex);
                        addlist = cracked_list;
                        while (addlist)
                        {
                            if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                            addlist = addlist->next;
                        }
                        pthread_mutex_unlock(&crackedmutex);
                        if (flag == 0)
                        {
                            add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                        }
    		    }
    		}
    		else if (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>=320)
    		{
    		    mylist = hash_list;
    		    while (mylist)
    		    {
            		if (memcmp(mylist->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
            		{
                    	    int flag = 0;
                    	    plain[0]=markov_charset[c];
                    	    plain[1]=markov_charset[b];
                    	    plain[2]=markov_charset[a];
                    	    plain[3]=reduced_charset[(a1+(d))%charset_size];
                    	    plain[4]=markov_charset[a2];
                    	    plain[5]=markov_charset[a3];
			    plain[6]=0;
                    	    pthread_mutex_lock(&crackedmutex);
                    	    addlist = cracked_list;
/*                    	    while (addlist)
                    	    {
                        	if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                        	addlist = addlist->next;
                    	    }
                    	    pthread_mutex_unlock(&crackedmutex);*/
                    	    if (flag == 0)
                    	    {
                        	add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    	    }
            		}
            		if (mylist) mylist = mylist->next;
        	    }
        	}
	    }
	    }
	clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
        bzero(found_ind,charset_size*charset_size*charset_size*sizeof(cl_uint));
        clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
	}
    }

    /* Markov, len=7 */
    csize=4;
    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);

    if (markov_max_len>=7)
    if (thread_restore[self].cstart<=3)
    for (a1=thread_restore[self].cpos[0];a1<reduced_size;a1+=(ocl_vector*ocl_threads*devicesnum))
    for (a2=thread_restore[self].cpos[1];a2<charset_size;a2++) if ((markov2[a1][a2] > markov_threshold)||(markov2[a1+1][a2] > markov_threshold)||(markov2[a1+2][a2] > markov_threshold)||(markov2[a1+3][a2] > markov_threshold))
    for (a3=thread_restore[self].cpos[2];a3<charset_size;a3++) if (markov1[a2][a3] > markov_threshold)
    for (a4=thread_restore[self].cpos[3];a4<charset_size;a4++) if (markov1[a3][a4] > markov_threshold)
    {
	if (attack_over!=0) return hash_ok;
	cbase.s0 = reduced_charset[a1];
	cbase.s1 = reduced_charset[a1+1];
	cbase.s2 = reduced_charset[a1+2];
	cbase.s3 = reduced_charset[a1+3];
	cbase.s4 = reduced_charset[a1+4];
	cbase.s5 = reduced_charset[a1+5];
	cbase.s6 = reduced_charset[a1+6];
	cbase.s7 = reduced_charset[a1+7];
	thread_save[self].cpos[0]=a1;
	thread_save[self].cpos[1]=a2;
	thread_save[self].cpos[2]=a3;
	thread_save[self].cpos[3]=a4;
	thread_save[self].cstr[0]=bruteforce_charset[a1];
	thread_save[self].cstr[1]=bruteforce_charset[a2];
	thread_save[self].cstr[2]=bruteforce_charset[a3];
	thread_save[self].cstr[3]=bruteforce_charset[a4];
	thread_save[self].cstr[4]=0;
	thread_save[self].cstart=3;

	clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
	image.x=0;
	image.y=markov_charset[a2]|(markov_charset[a3]<<8)|(markov_charset[a4]<<16)|(0x80<<24);
	image.z=0;image.w=0;
	REV(image.y);
	clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);
        if (found==1)
        {
    	    found = 0;
    	    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    	}
        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
        //clFlush(queue);
	invocations++;
	if (err != CL_SUCCESS)
	{
	    elog("clEnqueueNDRangeKernel error (%d)\n",err);
	    return hash_err;
	}
	//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
        int *fnd;
        fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
        found = *fnd;

        if (session_restore_lflag==1) ocl_markov_init_restore(self);
	session_restore_lflag = 0;

	if (found==1) 
	{
	    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*sizeof(uint), found_ind, 0, NULL, NULL);
	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    e=-1;

	    for (a=0;a<charset_size;a++)
	    for (b=0;b<charset_size;b++)
	    for (c=0;c<charset_size;c++)
	    if (found_ind[a*charset_size*charset_size+b*charset_size+c]==1)
	    {
	    e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector;
    	    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, e*32, 32*ocl_vector, ptr+e*32, 0, NULL, NULL);
	    for (d=0;d<ocl_vector;d++)
	    {
		e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector+d;
    		if ( (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count<320) && (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>0))
    		{
    		    int i;
    		    for (i=0;i<hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count;i++)
    		    if (memcmp(hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i]->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
    		    {
                        mylist = hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i];
                        int flag = 0;
                        plain[0]=markov_charset[c];
                        plain[1]=markov_charset[b];
                        plain[2]=markov_charset[a];
                        plain[3]=reduced_charset[(a1+(d))%charset_size];
                        plain[4]=markov_charset[a2];
                        plain[5]=markov_charset[a3];
                        plain[6]=markov_charset[a4];
			plain[7]=0;
                        pthread_mutex_lock(&crackedmutex);
                        addlist = cracked_list;
                        while (addlist)
                        {
                            if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                            addlist = addlist->next;
                        }
                        pthread_mutex_unlock(&crackedmutex);
                        if (flag == 0)
                        {
                            add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                        }
    		    }
    		}
    		else if (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>=320)
    		{
    		    mylist = hash_list;
    		    while (mylist)
    		    {
            		if (memcmp(mylist->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
            		{
                    	    int flag = 0;
                    	    plain[0]=markov_charset[c];
                    	    plain[1]=markov_charset[b];
                    	    plain[2]=markov_charset[a];
                    	    plain[3]=reduced_charset[(a1+(d))%charset_size];
                    	    plain[4]=markov_charset[a2];
                    	    plain[5]=markov_charset[a3];
                    	    plain[6]=markov_charset[a4];
			    plain[7]=0;
                    	    pthread_mutex_lock(&crackedmutex);
                    	    addlist = cracked_list;
                    	    while (addlist)
                    	    {
                        	if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                        	addlist = addlist->next;
                    	    }
                    	    pthread_mutex_unlock(&crackedmutex);
                    	    if (flag == 0)
                    	    {
                        	add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    	    }
            		}
            		if (mylist) mylist = mylist->next;
        	    }
        	}
	    }
	    }
	    clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
	    bzero(found_ind,charset_size*charset_size*charset_size*sizeof(cl_uint));
	    clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
	}
    }



    /* Markov, len=8 */
    csize=5;
    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);

    if (markov_max_len>=8)
    if (thread_restore[self].cstart<=4)
    for (a1=thread_restore[self].cpos[0];a1<reduced_size;a1+=(ocl_vector*ocl_threads*devicesnum))
    for (a2=thread_restore[self].cpos[1];a2<charset_size;a2++) if ((markov2[a1][a2] > markov_threshold)||(markov2[a1+1][a2] > markov_threshold)||(markov2[a1+2][a2] > markov_threshold)||(markov2[a1+3][a2] > markov_threshold))
    for (a3=thread_restore[self].cpos[2];a3<charset_size;a3++) if (markov1[a2][a3] > markov_threshold)
    for (a4=thread_restore[self].cpos[3];a4<charset_size;a4++) if (markov1[a3][a4] > markov_threshold)
    for (a5=thread_restore[self].cpos[4];a5<charset_size;a5++) if (markov1[a4][a5] > markov_threshold)
    {
	if (attack_over!=0) return hash_ok;
	cbase.s0 = reduced_charset[a1];
	cbase.s1 = reduced_charset[a1+1];
	cbase.s2 = reduced_charset[a1+2];
	cbase.s3 = reduced_charset[a1+3];
	cbase.s4 = reduced_charset[a1+4];
	cbase.s5 = reduced_charset[a1+5];
	cbase.s6 = reduced_charset[a1+6];
	cbase.s7 = reduced_charset[a1+7];
	thread_save[self].cpos[0]=a1;
	thread_save[self].cpos[1]=a2;
	thread_save[self].cpos[2]=a3;
	thread_save[self].cpos[3]=a4;
	thread_save[self].cpos[4]=a5;
	thread_save[self].cstr[0]=bruteforce_charset[a1];
	thread_save[self].cstr[1]=bruteforce_charset[a2];
	thread_save[self].cstr[2]=bruteforce_charset[a3];
	thread_save[self].cstr[3]=bruteforce_charset[a4];
	thread_save[self].cstr[4]=bruteforce_charset[a5];
	thread_save[self].cstr[5]=0;
	thread_save[self].cstart=4;

	clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
	image.x=0;
	image.y=markov_charset[a2]|(markov_charset[a3]<<8)|(markov_charset[a4]<<16)|(markov_charset[a5]<<24);
	image.z=0x80;image.w=0;
	REV(image.y);
	REV(image.z);
	clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);
        if (found==1)
        {
    	    found = 0;
    	    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    	}
        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
        //clFlush(queue);
	invocations++;
	if (err != CL_SUCCESS)
	{
	    elog("clEnqueueNDRangeKernel error (%d)\n",err);
	    return hash_err;
	}
	//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
        int *fnd;
        fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
        found = *fnd;

        if (session_restore_lflag==1) ocl_markov_init_restore(self);
	session_restore_lflag = 0;

	if (found==1) 
	{

	    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*sizeof(uint), found_ind, 0, NULL, NULL);
	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    e=-1;

	    for (a=0;a<charset_size;a++)
	    for (b=0;b<charset_size;b++)
	    for (c=0;c<charset_size;c++)
	    if (found_ind[a*charset_size*charset_size+b*charset_size+c]==1)
	    {
	    e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector;
    	    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, e*32, 32*ocl_vector, ptr+e*32, 0, NULL, NULL);
	    for (d=0;d<ocl_vector;d++)
	    {
		e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector+d;
    		if ( (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count<320) && (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>0))
    		{
    		    int i;
    		    for (i=0;i<hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count;i++)
    		    if (memcmp(hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i]->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
    		    {
                        mylist = hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i];
                        int flag = 0;
                        plain[0]=markov_charset[c];
                        plain[1]=markov_charset[b];
                        plain[2]=markov_charset[a];
                        plain[3]=reduced_charset[(a1+(d))%charset_size];
                        plain[4]=markov_charset[a2];
                        plain[5]=markov_charset[a3];
                        plain[6]=markov_charset[a4];
                        plain[7]=markov_charset[a5];
			plain[8]=0;
                        pthread_mutex_lock(&crackedmutex);
                        addlist = cracked_list;
                        while (addlist)
                        {
                            if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                            addlist = addlist->next;
                        }
                        pthread_mutex_unlock(&crackedmutex);
                        if (flag == 0)
                        {
                            add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                        }
    		    }
    		}
    		else if (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>=320)
    		{
    		    mylist = hash_list;
    		    while (mylist)
    		    {
            		if (memcmp(mylist->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
            		{
                    	    int flag = 0;
                    	    plain[0]=markov_charset[c];
                    	    plain[1]=markov_charset[b];
                    	    plain[2]=markov_charset[a];
                    	    plain[3]=reduced_charset[(a1+(d))%charset_size];
                    	    plain[4]=markov_charset[a2];
                    	    plain[5]=markov_charset[a3];
                    	    plain[6]=markov_charset[a4];
                    	    plain[7]=markov_charset[a5];
			    plain[8]=0;
                    	    pthread_mutex_lock(&crackedmutex);
                    	    addlist = cracked_list;
                    	    while (addlist)
                    	    {
                        	if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                        	addlist = addlist->next;
                    	    }
                    	    pthread_mutex_unlock(&crackedmutex);
                    	    if (flag == 0)
                    	    {
                        	add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    	    }
            		}
            		if (mylist) mylist = mylist->next;
        	    }
        	}
	    }
	    }
	    clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
	    bzero(found_ind,charset_size*charset_size*charset_size*sizeof(cl_uint));
	    clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
	}
    }


    /* Markov, len=9 */
    csize=6;
    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);

    if (markov_max_len>=9)
    if (thread_restore[self].cstart<=5)
    for (a1=thread_restore[self].cpos[0];a1<reduced_size;a1+=(ocl_vector*ocl_threads*devicesnum))
    for (a2=thread_restore[self].cpos[1];a2<charset_size;a2++) if ((markov2[a1][a2] > markov_threshold)||(markov2[a1+1][a2] > markov_threshold)||(markov2[a1+2][a2] > markov_threshold)||(markov2[a1+3][a2] > markov_threshold))
    for (a3=thread_restore[self].cpos[2];a3<charset_size;a3++) if (markov1[a2][a3] > markov_threshold)
    for (a4=thread_restore[self].cpos[3];a4<charset_size;a4++) if (markov1[a3][a4] > markov_threshold)
    for (a5=thread_restore[self].cpos[4];a5<charset_size;a5++) if (markov1[a4][a5] > markov_threshold)
    for (a6=thread_restore[self].cpos[5];a6<charset_size;a6++) if (markov1[a5][a6] > markov_threshold)
    {
	if (attack_over!=0) return hash_ok;
	cbase.s0 = reduced_charset[a1];
	cbase.s1 = reduced_charset[a1+1];
	cbase.s2 = reduced_charset[a1+2];
	cbase.s3 = reduced_charset[a1+3];
	cbase.s4 = reduced_charset[a1+4];
	cbase.s5 = reduced_charset[a1+5];
	cbase.s6 = reduced_charset[a1+6];
	cbase.s7 = reduced_charset[a1+7];
	thread_save[self].cpos[0]=a1;
	thread_save[self].cpos[1]=a2;
	thread_save[self].cpos[2]=a3;
	thread_save[self].cpos[3]=a4;
	thread_save[self].cpos[4]=a5;
	thread_save[self].cpos[5]=a6;
	thread_save[self].cstr[0]=bruteforce_charset[a1];
	thread_save[self].cstr[1]=bruteforce_charset[a2];
	thread_save[self].cstr[2]=bruteforce_charset[a3];
	thread_save[self].cstr[3]=bruteforce_charset[a4];
	thread_save[self].cstr[4]=bruteforce_charset[a5];
	thread_save[self].cstr[5]=bruteforce_charset[a6];
	thread_save[self].cstr[6]=0;
	thread_save[self].cstart=5;

	clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
	image.x=0;
	image.y=markov_charset[a2]|(markov_charset[a3]<<8)|(markov_charset[a4]<<16)|(markov_charset[a5]<<24);
	image.z=markov_charset[a6]|(0x80<<8);image.w=0;
	REV(image.y);
	REV(image.z);
	clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);
        if (found==1)
        {
    	    found = 0;
    	    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    	}
        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
        //clFlush(queue);
	invocations++;
	if (err != CL_SUCCESS)
	{
	    elog("clEnqueueNDRangeKernel error (%d)\n",err);
	    return hash_err;
	}
	//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
        int *fnd;
        fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
        found = *fnd;

        if (session_restore_lflag==1) ocl_markov_init_restore(self);
	session_restore_lflag = 0;

	if (found==1) 
	{
	    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*sizeof(uint), found_ind, 0, NULL, NULL);
	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    e=-1;

	    for (a=0;a<charset_size;a++)
	    for (b=0;b<charset_size;b++)
	    for (c=0;c<charset_size;c++)
	    if (found_ind[a*charset_size*charset_size+b*charset_size+c]==1)
	    {
	    e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector;
    	    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, e*32, 32*ocl_vector, ptr+e*32, 0, NULL, NULL);
	    for (d=0;d<ocl_vector;d++)
	    {
		e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector+d;
    		if ( (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count<320) && (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>0))
    		{
    		    int i;
    		    for (i=0;i<hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count;i++)
    		    if (memcmp(hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i]->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
    		    {
                        mylist = hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i];
                        int flag = 0;
                        plain[0]=markov_charset[c];
                        plain[1]=markov_charset[b];
                        plain[2]=markov_charset[a];
                        plain[3]=reduced_charset[(a1+(d))%charset_size];
                        plain[4]=markov_charset[a2];
                        plain[5]=markov_charset[a3];
                        plain[6]=markov_charset[a4];
                        plain[7]=markov_charset[a5];
                        plain[8]=markov_charset[a6];
			plain[9]=0;
                        pthread_mutex_lock(&crackedmutex);
                        addlist = cracked_list;
                        while (addlist)
                        {
                            if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                            addlist = addlist->next;
                        }
                        pthread_mutex_unlock(&crackedmutex);
                        if (flag == 0)
                        {
                            add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                        }
    		    }
    		}
    		else if (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>=320)
    		{
    		    mylist = hash_list;
    		    while (mylist)
    		    {
            		if (memcmp(mylist->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
            		{
                    	    int flag = 0;
                    	    plain[0]=markov_charset[c];
                    	    plain[1]=markov_charset[b];
                    	    plain[2]=markov_charset[a];
                    	    plain[3]=reduced_charset[(a1+(d))%charset_size];
                    	    plain[4]=markov_charset[a2];
                    	    plain[5]=markov_charset[a3];
                    	    plain[6]=markov_charset[a4];
                    	    plain[7]=markov_charset[a5];
                    	    plain[8]=markov_charset[a6];
			    plain[9]=0;
                    	    pthread_mutex_lock(&crackedmutex);
                    	    addlist = cracked_list;
                    	    while (addlist)
                    	    {
                        	if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                        	addlist = addlist->next;
                    	    }
                    	    pthread_mutex_unlock(&crackedmutex);
                    	    if (flag == 0)
                    	    {
                        	add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    	    }
            		}
            		if (mylist) mylist = mylist->next;
        	    }
        	}
	    }
	    }
	clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
        bzero(found_ind,charset_size*charset_size*charset_size*sizeof(cl_uint));
        clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
	}
    }


    /* Markov, len=10 */
    csize=6;
    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);

    if (markov_max_len>=10)
    if (thread_restore[self].cstart<=6)
    for (a1=thread_restore[self].cpos[0];a1<reduced_size;a1+=(ocl_vector*ocl_threads*devicesnum))
    for (a2=thread_restore[self].cpos[1];a2<charset_size;a2++) if ((markov2[a1][a2] > markov_threshold)||(markov2[a1+1][a2] > markov_threshold)||(markov2[a1+2][a2] > markov_threshold)||(markov2[a1+3][a2] > markov_threshold))
    for (a3=thread_restore[self].cpos[2];a3<charset_size;a3++) if (markov1[a2][a3] > markov_threshold)
    for (a4=thread_restore[self].cpos[3];a4<charset_size;a4++) if (markov1[a3][a4] > markov_threshold)
    for (a5=thread_restore[self].cpos[4];a5<charset_size;a5++) if (markov1[a4][a5] > markov_threshold)
    for (a6=thread_restore[self].cpos[5];a6<charset_size;a6++) if (markov1[a5][a6] > markov_threshold)
    for (a7=thread_restore[self].cpos[6];a7<charset_size;a7++) if (markov1[a6][a7] > markov_threshold)
    {
	if (attack_over!=0) return hash_ok;
	cbase.s0 = reduced_charset[a1];
	cbase.s1 = reduced_charset[a1+1];
	cbase.s2 = reduced_charset[a1+2];
	cbase.s3 = reduced_charset[a1+3];
	cbase.s4 = reduced_charset[a1+4];
	cbase.s5 = reduced_charset[a1+5];
	cbase.s6 = reduced_charset[a1+6];
	cbase.s7 = reduced_charset[a1+7];

	thread_save[self].cpos[0]=a1;
	thread_save[self].cpos[1]=a2;
	thread_save[self].cpos[2]=a3;
	thread_save[self].cpos[3]=a4;
	thread_save[self].cpos[4]=a5;
	thread_save[self].cpos[5]=a6;
	thread_save[self].cpos[6]=a7;
	thread_save[self].cstr[0]=bruteforce_charset[a1];
	thread_save[self].cstr[1]=bruteforce_charset[a2];
	thread_save[self].cstr[2]=bruteforce_charset[a3];
	thread_save[self].cstr[3]=bruteforce_charset[a4];
	thread_save[self].cstr[4]=bruteforce_charset[a5];
	thread_save[self].cstr[5]=bruteforce_charset[a6];
	thread_save[self].cstr[6]=bruteforce_charset[a7];
	thread_save[self].cstr[7]=0;
	thread_save[self].cstart=6;

	clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
	image.x=0;
	image.y=markov_charset[a2]|(markov_charset[a3]<<8)|(markov_charset[a4]<<16)|(markov_charset[a5]<<24);
	image.z=markov_charset[a6]|(markov_charset[a7]<<8)|(0x80<<16);image.w=0;
	REV(image.y);
	REV(image.z);
	clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);
        if (found==1)
        {
    	    found = 0;
    	    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    	}
        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
        //clFlush(queue);
	invocations++;
	if (err != CL_SUCCESS)
	{
	    elog("clEnqueueNDRangeKernel error (%d)\n",err);
	    return hash_err;
	}
	//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
        int *fnd;
        fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
        found = *fnd;

        if (session_restore_lflag==1) ocl_markov_init_restore(self);
	session_restore_lflag = 0;

	if (found==1) 
	{
	    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*sizeof(uint), found_ind, 0, NULL, NULL);
	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    e=-1;

	    for (a=0;a<charset_size;a++)
	    for (b=0;b<charset_size;b++)
	    for (c=0;c<charset_size;c++)
	    if (found_ind[a*charset_size*charset_size+b*charset_size+c]==1)
	    {
	    e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector;
    	    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, e*32, 32*ocl_vector, ptr+e*32, 0, NULL, NULL);
	    for (d=0;d<ocl_vector;d++)
	    {
		e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector+d;
    		if ( (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count<320) && (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>0))
    		{
    		    int i;
    		    for (i=0;i<hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count;i++)
    		    if (memcmp(hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i]->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
    		    {
                        mylist = hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i];
                        int flag = 0;
                        plain[0]=markov_charset[c];
                        plain[1]=markov_charset[b];
                        plain[2]=markov_charset[a];
                        plain[3]=reduced_charset[(a1+(d))%charset_size];
                        plain[4]=markov_charset[a2];
                        plain[5]=markov_charset[a3];
                        plain[6]=markov_charset[a4];
                        plain[7]=markov_charset[a5];
                        plain[8]=markov_charset[a6];
                        plain[9]=markov_charset[a7];
			plain[10]=0;
                        pthread_mutex_lock(&crackedmutex);
                        addlist = cracked_list;
                        while (addlist)
                        {
                            if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                            addlist = addlist->next;
                        }
                        pthread_mutex_unlock(&crackedmutex);
                        if (flag == 0)
                        {
                            add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                        }
    		    }
    		}
    		else if (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>=320)
    		{
    		    mylist = hash_list;
    		    while (mylist)
    		    {
            		if (memcmp(mylist->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
            		{
                    	    int flag = 0;
                    	    plain[0]=markov_charset[c];
                    	    plain[1]=markov_charset[b];
                    	    plain[2]=markov_charset[a];
                    	    plain[3]=reduced_charset[(a1+(d))%charset_size];
                    	    plain[4]=markov_charset[a2];
                    	    plain[5]=markov_charset[a3];
                    	    plain[6]=markov_charset[a4];
                    	    plain[7]=markov_charset[a5];
                    	    plain[8]=markov_charset[a6];
                    	    plain[9]=markov_charset[a7];
			    plain[10]=0;
                    	    pthread_mutex_lock(&crackedmutex);
                    	    addlist = cracked_list;
                    	    while (addlist)
                    	    {
                        	if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                        	addlist = addlist->next;
                    	    }
                    	    pthread_mutex_unlock(&crackedmutex);
                    	    if (flag == 0)
                    	    {
                        	add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    	    }
            		}
            		if (mylist) mylist = mylist->next;
        	    }
        	}
	    }
	    }
	clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
        bzero(found_ind,charset_size*charset_size*charset_size*sizeof(cl_uint));
        clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
	}
    }


    /* Markov, len=11 */
    csize=6;
    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);

    if (markov_max_len>=11)
    if (thread_restore[self].cstart<=7)
    for (a1=thread_restore[self].cpos[0];a1<reduced_size;a1+=(ocl_vector*ocl_threads*devicesnum))
    for (a2=thread_restore[self].cpos[1];a2<charset_size;a2++) if ((markov2[a1][a2] > markov_threshold)||(markov2[a1+1][a2] > markov_threshold)||(markov2[a1+2][a2] > markov_threshold)||(markov2[a1+3][a2] > markov_threshold))
    for (a3=thread_restore[self].cpos[2];a3<charset_size;a3++) if (markov1[a2][a3] > markov_threshold)
    for (a4=thread_restore[self].cpos[3];a4<charset_size;a4++) if (markov1[a3][a4] > markov_threshold)
    for (a5=thread_restore[self].cpos[4];a5<charset_size;a5++) if (markov1[a4][a5] > markov_threshold)
    for (a6=thread_restore[self].cpos[5];a6<charset_size;a6++) if (markov1[a5][a6] > markov_threshold)
    for (a7=thread_restore[self].cpos[6];a7<charset_size;a7++) if (markov1[a6][a7] > markov_threshold)
    for (a8=thread_restore[self].cpos[7];a8<charset_size;a8++) if (markov1[a7][a8] > markov_threshold)
    {
	if (attack_over!=0) return hash_ok;
	cbase.s0 = reduced_charset[a1];
	cbase.s1 = reduced_charset[a1+1];
	cbase.s2 = reduced_charset[a1+2];
	cbase.s3 = reduced_charset[a1+3];
	cbase.s4 = reduced_charset[a1+4];
	cbase.s5 = reduced_charset[a1+5];
	cbase.s6 = reduced_charset[a1+6];
	cbase.s7 = reduced_charset[a1+7];
	thread_save[self].cpos[0]=a1;
	thread_save[self].cpos[1]=a2;
	thread_save[self].cpos[2]=a3;
	thread_save[self].cpos[3]=a4;
	thread_save[self].cpos[4]=a5;
	thread_save[self].cpos[5]=a6;
	thread_save[self].cpos[6]=a7;
	thread_save[self].cpos[7]=a8;
	thread_save[self].cstr[0]=bruteforce_charset[a1];
	thread_save[self].cstr[1]=bruteforce_charset[a2];
	thread_save[self].cstr[2]=bruteforce_charset[a3];
	thread_save[self].cstr[3]=bruteforce_charset[a4];
	thread_save[self].cstr[4]=bruteforce_charset[a5];
	thread_save[self].cstr[5]=bruteforce_charset[a6];
	thread_save[self].cstr[6]=bruteforce_charset[a7];
	thread_save[self].cstr[7]=bruteforce_charset[a8];
	thread_save[self].cstr[8]=0;
	thread_save[self].cstart=7;

	clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
	image.x=0;
	image.y=markov_charset[a2]|(markov_charset[a3]<<8)|(markov_charset[a4]<<16)|(markov_charset[a5]<<24);
	image.z=markov_charset[a6]|(markov_charset[a7]<<8)|(markov_charset[a8]<<16)|(0x80<<24);image.w=0;
	REV(image.y);
	REV(image.z);
	clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);
        if (found==1)
        {
    	    found = 0;
    	    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    	}
        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
        //clFlush(queue);
	invocations++;
	if (err != CL_SUCCESS)
	{
	    elog("clEnqueueNDRangeKernel error (%d)\n",err);
	    return hash_err;
	}
	//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
        int *fnd;
        fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
        found = *fnd;

        if (session_restore_lflag==1) ocl_markov_init_restore(self);
	session_restore_lflag = 0;

	if (found==1) 
	{
	    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*sizeof(uint), found_ind, 0, NULL, NULL);
	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    e=-1;

	    for (a=0;a<charset_size;a++)
	    for (b=0;b<charset_size;b++)
	    for (c=0;c<charset_size;c++)
	    if (found_ind[a*charset_size*charset_size+b*charset_size+c]==1)
	    {
	    e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector;
    	    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, e*32, 32*ocl_vector, ptr+e*32, 0, NULL, NULL);
	    for (d=0;d<ocl_vector;d++)
	    {
		e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector+d;
    		if ( (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count<320) && (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>0))
    		{
    		    int i;
    		    for (i=0;i<hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count;i++)
    		    if (memcmp(hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i]->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
    		    {
                        mylist = hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i];
                        int flag = 0;
                        plain[0]=markov_charset[c];
                        plain[1]=markov_charset[b];
                        plain[2]=markov_charset[a];
                        plain[3]=reduced_charset[(a1+(d))%charset_size];
                        plain[4]=markov_charset[a2];
                        plain[5]=markov_charset[a3];
                        plain[6]=markov_charset[a4];
                        plain[7]=markov_charset[a5];
                        plain[8]=markov_charset[a6];
                        plain[9]=markov_charset[a7];
                        plain[10]=markov_charset[a8];
			plain[11]=0;
                        pthread_mutex_lock(&crackedmutex);
                        addlist = cracked_list;
                        while (addlist)
                        {
                            if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                            addlist = addlist->next;
                        }
                        pthread_mutex_unlock(&crackedmutex);
                        if (flag == 0)
                        {
                            add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                        }
    		    }
    		}
    		else if (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>=320)
    		{
    		    mylist = hash_list;
    		    while (mylist)
    		    {
            		if (memcmp(mylist->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
            		{
                    	    int flag = 0;
                    	    plain[0]=markov_charset[c];
                    	    plain[1]=markov_charset[b];
                    	    plain[2]=markov_charset[a];
                    	    plain[3]=reduced_charset[(a1+(d))%charset_size];
                    	    plain[4]=markov_charset[a2];
                    	    plain[5]=markov_charset[a3];
                    	    plain[6]=markov_charset[a4];
                    	    plain[7]=markov_charset[a5];
                    	    plain[8]=markov_charset[a6];
                    	    plain[9]=markov_charset[a7];
                    	    plain[10]=markov_charset[a8];
			    plain[11]=0;
                    	    pthread_mutex_lock(&crackedmutex);
                    	    addlist = cracked_list;
                    	    while (addlist)
                    	    {
                        	if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                        	addlist = addlist->next;
                    	    }
                    	    pthread_mutex_unlock(&crackedmutex);
                    	    if (flag == 0)
                    	    {
                        	add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    	    }
            		}
            		if (mylist) mylist = mylist->next;
        	    }
        	}
	    }
	    }
	clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
        bzero(found_ind,charset_size*charset_size*charset_size*sizeof(cl_uint));
        clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
	}
    }


    /* Markov, len=12 */
    csize=7;
    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);

    if (markov_max_len>=12)
    if (thread_restore[self].cstart<=8)
    for (a1=thread_restore[self].cpos[0];a1<reduced_size;a1+=(ocl_vector*ocl_threads*devicesnum))
    for (a2=thread_restore[self].cpos[1];a2<charset_size;a2++) if ((markov2[a1][a2] > markov_threshold)||(markov2[a1+1][a2] > markov_threshold)||(markov2[a1+2][a2] > markov_threshold)||(markov2[a1+3][a2] > markov_threshold))
    for (a3=thread_restore[self].cpos[2];a3<charset_size;a3++) if (markov1[a2][a3] > markov_threshold)
    for (a4=thread_restore[self].cpos[3];a4<charset_size;a4++) if (markov1[a3][a4] > markov_threshold)
    for (a5=thread_restore[self].cpos[4];a5<charset_size;a5++) if (markov1[a4][a5] > markov_threshold)
    for (a6=thread_restore[self].cpos[5];a6<charset_size;a6++) if (markov1[a5][a6] > markov_threshold)
    for (a7=thread_restore[self].cpos[6];a7<charset_size;a7++) if (markov1[a6][a7] > markov_threshold)
    for (a8=thread_restore[self].cpos[7];a8<charset_size;a8++) if (markov1[a7][a8] > markov_threshold)
    for (a9=thread_restore[self].cpos[8];a9<charset_size;a9++) if (markov1[a8][a9] > markov_threshold)
    {
	if (attack_over!=0) return hash_ok;
	cbase.s0 = reduced_charset[a1];
	cbase.s1 = reduced_charset[a1+1];
	cbase.s2 = reduced_charset[a1+2];
	cbase.s3 = reduced_charset[a1+3];
	cbase.s4 = reduced_charset[a1+4];
	cbase.s5 = reduced_charset[a1+5];
	cbase.s6 = reduced_charset[a1+6];
	cbase.s7 = reduced_charset[a1+7];

	thread_save[self].cpos[0]=a1;
	thread_save[self].cpos[1]=a2;
	thread_save[self].cpos[2]=a3;
	thread_save[self].cpos[3]=a4;
	thread_save[self].cpos[4]=a5;
	thread_save[self].cpos[5]=a6;
	thread_save[self].cpos[6]=a7;
	thread_save[self].cpos[7]=a8;
	thread_save[self].cpos[8]=a9;
	thread_save[self].cstr[0]=bruteforce_charset[a1];
	thread_save[self].cstr[1]=bruteforce_charset[a2];
	thread_save[self].cstr[2]=bruteforce_charset[a3];
	thread_save[self].cstr[3]=bruteforce_charset[a4];
	thread_save[self].cstr[4]=bruteforce_charset[a5];
	thread_save[self].cstr[5]=bruteforce_charset[a6];
	thread_save[self].cstr[6]=bruteforce_charset[a7];
	thread_save[self].cstr[7]=bruteforce_charset[a8];
	thread_save[self].cstr[8]=bruteforce_charset[a9];
	thread_save[self].cstr[9]=0;
	thread_save[self].cstart=8;

	clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
	image.x=0;
	image.y=markov_charset[a2]|(markov_charset[a3]<<8)|(markov_charset[a4]<<16)|(markov_charset[a5]<<24);
	image.z=markov_charset[a6]|(markov_charset[a7]<<8)|(markov_charset[a8]<<16)|(markov_charset[a9]<<24);image.w=0x80;
	REV(image.y);
	REV(image.z);
	REV(image.w);
	clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);
        if (found==1)
        {
    	    found = 0;
    	    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    	}
        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
        //clFlush(queue);
	invocations++;
	if (err != CL_SUCCESS)
	{
	    elog("clEnqueueNDRangeKernel error (%d)\n",err);
	    return hash_err;
	}
	//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
        int *fnd;
        fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
        found = *fnd;

        if (session_restore_lflag==1) ocl_markov_init_restore(self);
	session_restore_lflag = 0;

	if (found==1) 
	{
	    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*sizeof(uint), found_ind, 0, NULL, NULL);
	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    e=-1;

	    for (a=0;a<charset_size;a++)
	    for (b=0;b<charset_size;b++)
	    for (c=0;c<charset_size;c++)
	    if (found_ind[a*charset_size*charset_size+b*charset_size+c]==1)
	    {
	    e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector;
    	    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, e*32, 32*ocl_vector, ptr+e*32, 0, NULL, NULL);
	    for (d=0;d<ocl_vector;d++)
	    {
		e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector+d;
    		if ( (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count<320) && (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>0))
    		{
    		    int i;
    		    for (i=0;i<hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count;i++)
    		    if (memcmp(hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i]->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
    		    {
                        mylist = hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i];
                        int flag = 0;
                        plain[0]=markov_charset[c];
                        plain[1]=markov_charset[b];
                        plain[2]=markov_charset[a];
                        plain[3]=reduced_charset[(a1+(d))%charset_size];
                        plain[4]=markov_charset[a2];
                        plain[5]=markov_charset[a3];
                        plain[6]=markov_charset[a4];
                        plain[7]=markov_charset[a5];
                        plain[8]=markov_charset[a6];
                        plain[9]=markov_charset[a7];
                        plain[10]=markov_charset[a8];
                        plain[11]=markov_charset[a9];
			plain[12]=0;
                        pthread_mutex_lock(&crackedmutex);
                        addlist = cracked_list;
                        while (addlist)
                        {
                            if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                            addlist = addlist->next;
                        }
                        pthread_mutex_unlock(&crackedmutex);
                        if (flag == 0)
                        {
                            add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                        }
    		    }
    		}
    		else if (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>=320)
    		{
    		    mylist = hash_list;
    		    while (mylist)
    		    {
            		if (memcmp(mylist->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
            		{
                    	    int flag = 0;
                    	    plain[0]=markov_charset[c];
                    	    plain[1]=markov_charset[b];
                    	    plain[2]=markov_charset[a];
                    	    plain[3]=reduced_charset[(a1+(d))%charset_size];
                    	    plain[4]=markov_charset[a2];
                    	    plain[5]=markov_charset[a3];
                    	    plain[6]=markov_charset[a4];
                    	    plain[7]=markov_charset[a5];
                    	    plain[8]=markov_charset[a6];
                    	    plain[9]=markov_charset[a7];
                    	    plain[10]=markov_charset[a8];
                    	    plain[11]=markov_charset[a9];
			    plain[12]=0;
                    	    pthread_mutex_lock(&crackedmutex);
                    	    addlist = cracked_list;
                    	    while (addlist)
                    	    {
                        	if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                        	addlist = addlist->next;
                    	    }
                    	    pthread_mutex_unlock(&crackedmutex);
                    	    if (flag == 0)
                    	    {
                        	add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    	    }
            		}
            		if (mylist) mylist = mylist->next;
        	    }
        	}
	    }
	    }
	clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
        bzero(found_ind,charset_size*charset_size*charset_size*sizeof(cl_uint));
        clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
	}
    }

    return hash_ok;
}


/* Worker thread - hybrid attack */
hash_stat ocl_hybrid_sha512_thread(void *arg)
{
    int err;
    cl_command_queue queue;
    cl_mem buffer;
    size_t global_work_size, global_work_size1=1;
    cl_uint4 image;
    int self;
    cl_kernel kernel;
    unsigned char hex1[16];
    char plain[64];
    cl_mem bitmaps_buf;
    int *bitmaps;
    int a,b,c,d,e;
    int a1,a2,a3,a4,a5,a6,a7,a8,a9,a10;
    char *ptr = malloc(strlen(bruteforce_charset)*strlen(bruteforce_charset)*strlen(bruteforce_charset)*32*ocl_vector);
    int charset_size = (int)strlen(bruteforce_charset);
    cl_mem found_ind_buf;
    cl_uint *found_ind;
    int found;
    cl_mem found_buf;
    struct  hash_list_s  *mylist, *addlist;
    unsigned int csize=1;
    cl_mem table_buf;
    unsigned int *table;
    cl_uint8 cbase;
    cl_uint4 singlehash;
    FILE *dictfile;
    int dictpos = -1;
    char *readline = alloca(HASHFILE_MAX_LINE_LENGTH);
    
    
    
    dictfile = fopen(hash_dictionary_file,"r");
    if ((!dictfile) && (self == 0))
    {
	elog("Cannot open dictionary file: %s\n",hash_dictionary_file);
	return hash_err;
    }
    if (session_restore_flag==1)
    {
	for (a=0;a<attack_current_count+self;a++) fgets(readline, HASHFILE_MAX_LINE_LENGTH, dictfile);
    }

    /* Thread initialization */
    pthread_mutex_lock(&biglock);
    kernel = clCreateKernel(program[0], "sha512", &err );
    if (err != CL_SUCCESS)
    {
	elog("clCreateKernel error (%d)\n",err);
	return hash_err;
    }
    memcpy(&self,arg,sizeof(int));
    queue = clCreateCommandQueue(context[0], device[self/ocl_threads], 0, &err );
    if (err != CL_SUCCESS)
    {
        elog("clCreateCommandQueue error (%d)\n",err);
	return hash_err;
    }
    buffer = clCreateBuffer(context[0], CL_MEM_WRITE_ONLY, charset_size*charset_size*charset_size*32*ocl_vector, NULL, &err );
    if (err != CL_SUCCESS)
    {
	elog("clCreateBuffer error (%d)\n",err);
	return hash_err;
    }
    found_buf = clCreateBuffer(context[0], CL_MEM_WRITE_ONLY, 4, NULL, &err );
    if (err != CL_SUCCESS)
    {
	elog("clCreateBuffer error (%d)\n",err);
	return hash_err;
    }
    if (err != CL_SUCCESS)
    {
	elog("clCreateBuffer error (%d)\n",err);
	return hash_err;
    }
    table = malloc(sizeof(uint)*charset_size*charset_size*charset_size);
    table_buf = clCreateBuffer(context[0], CL_MEM_READ_ONLY, sizeof(uint)*charset_size*charset_size*charset_size, NULL, &err );
    bitmaps_buf = clCreateBuffer(context[0], CL_MEM_READ_ONLY, 256*256*32*4, NULL, &err );
    found_ind = NULL;
    bitmaps = NULL;
    
    pthread_mutex_unlock(&biglock); 


    /* Good. Now let's do some work */
    while (!feof(dictfile))
    {
	
	for (a=0;a<ocl_threads*devicesnum;a++) fgets(readline, HASHFILE_MAX_LINE_LENGTH, dictfile);
	readline[HASHFILE_MAX_LINE_LENGTH-1] = 0;
	/* Limit the dictionary word to 12 characters */
	readline[strlen(readline)-1]=0;
	if (strlen(readline)>11) readline[11] = 0;
	/* this line is our thread's job */
	if (((dictpos % (devicesnum*ocl_threads))==self)||(dictpos==(self-1)))
	{
	    attack_current_count++;
	    /* setup table */
	    for (a=0;a<charset_size*charset_size*charset_size;a++)
	    {
		unsigned char x,y,z,t1;
		t1 = (a / (charset_size*charset_size));
		x = (bruteforce_charset[(a / (charset_size*charset_size))]);
		y = (bruteforce_charset[((a-t1*(charset_size*charset_size))/charset_size)]);
		z = (bruteforce_charset[(a % charset_size)]);
		table[a] = (x<<16)|(y<<8)|(z);
	    }


	    if (!bitmaps)
	    {
		bitmaps = malloc(256*256*32*4);
		mylist = hash_list;
		a=0;
		while (mylist)
		{
    		    memcpy(hex1,mylist->hash,16);
		    a++;
		    unsigned int *b_a = (unsigned int *)hex1;
		    unsigned int *b_b = (unsigned int *)&hex1[4];
		    unsigned int *b_c = (unsigned int *)&hex1[8];
		    unsigned int *b_d = (unsigned int *)&hex1[12];
		    unsigned int bind_a = (*b_a)>>13;
		    unsigned int bval_a = (1<<((*b_a)&31));
		    unsigned int bind_b = (*b_b)>>13;
		    unsigned int bval_b = (1<<((*b_b)&31));
		    unsigned int bind_c = (*b_c)>>13;
		    unsigned int bval_c = (1<<((*b_c)&31));
		    unsigned int bind_d = (*b_d)>>13;
		    unsigned int bval_d = (1<<((*b_d)&31));
		    bitmaps[bind_a] |=bval_a;
		    bitmaps[bind_b+65535*8] |=bval_b;
		    bitmaps[bind_c+65535*16] |=bval_c;
		    bitmaps[bind_d+65535*24] |=bval_d;
	    	    singlehash.x |= (1<<((*b_b)&31));
		    singlehash.y |= (1<<((*b_c)&31));
		    singlehash.z |= (1<<((*b_d)&31));
    		    if (mylist) mylist = mylist->next;
		}
		if (a==1)
		{
		    mylist = hash_list;
		    memcpy(hex1,mylist->hash,4);
		    memcpy(&singlehash.x, hex1, 4);
		    memcpy(hex1,mylist->hash+4,4);
		    memcpy(&singlehash.y, hex1, 4);
		    memcpy(hex1,mylist->hash+8,4);
		    memcpy(&singlehash.z, hex1, 4);
		    memcpy(hex1,mylist->hash+12,4);
		    memcpy(&singlehash.w, hex1, 4);
	
		    clSetKernelArg(kernel, 8, sizeof(cl_uint4), (void*) &singlehash);
		}
		else 
		{
		    singlehash.w=0;
		    clSetKernelArg(kernel, 8, sizeof(cl_uint4), (void*) &singlehash);
		}
	    }
	    
	    if (!found_ind) 
	    {
		found_ind=malloc(charset_size*charset_size*charset_size*sizeof(cl_uint));
		bzero(found_ind,sizeof(uint)*charset_size*charset_size*charset_size);
		found_ind_buf = clCreateBuffer(context[0], CL_MEM_WRITE_ONLY, charset_size*charset_size*charset_size*sizeof(cl_uint), NULL, &err );
		clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
	    }

	    clSetKernelArg(kernel, 0, sizeof(cl_mem), (void*) &buffer);
	    clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);
	    clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
	    clSetKernelArg(kernel, 4, sizeof(cl_mem), (void*) &found_ind_buf);
	    clSetKernelArg(kernel, 5, sizeof(cl_mem), (void*) &bitmaps_buf);
	    clSetKernelArg(kernel, 6, sizeof(cl_mem), (void*) &found_buf);
	    clSetKernelArg(kernel, 7, sizeof(cl_mem), (void*) &table_buf);
	    global_work_size = charset_size*charset_size*charset_size;
	    while ((global_work_size %256)!=0) global_work_size++;

	    if (err != CL_SUCCESS)
	    {
		elog("clEnqueueBuffer error (%d)\n",err);
		return hash_err;
	    }
	    /* dictionary word -> image */
	    switch (strlen(readline))
	    {
		case 1:
		    image.x=readline[0];
		    image.y=image.z=image.w=0;
		    break;
		case 2:
		    image.x=readline[0]|(readline[1]<<8);
		    image.y=image.z=image.w=0;
		    break;
		case 3:
		    image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16);
		    image.y=image.z=image.w=0;
		    break;
		case 4:
		    image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
		    image.y=image.z=image.w=0;
		    break;
		case 5:
		    image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
		    image.y=readline[4];
		    image.z=image.w=0;
		    break;
		case 6:
		    image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
		    image.y=readline[4]|(readline[5]<<8);
		    image.z=image.w=0;
		    break;
		case 7:
		    image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
		    image.y=readline[4]|(readline[5]<<8)|(readline[6]<<16);
		    image.z=image.w=0;
		    break;
		case 8:
		    image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
		    image.y=readline[4]|(readline[5]<<8)|(readline[6]<<16)|(readline[7]<<24);
		    image.z=image.w=0;
		    break;
		case 9:
		    image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
		    image.y=readline[4]|(readline[5]<<8)|(readline[6]<<16)|(readline[7]<<24);
		    image.z=readline[8];
		    image.w=0;
		    break;
		case 10:
		    image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
		    image.y=readline[4]|(readline[5]<<8)|(readline[6]<<16)|(readline[7]<<24);
		    image.z=readline[8]|(readline[9]<<8);
		    image.w=0;
		    break;
		case 11:
		    image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
		    image.y=readline[4]|(readline[5]<<8)|(readline[6]<<16)|(readline[7]<<24);
		    image.z=readline[8]|(readline[9]<<8)|(readline[10]<<16);
		    image.w=0;
		    break;
		case 12:
		    image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
		    image.y=readline[4]|(readline[5]<<8)|(readline[6]<<16)|(readline[7]<<24);
		    image.z=readline[8]|(readline[9]<<8)|(readline[10]<<16)|(readline[11]<<24);
		    image.w=0;
		    break;
		default:
		    image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
		    image.y=readline[4]|(readline[5]<<8)|(readline[6]<<16)|(readline[7]<<24);
		    image.z=readline[8]|(readline[9]<<8)|(readline[10]<<16)|(readline[11]<<24);
		    image.w=0;
		    break;
	    }
	    clEnqueueWriteBuffer(queue, table_buf, CL_TRUE, 0, sizeof(uint)*charset_size*charset_size*charset_size, table, 0, NULL, NULL);
	    clEnqueueWriteBuffer(queue, bitmaps_buf, CL_TRUE, 0, 256*256*32*4 , bitmaps, 0, NULL, NULL);

	    /* hybrid, len=0 (aka dictionary) */
	    for (a=0;a<charset_size*charset_size*charset_size;a++)
	    {
		unsigned char x,y,z,t1;
		t1 = (a / (charset_size*charset_size));
		x = readline[strlen(readline)-1];
		y = readline[strlen(readline)-2];
		z = readline[strlen(readline)-3];
		table[a] = (x<<16)|(y<<8)|(z);
	    }
	    clEnqueueWriteBuffer(queue, table_buf, CL_TRUE, 0, sizeof(uint)*charset_size*charset_size*charset_size, table, 0, NULL, NULL);
	    clSetKernelArg(kernel, 7, sizeof(cl_mem), (void*) &table_buf);

	    csize=strlen(readline)-3;
	    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);
	    clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);
	    if ((bruteforce_start<=0) && (bruteforce_end>=1) && ((strlen(readline)+1)<16))
	    for (a1=0;a1<1;a1++)
	    {
		if (attack_over!=0) return hash_ok;
		if (strlen(readline)>2) { cbase.s0 = cbase.s1 = cbase.s2 = cbase.s3 = cbase.s4 = cbase.s5 = cbase.s6 = cbase.s7 = readline[strlen(readline)-4];}
		else {  cbase.s0 = cbase.s1 = cbase.s2 = cbase.s3 = cbase.s4 = cbase.s5 = cbase.s6 = cbase.s7  = readline[0];}
		
		clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
    		if (found==1)
    		{
    		    found = 0;
    		    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    		}
	        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size1, NULL, 0, NULL, NULL);
	        //clFlush(queue);
		invocations++;
		if (err != CL_SUCCESS)
		{
		    elog("clEnqueueNDRangeKernel error (%d)\n",err);
		    return hash_err;
		}
		//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
    		int *fnd;
    		fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
    		found = *fnd;

		if (found==1) 
		{
		    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*sizeof(uint), found_ind, 0, NULL, NULL);
		    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
		    e=-1;
		    for (a=0;a<charset_size;a++)
		    for (b=0;b<charset_size;b++)
		    for (c=0;c<charset_size;c++)
		    if (found_ind[a*charset_size*charset_size+b*charset_size+c]==1)
		    {
		    e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector;
    		    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, e*32, 32*ocl_vector, ptr+e*32, 0, NULL, NULL);
		    for (d=0;d<ocl_vector;d++)
		    {
	    		e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector+d;
    			if ( (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count<320) && (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>0))
    			{
    			    int i;
    			    for (i=0;i<hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count;i++)
    			    if (memcmp(hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i]->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
    			    {
                    		mylist = hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i];
                    		int flag = 0;
                    		strcpy(plain,readline);
                    		pthread_mutex_lock(&crackedmutex);
                    		addlist = cracked_list;
                    		while (addlist)
                    		{
                        	    if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                	flag = 1;
                        	    addlist = addlist->next;
                    		}
                    		pthread_mutex_unlock(&crackedmutex);
                    		if (flag == 0)
                    		{
                        	    add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    		}
    			    }
    			}
    			else if (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>=320)
    			{
    			    mylist = hash_list;
    			    while (mylist)
    			    {
            			if (memcmp(mylist->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
            			{
                    		    int flag = 0;
                    		    strcpy(plain,readline);
                    		    pthread_mutex_lock(&crackedmutex);
                    		    addlist = cracked_list;
                    		    while (addlist)
                    		    {
                        		if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                	flag = 1;
                        		addlist = addlist->next;
                    		    }
                    		    pthread_mutex_unlock(&crackedmutex);
                    		    if (flag == 0)
                    		    {
                        		add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    		    }
            			}
            			if (mylist) mylist = mylist->next;
        		    }
        		}
		    }
		    }
		clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
    		bzero(found_ind,charset_size*charset_size*charset_size*sizeof(cl_uint));
    		clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
		}
	    }




	    /* hybrid, len=1 */
	    for (a=0;a<charset_size*charset_size*charset_size;a++)
	    {
		unsigned char x,y,z,t1;
		t1 = (a / (charset_size*charset_size));
		x = (bruteforce_charset[(a / (charset_size*charset_size))]);
		y = readline[strlen(readline)-1];
		z = readline[strlen(readline)-2];
		table[a] = (x<<16)|(y<<8)|(z);
	    }
	    clEnqueueWriteBuffer(queue, table_buf, CL_TRUE, 0, sizeof(uint)*charset_size*charset_size*charset_size, table, 0, NULL, NULL);
	    clSetKernelArg(kernel, 7, sizeof(cl_mem), (void*) &table_buf);

	    csize=strlen(readline)-2;
	    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);
	    clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);
	    if ((bruteforce_start<=1) && (bruteforce_end>=1) && ((strlen(readline)+1)<16))
	    for (a1=0;a1<1;a1++)
	    {
		if (attack_over!=0) return hash_ok;
		if (strlen(readline)>2) { cbase.s0 = cbase.s1 = cbase.s2 = cbase.s3 = cbase.s4 = cbase.s5 = cbase.s6 = cbase.s7 = readline[strlen(readline)-3];}
		else { cbase.s0 = cbase.s1 = cbase.s2 = cbase.s3 = cbase.s4 = cbase.s5 = cbase.s6 = cbase.s7 = readline[0];}
		
		clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
    		if (found==1)
    		{
    		    found = 0;
    		    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    		}
	        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
	        //clFlush(queue);
		invocations++;
		if (err != CL_SUCCESS)
		{
		    elog("clEnqueueNDRangeKernel error (%d)\n",err);
		    return hash_err;
		}
		//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
    		int *fnd;
    		fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
    		found = *fnd;

		if (found==1) 
		{
		    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*sizeof(uint), found_ind, 0, NULL, NULL);
		    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
		    e=-1;
		    for (a=0;a<charset_size;a++)
		    for (b=0;b<charset_size;b++)
		    for (c=0;c<charset_size;c++)
		    if (found_ind[a*charset_size*charset_size+b*charset_size+c]==1)
		    {
		    e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector;
    		    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, e*32, 32*ocl_vector, ptr+e*32, 0, NULL, NULL);
		    for (d=0;d<ocl_vector;d++)
		    {
	    		e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector+d;
    			if ( (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count<320) && (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>0))
    			{
    			    int i;
    			    for (i=0;i<hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count;i++)
    			    if (memcmp(hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i]->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
    			    {
                    		mylist = hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i];
                    		int flag = 0;
                    		strcpy(plain,readline);
                    		plain[strlen(readline)]=bruteforce_charset[a];
                    		plain[strlen(readline)+1]=0;
                    		pthread_mutex_lock(&crackedmutex);
                    		addlist = cracked_list;
                    		while (addlist)
                    		{
                        	    if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
                            	    && (strcmp(addlist->salt, mylist->salt) == 0))
                                	flag = 1;
                        	    addlist = addlist->next;
                    		}
                    		pthread_mutex_unlock(&crackedmutex);
                    		if (flag == 0)
                    		{
                        	    add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    		}
    			    }
    			}
    			else if (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>=320)
    			{
    			    mylist = hash_list;
    			    while (mylist)
    			    {
            			if (memcmp(mylist->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
            			{
                    		    int flag = 0;
                    		    strcpy(plain,readline);
                    		    plain[strlen(readline)]=bruteforce_charset[a];
                    		    plain[strlen(readline)+1]=0;
                    		    pthread_mutex_lock(&crackedmutex);
                    		    addlist = cracked_list;
                    		    while (addlist)
                    		    {
                        		if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
                            		&& (strcmp(addlist->salt, mylist->salt) == 0))
                                	flag = 1;
                        		addlist = addlist->next;
                    		    }
                    		    pthread_mutex_unlock(&crackedmutex);
                    		    if (flag == 0)
                    		    {
                        		add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    		    }
            			}
            			if (mylist) mylist = mylist->next;
        		    }
        		}
		    }
		    }
		clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
    		bzero(found_ind,charset_size*charset_size*charset_size*sizeof(cl_uint));
    		clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
		}
	    }


	    /* hybrid, len=2 */

	    for (a=0;a<charset_size*charset_size*charset_size;a++)
	    {
		unsigned char x,y,z,t1;
		t1 = (a / (charset_size*charset_size));
		x = (bruteforce_charset[(a / (charset_size*charset_size))]);
		y = (bruteforce_charset[((a-t1*(charset_size*charset_size))/charset_size)]);
		z = (bruteforce_charset[(a % charset_size)]);
		z = readline[strlen(readline)-1];
		table[a] = (x<<16)|(y<<8)|(z);
	    }
	    clEnqueueWriteBuffer(queue, table_buf, CL_TRUE, 0, sizeof(uint)*charset_size*charset_size*charset_size, table, 0, NULL, NULL);
	    clSetKernelArg(kernel, 7, sizeof(cl_mem), (void*) &table_buf);

	    csize=strlen(readline)-1;
	    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);
	    clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);
	    if ((bruteforce_start<=2) && (bruteforce_end>=2) && ((strlen(readline)+2)<16))
	    for (a1=0;a1<charset_size;a1++)
	    {
		if (attack_over!=0) return hash_ok;
		if (strlen(readline)>2) { cbase.s0 = cbase.s1 = cbase.s2 = cbase.s3 = cbase.s4 = cbase.s5 = cbase.s6 = cbase.s7 = readline[strlen(readline)-2];}
		else { cbase.s0 = cbase.s1 = cbase.s2 = cbase.s3 = cbase.s4 = cbase.s5 = cbase.s6 = cbase.s7 = readline[0];}
		
		clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
    		if (found==1)
    		{
    		    found = 0;
    		    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    		}
	        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
	        //clFlush(queue);
		invocations++;
		if (err != CL_SUCCESS)
		{
		    elog("clEnqueueNDRangeKernel error (%d)\n",err);
		    return hash_err;
		}
		//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
    		int *fnd;
    		fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
    		found = *fnd;

		if (found==1) 
		{
		    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*sizeof(uint), found_ind, 0, NULL, NULL);
		    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
		    e=-1;
		    for (a=0;a<charset_size;a++)
		    for (b=0;b<charset_size;b++)
		    for (c=0;c<charset_size;c++)
		    if (found_ind[a*charset_size*charset_size+b*charset_size+c]==1)
		    {
		    e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector;
    		    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, e*32, 32*ocl_vector, ptr+e*32, 0, NULL, NULL);
		    for (d=0;d<ocl_vector;d++)
		    {
	    		e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector+d;
    			if ( (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count<320) && (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>0))
    			{
    			    int i;
    			    for (i=0;i<hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count;i++)
    			    if (memcmp(hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i]->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
    			    {
                    		mylist = hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i];
                    		int flag = 0;
                    		strcpy(plain,readline);
                    		plain[strlen(readline)]=bruteforce_charset[b];
                    		plain[strlen(readline)+1]=bruteforce_charset[a];
                    		plain[strlen(readline)+2]=0;
                    		pthread_mutex_lock(&crackedmutex);
                    		addlist = cracked_list;
                    		while (addlist)
                    		{
                        	    if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
                            	    && (strcmp(addlist->salt, mylist->salt) == 0))
                                	flag = 1;
                        	    addlist = addlist->next;
                    		}
                    		pthread_mutex_unlock(&crackedmutex);
                    		if (flag == 0)
                    		{
                        	    add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    		}
    			    }
    			}
    			else if (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>=320)
    			{
    			    mylist = hash_list;
    			    while (mylist)
    			    {
            			if (memcmp(mylist->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
            			{
                    		    int flag = 0;
                    		    strcpy(plain,readline);
                    		    plain[strlen(readline)]=bruteforce_charset[b];
                    		    plain[strlen(readline)+1]=bruteforce_charset[a];
                    		    plain[strlen(readline)+2]=0;
                    		    pthread_mutex_lock(&crackedmutex);
                    		    addlist = cracked_list;
                    		    while (addlist)
                    		    {
                        		if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
                            		&& (strcmp(addlist->salt, mylist->salt) == 0))
                                	flag = 1;
                        		addlist = addlist->next;
                    		    }
                    		    pthread_mutex_unlock(&crackedmutex);
                    		    if (flag == 0)
                    		    {
                        		add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    		    }
            			}
            			if (mylist) mylist = mylist->next;
        		    }
        		}
		    }
		    }
		clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
    		bzero(found_ind,charset_size*charset_size*charset_size*sizeof(cl_uint));
    		clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
		}
	    }



	    /* setup table - for the last time*/
	    for (a=0;a<charset_size*charset_size*charset_size;a++)
	    {
		unsigned char x,y,z,t1;
		t1 = (a / (charset_size*charset_size));
		x = (bruteforce_charset[(a / (charset_size*charset_size))]);
		y = (bruteforce_charset[((a-t1*(charset_size*charset_size))/charset_size)]);
		z = (bruteforce_charset[(a % charset_size)]);
		table[a] = (x<<16)|(y<<8)|(z);
	    }
	    clEnqueueWriteBuffer(queue, table_buf, CL_TRUE, 0, sizeof(uint)*charset_size*charset_size*charset_size, table, 0, NULL, NULL);
	    clSetKernelArg(kernel, 7, sizeof(cl_mem), (void*) &table_buf);

	    /* Hybrid, len=3 */
	    csize=0+strlen(readline);
	    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);
	    clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);
	    if ((bruteforce_start<=3) && (bruteforce_end>=3) && ((strlen(readline)+3)<16))
	    for (a1=0;a1<charset_size;a1++)
	    {
		if (attack_over!=0) return hash_ok;
		cbase.s0 = cbase.s1 = cbase.s2 = cbase.s3 = cbase.s4 = cbase.s5 = cbase.s6 = cbase.s7 = readline[strlen(readline)-3];
		
		clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
    		if (found==1)
    		{
    		    found = 0;
    		    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    		}
	        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
	        //clFlush(queue);
		invocations++;
		if (err != CL_SUCCESS)
		{
		    elog("clEnqueueNDRangeKernel error (%d)\n",err);
		    return hash_err;
		}
		//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
    		int *fnd;
    		fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
    		found = *fnd;

		if (found==1) 
		{
		    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*sizeof(uint), found_ind, 0, NULL, NULL);
		    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
		    e=-1;
		    for (a=0;a<charset_size;a++)
		    for (b=0;b<charset_size;b++)
		    for (c=0;c<charset_size;c++)
		    if (found_ind[a*charset_size*charset_size+b*charset_size+c]==1)
		    {
		    e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector;
    		    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, e*32, 32*ocl_vector, ptr+e*32, 0, NULL, NULL);
		    for (d=0;d<ocl_vector;d++)
		    {
	    		e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector+d;
    			if ( (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count<320) && (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>0))
    			{
    			    int i;
    			    for (i=0;i<hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count;i++)
    			    if (memcmp(hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i]->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
    			    {
                    		mylist = hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i];
                    		int flag = 0;
                    		strcpy(plain,readline);
                    		plain[strlen(readline)]=bruteforce_charset[(a1+(d))%charset_size];
                    		plain[strlen(readline)+1]=bruteforce_charset[b];
                    		plain[strlen(readline)+2]=bruteforce_charset[a];
                    		plain[strlen(readline)+3]=0;
                    		pthread_mutex_lock(&crackedmutex);
                    		addlist = cracked_list;
                    		while (addlist)
                    		{
                        	    if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
                            	    && (strcmp(addlist->salt, mylist->salt) == 0))
                                	flag = 1;
                        	    addlist = addlist->next;
                    		}
                    		pthread_mutex_unlock(&crackedmutex);
                    		if (flag == 0)
                    		{
                        	    add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    		}
    			    }
    			}
    			else if (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>=320)
    			{
    			    mylist = hash_list;
    			    while (mylist)
    			    {
            			if (memcmp(mylist->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
            			{
                    		    int flag = 0;
                    		    strcpy(plain,readline);
                    		    plain[strlen(readline)]=bruteforce_charset[(a1+(d))%charset_size];
                    		    plain[strlen(readline)+1]=bruteforce_charset[c];
                    		    plain[strlen(readline)+2]=bruteforce_charset[b];
                    		    plain[strlen(readline)+3]=bruteforce_charset[a];
                    		    plain[strlen(readline)+4]=0;
                    		    pthread_mutex_lock(&crackedmutex);
                    		    addlist = cracked_list;
                    		    while (addlist)
                    		    {
                        		if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
                            		&& (strcmp(addlist->salt, mylist->salt) == 0))
                                	flag = 1;
                        		addlist = addlist->next;
                    		    }
                    		    pthread_mutex_unlock(&crackedmutex);
                    		    if (flag == 0)
                    		    {
                        		add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    		    }
            			}
            			if (mylist) mylist = mylist->next;
        		    }
        		}
		    }
		    }
		clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
    		bzero(found_ind,charset_size*charset_size*charset_size*sizeof(cl_uint));
    		clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
		}
	    }


	    /* Hybrid, len=4 */
	    csize=1+strlen(readline);
	    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);
	    clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);
	    if ((bruteforce_start<=4) && (bruteforce_end>=4) && ((strlen(readline)+4)<16))
	    for (a1=0;a1<charset_size;a1++)
	    {
		if (attack_over!=0) return hash_ok;

		cbase.s0 = bruteforce_charset[a1];
		cbase.s1 = bruteforce_charset[a1+1];
		cbase.s2 = bruteforce_charset[a1+2];
		cbase.s3 = bruteforce_charset[a1+3];
		cbase.s4 = bruteforce_charset[a1+4];
		cbase.s5 = bruteforce_charset[a1+5];
		cbase.s6 = bruteforce_charset[a1+6];
		cbase.s7 = bruteforce_charset[a1+7];
		clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
    		if (found==1)
    		{
    		    found = 0;
    		    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    		}
	        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
	        //clFlush(queue);
		invocations++;
		if (err != CL_SUCCESS)
		{
		    elog("clEnqueueNDRangeKernel error (%d)\n",err);
		    return hash_err;
		}
		//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
    		int *fnd;
    		fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
    		found = *fnd;

		if (found==1) 
		{
		    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, charset_size*charset_size*charset_size*sizeof(uint), found_ind, 0, NULL, NULL);
		    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
		    e=-1;
		    for (a=0;a<charset_size;a++)
		    for (b=0;b<charset_size;b++)
		    for (c=0;c<charset_size;c++)
		    if (found_ind[a*charset_size*charset_size+b*charset_size+c]==1)
		    {
		    e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector;
    		    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, e*32, 32*ocl_vector, ptr+e*32, 0, NULL, NULL);
		    for (d=0;d<ocl_vector;d++)
		    {
	    		e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector+d;
    			if ( (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count<320) && (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>0))
    			{
    			    int i;
    			    for (i=0;i<hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count;i++)
    			    if (memcmp(hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i]->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
    			    {
                    		mylist = hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i];
                    		int flag = 0;
                    		strcpy(plain,readline);
                    		plain[strlen(readline)]=bruteforce_charset[(a1+(d))%charset_size];
                    		plain[strlen(readline)+1]=bruteforce_charset[c];
                    		plain[strlen(readline)+2]=bruteforce_charset[b];
                    		plain[strlen(readline)+3]=bruteforce_charset[a];
                    		plain[strlen(readline)+4]=0;
                    		pthread_mutex_lock(&crackedmutex);
                    		addlist = cracked_list;
                    		while (addlist)
                    		{
                        	    if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
                            	    && (strcmp(addlist->salt, mylist->salt) == 0))
                                	flag = 1;
                        	    addlist = addlist->next;
                    		}
                    		pthread_mutex_unlock(&crackedmutex);
                    		if (flag == 0)
                    		{
                        	    add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    		}
    			    }
    			}
    			else if (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>=320)
    			{
    			    mylist = hash_list;
    			    while (mylist)
    			    {
            			if (memcmp(mylist->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
            			{
                    		    int flag = 0;
                    		    strcpy(plain,readline);
                    		    plain[strlen(readline)]=bruteforce_charset[(a1+(d))%charset_size];
                    		    plain[strlen(readline)+1]=bruteforce_charset[c];
                    		    plain[strlen(readline)+2]=bruteforce_charset[b];
                    		    plain[strlen(readline)+3]=bruteforce_charset[a];
                    		    plain[strlen(readline)+4]=0;
                    		    pthread_mutex_lock(&crackedmutex);
                    		    addlist = cracked_list;
                    		    while (addlist)
                    		    {
                        		if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
                            		&& (strcmp(addlist->salt, mylist->salt) == 0))
                                	flag = 1;
                        		addlist = addlist->next;
                    		    }
                    		    pthread_mutex_unlock(&crackedmutex);
                    		    if (flag == 0)
                    		    {
                        		add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    		    }
            			}
            			if (mylist) mylist = mylist->next;
        		    }
        		}
		    }
		    }
		clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
    		bzero(found_ind,charset_size*charset_size*charset_size*sizeof(cl_uint));
    		clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
		}
	    }

	    /* Hybrid, len=5 */
	    csize=2+strlen(readline);
	    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);
	    clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);
	    if ((bruteforce_start<=5) && (bruteforce_end>=5) && ((strlen(readline)+5)<16))
	    for (a1=0;a1<charset_size;a1++)
	    for (a2=0;a2<charset_size;a2++)
	    {
		if (attack_over!=0) return hash_ok;

		cbase.s0 = bruteforce_charset[a2];
		cbase.s1 = bruteforce_charset[a2+1];
		cbase.s2 = bruteforce_charset[a2+2];
		cbase.s3 = bruteforce_charset[a2+3];
		cbase.s4 = bruteforce_charset[a2+4];
		cbase.s5 = bruteforce_charset[a2+5];
		cbase.s6 = bruteforce_charset[a2+6];
		cbase.s7 = bruteforce_charset[a2+7];

		switch (strlen(readline))
		{
		    case 1:
			image.x=readline[0]|(bruteforce_charset[a1]<<8);
			image.y=image.z=image.w=0;
			break;
		    case 2:
			image.x=readline[0]|(readline[1]<<8)|(bruteforce_charset[a1]<<16);
			image.y=image.z=image.w=0;
			break;
		    case 3:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(bruteforce_charset[a1]<<24);
			image.y=image.z=image.w=0;
			break;
		    case 4:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=image.z=image.w=0;
			image.y=bruteforce_charset[a1];
			break;
		    case 5:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=readline[4]|(bruteforce_charset[a1]<<8);
			image.z=image.w=0;
			break;
		    case 6:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=readline[4]|(readline[5]<<8)|(bruteforce_charset[a1]<<16);
			image.z=image.w=0;
			break;
		    case 7:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=readline[4]|(readline[5]<<8)|(readline[6]<<16)|(bruteforce_charset[a1]<<24);
			image.z=image.w=0;
			break;
		    case 8:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=readline[4]|(readline[5]<<8)|(readline[6]<<16)|(readline[7]<<24);
			image.z=image.w=0;
			image.z=bruteforce_charset[a1];
			break;
		    case 9:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=readline[4]|(readline[5]<<8)|(readline[6]<<16)|(readline[7]<<24);
			image.z=readline[8]|(bruteforce_charset[a1]<<8);
			image.w=0;
			break;
		    case 10:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=readline[4]|(readline[5]<<8)|(readline[6]<<16)|(readline[7]<<24);
			image.z=readline[8]|(readline[9]<<8)|(bruteforce_charset[a1]<<16);
			image.w=0;
			break;
		    case 11:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=readline[4]|(readline[5]<<8)|(readline[6]<<16)|(readline[7]<<24);
			image.z=readline[8]|(readline[9]<<8)|(readline[10]<<16)|(bruteforce_charset[a1]<<24);
			image.w=0;
			break;
		    default:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=readline[4]|(readline[5]<<8)|(readline[6]<<16)|(readline[7]<<24);
			image.z=readline[8]|(readline[9]<<8)|(readline[10]<<16)|(readline[11]<<24);
			image.w=0;
			break;
		}
		clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);

		clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
    		if (found==1)
    		{
    		    found = 0;
    		    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    		}
	        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
	        //clFlush(queue);
		invocations++;
		if (err != CL_SUCCESS)
		{
		    elog("clEnqueueNDRangeKernel error (%d)\n",err);
		    return hash_err;
		}
		//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
    		int *fnd;
    		fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
    		found = *fnd;

		if (found==1) 
		{
		    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, 0, charset_size*charset_size*charset_size*32*ocl_vector, ptr, 0, NULL, NULL);
		    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
		    e=-1;
		    for (a=0;a<charset_size;a++)
		    for (b=0;b<charset_size;b++)
		    for (c=0;c<charset_size;c++)
		    if (found_ind[a*charset_size*charset_size+b*charset_size+c]==1)
		    {
		    e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector;
    		    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, e*32, 32*ocl_vector, ptr+e*32, 0, NULL, NULL);
		    for (d=0;d<ocl_vector;d++)
		    {
	    		e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector+d;
    			if ( (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count<320) && (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>0))
    			{
    			    int i;
    			    for (i=0;i<hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count;i++)
    			    if (memcmp(hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i]->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
    			    {
                    		mylist = hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i];
                    		int flag = 0;
                    		strcpy(plain,readline);
                    		plain[strlen(readline)]=bruteforce_charset[a1];
                    		plain[strlen(readline)+1]=bruteforce_charset[(a2+(d))%charset_size];
                    		plain[strlen(readline)+2]=bruteforce_charset[c];
                    		plain[strlen(readline)+3]=bruteforce_charset[b];
                    		plain[strlen(readline)+4]=bruteforce_charset[a];
                    		plain[strlen(readline)+5]=0;
                    		pthread_mutex_lock(&crackedmutex);
                    		addlist = cracked_list;
                    		while (addlist)
                    		{
                        	    if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
                            	    && (strcmp(addlist->salt, mylist->salt) == 0))
                                	flag = 1;
                        	    addlist = addlist->next;
                    		}
                    		pthread_mutex_unlock(&crackedmutex);
                    		if (flag == 0)
                    		{
                        	    add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    		}
    			    }
    			}
    			else if (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>=320)
    			{
    			    mylist = hash_list;
    			    while (mylist)
    			    {
            			if (memcmp(mylist->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
            			{
                    		    int flag = 0;
                    		    strcpy(plain,readline);
                    		    plain[strlen(readline)]=bruteforce_charset[(a1+(d))%charset_size];
                    		    plain[strlen(readline)+1]=bruteforce_charset[c];
                    		    plain[strlen(readline)+2]=bruteforce_charset[b];
                    		    plain[strlen(readline)+3]=bruteforce_charset[a];
                    		    plain[strlen(readline)+4]=0;
                    		    pthread_mutex_lock(&crackedmutex);
                    		    addlist = cracked_list;
                    		    while (addlist)
                    		    {
                        		if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
                            		&& (strcmp(addlist->salt, mylist->salt) == 0))
                                	flag = 1;
                        		addlist = addlist->next;
                    		    }
                    		    pthread_mutex_unlock(&crackedmutex);
                    		    if (flag == 0)
                    		    {
                        		add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    		    }
            			}
            			if (mylist) mylist = mylist->next;
        		    }
        		}
		    }
		    }
		clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
    		bzero(found_ind,charset_size*charset_size*charset_size*sizeof(cl_uint));
    		clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
		}
	    }

	    /* Hybrid, len=6 */
	    csize=3+strlen(readline);
	    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);
	    clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);
	    if ((bruteforce_start<=6) && (bruteforce_end>=6) && ((strlen(readline)+6)<16))
	    for (a1=0;a1<charset_size;a1++)
	    for (a2=0;a2<charset_size;a2++)
	    for (a3=0;a3<charset_size;a3++)
	    {
		if (attack_over!=0) return hash_ok;

		cbase.s0 = bruteforce_charset[a3];
		cbase.s1 = bruteforce_charset[a3+1];
		cbase.s2 = bruteforce_charset[a3+2];
		cbase.s3 = bruteforce_charset[a3+3];
		cbase.s4 = bruteforce_charset[a3+4];
		cbase.s5 = bruteforce_charset[a3+5];
		cbase.s6 = bruteforce_charset[a3+6];
		cbase.s7 = bruteforce_charset[a3+7];
		switch (strlen(readline))
		{
		    case 1:
			image.x=readline[0]|(bruteforce_charset[a1]<<8)|(bruteforce_charset[a2]<<16);
			image.y=image.z=image.w=0;
			break;
		    case 2:
			image.x=readline[0]|(readline[1]<<8)|(bruteforce_charset[a1]<<16)|(bruteforce_charset[a2]<<24);
			image.y=image.z=image.w=0;
			break;
		    case 3:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(bruteforce_charset[a1]<<24);
			image.y=(bruteforce_charset[a2]);
			image.z=image.w=0;
			break;
		    case 4:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=image.z=image.w=0;
			image.y=bruteforce_charset[a1]|(bruteforce_charset[a2]<<8);
			break;
		    case 5:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=readline[4]|(bruteforce_charset[a1]<<8)|(bruteforce_charset[a2]<<16);
			image.z=image.w=0;
			break;
		    case 6:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=readline[4]|(readline[5]<<8)|(bruteforce_charset[a1]<<16)|(bruteforce_charset[a2]<<24);
			image.z=image.w=0;
			break;
		    case 7:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=readline[4]|(readline[5]<<8)|(readline[6]<<16)|(bruteforce_charset[a1]<<24);
			image.z=bruteforce_charset[a2];
			image.w=0;
			break;
		    case 8:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=readline[4]|(readline[5]<<8)|(readline[6]<<16)|(readline[7]<<24);
			image.z=image.w=0;
			image.z=bruteforce_charset[a1]|(bruteforce_charset[a2]<<8);
			break;
		    case 9:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=readline[4]|(readline[5]<<8)|(readline[6]<<16)|(readline[7]<<24);
			image.z=readline[8]|(bruteforce_charset[a1]<<8)|(bruteforce_charset[a2]<<16);
			image.w=0;
			break;
		    case 10:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=readline[4]|(readline[5]<<8)|(readline[6]<<16)|(readline[7]<<24);
			image.z=readline[8]|(readline[9]<<8)|(bruteforce_charset[a1]<<16)|(bruteforce_charset[a2]<<24);
			image.w=0;
			break;
		    case 11:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=readline[4]|(readline[5]<<8)|(readline[6]<<16)|(readline[7]<<24);
			image.z=readline[8]|(readline[9]<<8)|(readline[10]<<16)|(bruteforce_charset[a1]<<24);
			image.w=bruteforce_charset[a2];
			break;
		    default:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=readline[4]|(readline[5]<<8)|(readline[6]<<16)|(readline[7]<<24);
			image.z=readline[8]|(readline[9]<<8)|(readline[10]<<16)|(readline[11]<<24);
			image.w=0;
			break;
		}
		clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);

		clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
    		if (found==1)
    		{
    		    found = 0;
    		    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    		}
	        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
	        //clFlush(queue);
		invocations++;
		if (err != CL_SUCCESS)
		{
		    elog("clEnqueueNDRangeKernel error (%d)\n",err);
		    return hash_err;
		}
		//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
    		int *fnd;
    		fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
    		found = *fnd;

		if (found==1) 
		{
		    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, 0, charset_size*charset_size*charset_size*32*ocl_vector, ptr, 0, NULL, NULL);
		    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
		    e=-1;
		    for (a=0;a<charset_size;a++)
		    for (b=0;b<charset_size;b++)
		    for (c=0;c<charset_size;c++)
		    if (found_ind[a*charset_size*charset_size+b*charset_size+c]==1)
		    {
		    e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector;
    		    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, e*32, 32*ocl_vector, ptr+e*32, 0, NULL, NULL);
		    for (d=0;d<ocl_vector;d++)
		    {
	    		e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector+d;
    			if ( (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count<320) && (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>0))
    			{
    			    int i;
    			    for (i=0;i<hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count;i++)
    			    if (memcmp(hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i]->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
    			    {
                    		mylist = hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i];
                    		int flag = 0;
                    		strcpy(plain,readline);
                    		plain[strlen(readline)]=bruteforce_charset[a1];
                    		plain[strlen(readline)+1]=bruteforce_charset[a2];
                    		plain[strlen(readline)+2]=bruteforce_charset[(a3+(d))%charset_size];
                    		plain[strlen(readline)+3]=bruteforce_charset[c];
                    		plain[strlen(readline)+4]=bruteforce_charset[b];
                    		plain[strlen(readline)+5]=bruteforce_charset[a];
                    		plain[strlen(readline)+6]=0;
                    		pthread_mutex_lock(&crackedmutex);
                    		addlist = cracked_list;
                    		while (addlist)
                    		{
                        	    if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
                            	    && (strcmp(addlist->salt, mylist->salt) == 0))
                                	flag = 1;
                        	    addlist = addlist->next;
                    		}
                    		pthread_mutex_unlock(&crackedmutex);
                    		if (flag == 0)
                    		{
                        	    add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    		}
    			    }
    			}
    			else if (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>=320)
    			{
    			    mylist = hash_list;
    			    while (mylist)
    			    {
            			if (memcmp(mylist->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
            			{
                    		    int flag = 0;
                    		    strcpy(plain,readline);
                    		    plain[strlen(readline)]=bruteforce_charset[a1];
                    		    plain[strlen(readline)+1]=bruteforce_charset[a2];
                    		    plain[strlen(readline)+2]=bruteforce_charset[(a3+(d))%charset_size];
                    		    plain[strlen(readline)+3]=bruteforce_charset[c];
                    		    plain[strlen(readline)+4]=bruteforce_charset[b];
                    		    plain[strlen(readline)+5]=bruteforce_charset[a];
                    		    plain[strlen(readline)+6]=0;
                    		    pthread_mutex_lock(&crackedmutex);
                    		    addlist = cracked_list;
                    		    while (addlist)
                    		    {
                        		if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
                            		&& (strcmp(addlist->salt, mylist->salt) == 0))
                                	flag = 1;
                        		addlist = addlist->next;
                    		    }
                    		    pthread_mutex_unlock(&crackedmutex);
                    		    if (flag == 0)
                    		    {
                        		add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    		    }
            			}
            			if (mylist) mylist = mylist->next;
        		    }
        		}
		    }
		    }
		clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
    		bzero(found_ind,charset_size*charset_size*charset_size*sizeof(cl_uint));
    		clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
		}
	    }

	    /* Hybrid, len=7 */
	    csize=4+strlen(readline);
	    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);
	    clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);
	    if ((bruteforce_start<=7) && (bruteforce_end>=7) && ((strlen(readline)+7)<16))
	    for (a1=0;a1<charset_size;a1++)
	    for (a2=0;a2<charset_size;a2++)
	    for (a3=0;a3<charset_size;a3++)
	    for (a4=0;a4<charset_size;a4++)
	    {
		if (attack_over!=0) return hash_ok;

		cbase.s0 = bruteforce_charset[a4];
		cbase.s1 = bruteforce_charset[a4+1];
		cbase.s2 = bruteforce_charset[a4+2];
		cbase.s3 = bruteforce_charset[a4+3];
		cbase.s4 = bruteforce_charset[a4+4];
		cbase.s5 = bruteforce_charset[a4+5];
		cbase.s6 = bruteforce_charset[a4+6];
		cbase.s7 = bruteforce_charset[a4+7];

		switch (strlen(readline))
		{
		    case 1:
			image.x=readline[0]|(bruteforce_charset[a1]<<8)|(bruteforce_charset[a2]<<16)|(bruteforce_charset[a3]<<24);
			image.y=image.z=image.w=0;
			break;
		    case 2:
			image.x=readline[0]|(readline[1]<<8)|(bruteforce_charset[a1]<<16)|(bruteforce_charset[a2]<<24);
			image.y=bruteforce_charset[a3];image.z=image.w=0;
			break;
		    case 3:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(bruteforce_charset[a1]<<24);
			image.y=(bruteforce_charset[a2])|(bruteforce_charset[a3]<<8);
			image.z=image.w=0;
			break;
		    case 4:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=image.z=image.w=0;
			image.y=bruteforce_charset[a1]|(bruteforce_charset[a2]<<8)|(bruteforce_charset[a3]<<16);
			break;
		    case 5:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=readline[4]|(bruteforce_charset[a1]<<8)|(bruteforce_charset[a2]<<16)|(bruteforce_charset[a3]<<24);
			image.z=image.w=0;
			break;
		    case 6:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=readline[4]|(readline[5]<<8)|(bruteforce_charset[a1]<<16)|(bruteforce_charset[a2]<<24);
			image.z=bruteforce_charset[a3];image.w=0;
			break;
		    case 7:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=readline[4]|(readline[5]<<8)|(readline[6]<<16)|(bruteforce_charset[a1]<<24);
			image.z=bruteforce_charset[a2]|(bruteforce_charset[a3]<<8);
			image.w=0;
			break;
		    case 8:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=readline[4]|(readline[5]<<8)|(readline[6]<<16)|(readline[7]<<24);
			image.z=image.w=0;
			image.z=bruteforce_charset[a1]|(bruteforce_charset[a2]<<8)|(bruteforce_charset[a3]<<16);
			break;
		    case 9:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=readline[4]|(readline[5]<<8)|(readline[6]<<16)|(readline[7]<<24);
			image.z=readline[8]|(bruteforce_charset[a1]<<8)|(bruteforce_charset[a2]<<16)|(bruteforce_charset[a3]<<24);
			image.w=0;
			break;
		    case 10:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=readline[4]|(readline[5]<<8)|(readline[6]<<16)|(readline[7]<<24);
			image.z=readline[8]|(readline[9]<<8)|(bruteforce_charset[a1]<<16)|(bruteforce_charset[a2]<<24);
			image.w=bruteforce_charset[a3];
			break;
		    case 11:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=readline[4]|(readline[5]<<8)|(readline[6]<<16)|(readline[7]<<24);
			image.z=readline[8]|(readline[9]<<8)|(readline[10]<<16)|(bruteforce_charset[a1]<<24);
			image.w=bruteforce_charset[a2]|(bruteforce_charset[a3]<<8);
			break;
		    default:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=readline[4]|(readline[5]<<8)|(readline[6]<<16)|(readline[7]<<24);
			image.z=readline[8]|(readline[9]<<8)|(readline[10]<<16)|(readline[11]<<24);
			image.w=0;
			break;
		}
		clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);

		clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
    		if (found==1)
    		{
    		    found = 0;
    		    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    		}
	        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
	        //clFlush(queue);
		invocations++;
		if (err != CL_SUCCESS)
		{
		    elog("clEnqueueNDRangeKernel error (%d)\n",err);
		    return hash_err;
		}
		//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
    		int *fnd;
    		fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
    		found = *fnd;

		if (found==1) 
		{
		    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, 0, charset_size*charset_size*charset_size*32*ocl_vector, ptr, 0, NULL, NULL);
		    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
		    e=-1;
		    for (a=0;a<charset_size;a++)
		    for (b=0;b<charset_size;b++)
		    for (c=0;c<charset_size;c++)
		    if (found_ind[a*charset_size*charset_size+b*charset_size+c]==1)
		    {
		    e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector;
    		    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, e*32, 32*ocl_vector, ptr+e*32, 0, NULL, NULL);
		    for (d=0;d<ocl_vector;d++)
		    {
	    		e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector+d;
    			if ( (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count<320) && (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>0))
    			{
    			    int i;
    			    for (i=0;i<hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count;i++)
    			    if (memcmp(hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i]->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
    			    {
                    		mylist = hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i];
                    		int flag = 0;
                    		strcpy(plain,readline);
                    		plain[strlen(readline)]=bruteforce_charset[a1];
                    		plain[strlen(readline)+1]=bruteforce_charset[a2];
                    		plain[strlen(readline)+2]=bruteforce_charset[a3];
                    		plain[strlen(readline)+3]=bruteforce_charset[(a4+(d))%charset_size];
                    		plain[strlen(readline)+4]=bruteforce_charset[c];
                    		plain[strlen(readline)+5]=bruteforce_charset[b];
                    		plain[strlen(readline)+6]=bruteforce_charset[a];
                    		plain[strlen(readline)+7]=0;
                    		pthread_mutex_lock(&crackedmutex);
                    		addlist = cracked_list;
                    		while (addlist)
                    		{
                        	    if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
                            	    && (strcmp(addlist->salt, mylist->salt) == 0))
                                	flag = 1;
                        	    addlist = addlist->next;
                    		}
                    		pthread_mutex_unlock(&crackedmutex);
                    		if (flag == 0)
                    		{
                        	    add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    		}
    			    }
    			}
    			else if (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>=320)
    			{
    			    mylist = hash_list;
    			    while (mylist)
    			    {
            			if (memcmp(mylist->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
            			{
                    		    int flag = 0;
                    		    strcpy(plain,readline);
                    		    plain[strlen(readline)]=bruteforce_charset[a1];
                    		    plain[strlen(readline)+1]=bruteforce_charset[a2];
                    		    plain[strlen(readline)+2]=bruteforce_charset[a3];
                    		    plain[strlen(readline)+3]=bruteforce_charset[(a4+(d))%charset_size];
                    		    plain[strlen(readline)+4]=bruteforce_charset[c];
                    		    plain[strlen(readline)+5]=bruteforce_charset[b];
                    		    plain[strlen(readline)+6]=bruteforce_charset[a];
                    		    plain[strlen(readline)+7]=0;
                    		    pthread_mutex_lock(&crackedmutex);
                    		    addlist = cracked_list;
                    		    while (addlist)
                    		    {
                        		if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
                            		&& (strcmp(addlist->salt, mylist->salt) == 0))
                                	flag = 1;
                        		addlist = addlist->next;
                    		    }
                    		    pthread_mutex_unlock(&crackedmutex);
                    		    if (flag == 0)
                    		    {
                        		add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    		    }
            			}
            			if (mylist) mylist = mylist->next;
        		    }
        		}
		    }
		    }
		clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
    		bzero(found_ind,charset_size*charset_size*charset_size*sizeof(cl_uint));
    		clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
		}
	    }

	    /* Hybrid, len=8 */
	    csize=4+strlen(readline);
	    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);
	    clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);
	    if ((bruteforce_start<=8) && (bruteforce_end>=8) && ((strlen(readline)+8)<16))
	    for (a1=0;a1<charset_size;a1++)
	    for (a2=0;a2<charset_size;a2++)
	    for (a3=0;a3<charset_size;a3++)
	    for (a4=0;a4<charset_size;a4++)
	    for (a5=0;a5<charset_size;a5++)
	    {
		if (attack_over!=0) return hash_ok;

		cbase.s0 = bruteforce_charset[a4];
		cbase.s1 = bruteforce_charset[a4+1];
		cbase.s2 = bruteforce_charset[a4+2];
		cbase.s3 = bruteforce_charset[a4+3];
		cbase.s4 = bruteforce_charset[a4+4];
		cbase.s5 = bruteforce_charset[a4+5];
		cbase.s6 = bruteforce_charset[a4+6];
		cbase.s7 = bruteforce_charset[a4+7];

		switch (strlen(readline))
		{
		    case 1:
			image.x=readline[0]|(bruteforce_charset[a1]<<8)|(bruteforce_charset[a2]<<16)|(bruteforce_charset[a3]<<24);
			image.y=bruteforce_charset[a4];image.z=image.w=0;
			break;
		    case 2:
			image.x=readline[0]|(readline[1]<<8)|(bruteforce_charset[a1]<<16)|(bruteforce_charset[a2]<<24);
			image.y=bruteforce_charset[a3]|(bruteforce_charset[a4]<<8);image.z=image.w=0;
			break;
		    case 3:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(bruteforce_charset[a1]<<24);
			image.y=(bruteforce_charset[a2])|(bruteforce_charset[a3]<<8)|(bruteforce_charset[a4]<<16);
			image.z=image.w=0;
			break;
		    case 4:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=image.z=image.w=0;
			image.y=bruteforce_charset[a1]|(bruteforce_charset[a2]<<8)|(bruteforce_charset[a3]<<16)|(bruteforce_charset[a4]<<24);
			break;
		    case 5:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=readline[4]|(bruteforce_charset[a1]<<8)|(bruteforce_charset[a2]<<16)|(bruteforce_charset[a3]<<24);
			image.z=bruteforce_charset[a4];image.w=0;
			break;
		    case 6:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=readline[4]|(readline[5]<<8)|(bruteforce_charset[a1]<<16)|(bruteforce_charset[a2]<<24);
			image.z=bruteforce_charset[a3]|(bruteforce_charset[a4]<<8);image.w=0;
			break;
		    case 7:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=readline[4]|(readline[5]<<8)|(readline[6]<<16)|(bruteforce_charset[a1]<<24);
			image.z=bruteforce_charset[a2]|(bruteforce_charset[a3]<<8)|(bruteforce_charset[a4]<<16);
			image.w=0;
			break;
		    case 8:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=readline[4]|(readline[5]<<8)|(readline[6]<<16)|(readline[7]<<24);
			image.z=image.w=0;
			image.z=bruteforce_charset[a1]|(bruteforce_charset[a2]<<8)|(bruteforce_charset[a3]<<16)|(bruteforce_charset[a4]<<24);
			break;
		    case 9:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=readline[4]|(readline[5]<<8)|(readline[6]<<16)|(readline[7]<<24);
			image.z=readline[8]|(bruteforce_charset[a1]<<8)|(bruteforce_charset[a2]<<16)|(bruteforce_charset[a3]<<24);
			image.w=bruteforce_charset[a4];
			break;
		    case 10:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=readline[4]|(readline[5]<<8)|(readline[6]<<16)|(readline[7]<<24);
			image.z=readline[8]|(readline[9]<<8)|(bruteforce_charset[a1]<<16)|(bruteforce_charset[a2]<<24);
			image.w=bruteforce_charset[a3]|(bruteforce_charset[a4]<<8);
			break;
		    case 11:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=readline[4]|(readline[5]<<8)|(readline[6]<<16)|(readline[7]<<24);
			image.z=readline[8]|(readline[9]<<8)|(readline[10]<<16)|(bruteforce_charset[a1]<<24);
			image.w=bruteforce_charset[a2]|(bruteforce_charset[a3]<<8)|(bruteforce_charset[a4]<<16);
			break;
		    default:
			image.x=readline[0]|(readline[1]<<8)|(readline[2]<<16)|(readline[3]<<24);
			image.y=readline[4]|(readline[5]<<8)|(readline[6]<<16)|(readline[7]<<24);
			image.z=readline[8]|(readline[9]<<8)|(readline[10]<<16)|(readline[11]<<24);
			image.w=0;
			break;
		}
		clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);

		clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
    		if (found==1)
    		{
    		    found = 0;
    		    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    		}
	        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, NULL, 0, NULL, NULL);
	        //clFlush(queue);
		invocations++;
		if (err != CL_SUCCESS)
		{
		    elog("clEnqueueNDRangeKernel error (%d)\n",err);
		    return hash_err;
		}
		//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
    		int *fnd;
    		fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
    		found = *fnd;

		if (found==1) 
		{
		    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, 0, charset_size*charset_size*charset_size*32*ocl_vector, ptr, 0, NULL, NULL);
		    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
		    e=-1;
		    for (a=0;a<charset_size;a++)
		    for (b=0;b<charset_size;b++)
		    for (c=0;c<charset_size;c++)
		    if (found_ind[a*charset_size*charset_size+b*charset_size+c]==1)
		    {
		    e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector;
    		    err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, e*32, 32*ocl_vector, ptr+e*32, 0, NULL, NULL);

		    for (d=0;d<ocl_vector;d++)
		    {
	    		e=(a*charset_size*charset_size+b*charset_size+c)*ocl_vector+d;
    			if ( (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count<320) && (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>0))
    			{
    			    int i;
    			    for (i=0;i<hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count;i++)
    			    if (memcmp(hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i]->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
    			    {
                    		mylist = hash_index[ptr[e*32]&255][ptr[e*32+1]&255].nodes[i];
                    		int flag = 0;
                    		strcpy(plain,readline);
                    		plain[strlen(readline)]=bruteforce_charset[a1];
                    		plain[strlen(readline)+1]=bruteforce_charset[a2];
                    		plain[strlen(readline)+2]=bruteforce_charset[a3];
                    		plain[strlen(readline)+3]=bruteforce_charset[a4];
                    		plain[strlen(readline)+4]=bruteforce_charset[(a5+(d))%charset_size];
                    		plain[strlen(readline)+5]=bruteforce_charset[c];
                    		plain[strlen(readline)+6]=bruteforce_charset[b];
                    		plain[strlen(readline)+7]=bruteforce_charset[a];
                    		plain[strlen(readline)+8]=0;
                    		pthread_mutex_lock(&crackedmutex);
                    		addlist = cracked_list;
                    		while (addlist)
                    		{
                        	    if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
                            	    && (strcmp(addlist->salt, mylist->salt) == 0))
                                	flag = 1;
                        	    addlist = addlist->next;
                    		}
                    		pthread_mutex_unlock(&crackedmutex);
                    		if (flag == 0)
                    		{
                        	    add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    		}
    			    }
    			}
    			else if (hash_index[ptr[e*32]&255][ptr[e*32+1]&255].count>=320)
    			{
    			    mylist = hash_list;
    			    while (mylist)
    			    {
            			if (memcmp(mylist->hash, (char *)ptr+(e)*32, hash_ret_len) == 0)
            			{
                    		    int flag = 0;
                    		    strcpy(plain,readline);
                    		    plain[strlen(readline)]=bruteforce_charset[a1];
                    		    plain[strlen(readline)+1]=bruteforce_charset[a2];
                    		    plain[strlen(readline)+2]=bruteforce_charset[a3];
                    		    plain[strlen(readline)+3]=bruteforce_charset[a4];
                    		    plain[strlen(readline)+4]=bruteforce_charset[(a5+(d))%charset_size];
                    		    plain[strlen(readline)+5]=bruteforce_charset[c];
                    		    plain[strlen(readline)+6]=bruteforce_charset[b];
                    		    plain[strlen(readline)+7]=bruteforce_charset[a];
                    		    plain[strlen(readline)+8]=0;
                    		    pthread_mutex_lock(&crackedmutex);
                    		    addlist = cracked_list;
                    		    while (addlist)
                    		    {
                        		if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
                            		&& (strcmp(addlist->salt, mylist->salt) == 0))
                                	flag = 1;
                        		addlist = addlist->next;
                    		    }
                    		    pthread_mutex_unlock(&crackedmutex);
                    		    if (flag == 0)
                    		    {
                        		add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                    		    }
            			}
            			if (mylist) mylist = mylist->next;
        		    }
        		}
		    }
		    }
		clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
    		bzero(found_ind,charset_size*charset_size*charset_size*sizeof(cl_uint));
    		clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, charset_size*charset_size*charset_size*sizeof(cl_uint), found_ind, 0, NULL, NULL);
		}
	    }

	/* end actual work */
	}
	dictpos++;
    }
    return hash_ok;
}





void ocl_bruteforce_init_restore(int self, int len)
{
    int b;

    for (b=0;b<16;b++) thread_restore[self].cpos[b] = 0;
    thread_restore[self].cpos[len] = self*ocl_vector;
    bzero(thread_restore[self].cstr,16);
    thread_restore[self].cstart = 0;
}


void ocl_markov_init_restore(int self)
{
    int b;

    for (b=0;b<16;b++) thread_restore[self].cpos[b] = 0;
    thread_restore[self].cpos[0] = self*ocl_vector;
    bzero(thread_restore[self].cstr,16);
    thread_restore[self].cstart = 0;
}







hash_stat ocl_bruteforce_sha512(void)
{
    int a,b;
    int err;
    int worker_thread_keys[32];
    FILE *kfile;
    char *source_str;
    size_t source_size;
    char source_path[255];

    /* we are not restoring a session? */
    if (session_restore_flag==0)
    {
	for (a=0;a<64;a++) 
	{
	    ocl_bruteforce_init_restore(a, 0);
	}
    }

    if (bruteforce_end<5) 
    {
	bruteforce_end = bruteforce_start=5;
	hlog("Raising bruteforce start and max len limit to %d\n",5);
	if (session_restore_flag==1) bruteforce_end -=5;
    }
    b=strlen(bruteforce_charset);
    for (a=5;a<=bruteforce_end;a++) b*=strlen(bruteforce_charset);
    b = b/ocl_vector;
    attack_overall_count = b;
    if (strlen(bruteforce_charset)<=26) attack_overall_count /= strlen(bruteforce_charset);

    /* setup initial OpenCL vars */
    err = clGetPlatformIDs(1, &platform, NULL);
    if (err != CL_SUCCESS)
    {
	elog("clGetPlatformIDs error (%d)\n",err);
	return hash_err;
    }

    err = clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, 16, device, &devicesnum);
    if (err != CL_SUCCESS)
    {
        elog("clGetDeviceIDs error (%d)\n",err);
        return hash_err;
    }

    for (a=0;a<devicesnum*ocl_threads;a++)
    {
	context[a] = clCreateContext(NULL, 1, &device[a/ocl_threads], NULL, NULL, &err);
	if (err != CL_SUCCESS)
	{
	    elog("clCreateContext error (%d)\n",err);
	    return hash_err;
	}
    }
    
    if (strlen(bruteforce_charset)>26)
    {
	if (ocl_dev_nvidia == 1)
	{
	    sprintf(source_path,"%s%s",KERNEL_PATH,"nvidia_sha512.cl");
	    hlog("Compiling OpenCL kernel source (%s)","nvidia_sha512.cl");
	}
	else
	{
	    sprintf(source_path,"%s%s",KERNEL_PATH,"amd_sha512.cl");
	    hlog("Compiling OpenCL kernel source (%s)","amd_sha512.cl");
	}
	kfile = fopen(source_path, "r");
        if (!kfile) 
        {
    	    elog("Cannot load kernel: %s\n",source_path);
    	    return hash_err;
        }
	source_str = (char*)malloc(MAX_SOURCE_SIZE);
        source_size = fread(source_str, 1, MAX_SOURCE_SIZE, kfile);
        fclose(kfile);
	
	for (a=0;a<devicesnum*ocl_threads;a++)
	{
	    program[a] = clCreateProgramWithSource(context[a], 1, (const char **)&source_str,  (const size_t *)&source_size, &err );
	    if (err != CL_SUCCESS)
	    {
		elog("clCreateProgramWithSource error (%d)\n",err);
		return hash_err;
	    }

    	    if ((ocl_have_old_ati==0)&&(ocl_dev_nvidia == 0))
    	    {
        	size_t nDevices=1;
        	size_t * binary_sizes;
        	char ** binaries;
        	int i;

        	if (hash_list->next) err = clBuildProgram(program[a], 1, &device[a/ocl_threads], NULL, NULL, NULL );
        	else err = clBuildProgram(program[a], 1, &device[a/ocl_threads], "-DSINGLE_MODE ", NULL, NULL );

        	if (err != CL_SUCCESS)
        	{
            	    elog("clBuildProgram error (%d)\n",err);
            	    return hash_err;
        	}

        	err = clGetProgramInfo( program[a], CL_PROGRAM_NUM_DEVICES, sizeof(nDevices), &nDevices, NULL );
        	binary_sizes = (size_t *)malloc( sizeof(size_t)*nDevices );
        	err = clGetProgramInfo( program[a], CL_PROGRAM_BINARY_SIZES, sizeof(size_t)*nDevices, binary_sizes, NULL );

        	binaries = (char **)malloc( sizeof(char *)*nDevices );
        	for( i = 0; i < nDevices; i++ )
        	{
            	    if( binary_sizes[i] != 0 )
            	    {
                	binaries[i] = (char *)malloc( sizeof(char)*binary_sizes[i]+100 );
            	    }
            	    else
            	    {
                	binaries[i] = NULL;
            	    }
        	}
            
        	clGetProgramInfo( program[a], CL_PROGRAM_BINARIES, sizeof(char *)*nDevices, binaries, NULL );

        	if (a==0) 
        	{
            	    printf("\n");
            	    hlog("Binary size: %d\n",binary_sizes[0]);
            	    hlog("Doing BFI_INT magic...%s\n","");
        	}
        	bfi_int_magic(binaries[0],binary_sizes[0]);
        	program[a] = clCreateProgramWithBinary(context[a], 1, &device[a/ocl_threads], &binary_sizes[0], (const unsigned char **)&binaries[0], NULL, &err );
        	if (err!=CL_SUCCESS) elog("Cannot compile binary!\n%s","");
        	err = clBuildProgram(program[a], 1, &device[a/ocl_threads], NULL, NULL, NULL );
        	if (err!=CL_SUCCESS) elog("Cannot build binary!\n%s","");
    	    }
    	    else
    	    {
        	if (hash_list->next) err = clBuildProgram(program[a], 1, &device[a/ocl_threads], "-DOLD_ATI", NULL, NULL );
        	else err = clBuildProgram(program[a], 1, &device[a/ocl_threads], "-DSINGLE_MODE -DOLD_ATI", NULL, NULL );
        	if (err != CL_SUCCESS)
        	{
            	    elog("clBuildProgram error (%d)\n",err);
            	    return hash_err;
        	}
    	    }
	}

	pthread_mutex_init(&biglock, NULL);

	for (a=0;a<devicesnum*ocl_threads;a++)
	{
	    worker_thread_keys[a]=a;
	    pthread_create(&crack_threads[a], NULL, ocl_bruteforce_sha512_thread, &worker_thread_keys[a]);
	}
    }
    else
    {
	if (ocl_dev_nvidia == 1)
	{
	    sprintf(source_path,"%s%s",KERNEL_PATH,"nvidia_sha512_short.cl");
	    hlog("Compiling OpenCL kernel source (%s)","nvidia_sha512_short.cl");
	}
	else
	{
	    sprintf(source_path,"%s%s",KERNEL_PATH,"amd_sha512_short.cl");
	    hlog("Compiling OpenCL kernel source (%s)","amd_sha512_short.cl");
	}
	kfile = fopen(source_path, "r");
        if (!kfile) 
        {
    	    elog("Cannot load kernel: %s\n",source_path);
    	    return hash_err;
        }
	source_str = (char*)malloc(MAX_SOURCE_SIZE);
        source_size = fread(source_str, 1, MAX_SOURCE_SIZE, kfile);
        fclose(kfile);
	
	for (a=0;a<devicesnum*ocl_threads;a++)
	{
	    program[a] = clCreateProgramWithSource(context[a], 1, (const char **)&source_str,  (const size_t *)&source_size, &err );
	    if (err != CL_SUCCESS)
	    {
		elog("clCreateProgramWithSource error (%d)\n",err);
		return hash_err;
	    }

    	    if ((ocl_have_old_ati==0)&&(ocl_dev_nvidia == 0))
    	    {
        	size_t nDevices=1;
        	size_t * binary_sizes;
        	char ** binaries;
        	int i;

        	if (hash_list->next) err = clBuildProgram(program[a], 1, &device[a/ocl_threads], NULL, NULL, NULL );
        	else err = clBuildProgram(program[a], 1, &device[a/ocl_threads], "-DSINGLE_MODE ", NULL, NULL );

        	if (err != CL_SUCCESS)
        	{
            	    elog("clBuildProgram error (%d)\n",err);
            	    return hash_err;
        	}

        	err = clGetProgramInfo( program[a], CL_PROGRAM_NUM_DEVICES, sizeof(nDevices), &nDevices, NULL );
        	binary_sizes = (size_t *)malloc( sizeof(size_t)*nDevices );
        	err = clGetProgramInfo( program[a], CL_PROGRAM_BINARY_SIZES, sizeof(size_t)*nDevices, binary_sizes, NULL );

        	binaries = (char **)malloc( sizeof(char *)*nDevices );
        	for( i = 0; i < nDevices; i++ )
        	{
            	    if( binary_sizes[i] != 0 )
            	    {
                	binaries[i] = (char *)malloc( sizeof(char)*binary_sizes[i]+100 );
            	    }
            	    else
            	    {
                	binaries[i] = NULL;
            	    }
        	}
            
        	clGetProgramInfo( program[a], CL_PROGRAM_BINARIES, sizeof(char *)*nDevices, binaries, NULL );

        	if (a==0) 
        	{
            	    printf("\n");
            	    hlog("Binary size: %d\n",binary_sizes[0]);
            	    hlog("Doing BFI_INT magic...%s\n","");
        	}
        	bfi_int_magic(binaries[0],binary_sizes[0]);
        	program[a] = clCreateProgramWithBinary(context[a], 1, &device[a/ocl_threads], &binary_sizes[0], (const unsigned char **)&binaries[0], NULL, &err );
        	if (err!=CL_SUCCESS) elog("Cannot compile binary!\n%s","");
        	err = clBuildProgram(program[a], 1, &device[a/ocl_threads], NULL, NULL, NULL );
        	if (err!=CL_SUCCESS) elog("Cannot build binary!\n%s","");
    	    }
    	    else
    	    {
        	if (hash_list->next) err = clBuildProgram(program[a], 1, &device[a/ocl_threads], "-DOLD_ATI", NULL, NULL );
        	else err = clBuildProgram(program[a], 1, &device[a/ocl_threads], "-DSINGLE_MODE -DOLD_ATI", NULL, NULL );
        	if (err != CL_SUCCESS)
        	{
            	    elog("clBuildProgram error (%d)\n",err);
            	    return hash_err;
        	}
    	    }
	}
	
	pthread_mutex_init(&biglock, NULL);

	for (a=0;a<devicesnum*ocl_threads;a++)
	{
	    worker_thread_keys[a]=a;
	    pthread_create(&crack_threads[a], NULL, ocl_bruteforce_sha512_short_thread, &worker_thread_keys[a]);
	}    
    }
    
    for (a=0;a<devicesnum*ocl_threads;a++) pthread_join(crack_threads[a], NULL);
    printf("\n");
    hlog("Done!\n%s","");
    attack_over=2;
}



hash_stat ocl_markov_sha512(void)
{
    int a;
    int err;
    int worker_thread_keys[32];
    FILE *kfile;
    char *source_str;
    size_t source_size;
    char source_path[255];


    if (fast_markov==1)  hlog("Fast markov attack mode enabled%s\n","");

    /* we are not restoring a session? */
    if (session_restore_flag==0)
    {
	for (a=0;a<64;a++) 
	{
	    ocl_markov_init_restore(a);
	}
    }

    /* setup initial OpenCL vars */
    err = clGetPlatformIDs(1, &platform, NULL);
    if (err != CL_SUCCESS)
    {
	elog("clGetPlatformIDs error (%d)\n",err);
	return hash_err;
    }

    err = clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, 16, device, &devicesnum);
    if (err != CL_SUCCESS)
    {
        elog("clGetDeviceIDs error (%d)\n",err);
        return hash_err;
    }

    for (a=0;a<devicesnum*ocl_threads;a++)
    {
	context[a] = clCreateContext(NULL, 1, &device[a/ocl_threads], NULL, NULL, &err);
	if (err != CL_SUCCESS)
	{
	    elog("clCreateContext error (%d)\n",err);
	    return hash_err;
	}
    }

    if (ocl_dev_nvidia == 1)
    {
        sprintf(source_path,"%s%s",KERNEL_PATH,"nvidia_sha512_markov.cl");
        hlog("Compiling OpenCL kernel source (%s)","nvidia_sha512_markov.cl");
    }
    else
    {
        sprintf(source_path,"%s%s",KERNEL_PATH,"amd_sha512_markov.cl");
        hlog("Compiling OpenCL kernel source (%s)","amd_sha512_markov.cl");
    }
    kfile = fopen(source_path, "r");
    if (!kfile) 
    {
        elog("Cannot load kernel: %s\n",source_path);
        return hash_err;
    }
    source_str = (char*)malloc(MAX_SOURCE_SIZE);
    source_size = fread(source_str, 1, MAX_SOURCE_SIZE, kfile);
    fclose(kfile);

    for (a=0;a<devicesnum*ocl_threads;a++)
    {
	program[a] = clCreateProgramWithSource(context[a], 1, (const char **)&source_str,  (const size_t *)&source_size, &err );
	if (err != CL_SUCCESS)
	{
	    elog("clCreateProgramWithSource error (%d)\n",err);
	    return hash_err;
	}

        if ((ocl_have_old_ati==0)&&(ocl_dev_nvidia == 0))
        {
            size_t nDevices=1;
            size_t * binary_sizes;
            char ** binaries;
            int i;

            if (hash_list->next) err = clBuildProgram(program[a], 1, &device[a/ocl_threads], NULL, NULL, NULL );
            else err = clBuildProgram(program[a], 1, &device[a/ocl_threads], "-DSINGLE_MODE ", NULL, NULL );

            if (err != CL_SUCCESS)
            {
                elog("clBuildProgram error (%d)\n",err);
                return hash_err;
            }

            err = clGetProgramInfo( program[a], CL_PROGRAM_NUM_DEVICES, sizeof(nDevices), &nDevices, NULL );
            binary_sizes = (size_t *)malloc( sizeof(size_t)*nDevices );
            err = clGetProgramInfo( program[a], CL_PROGRAM_BINARY_SIZES, sizeof(size_t)*nDevices, binary_sizes, NULL );

            binaries = (char **)malloc( sizeof(char *)*nDevices );
            for( i = 0; i < nDevices; i++ )
            {
                if( binary_sizes[i] != 0 )
                {
                    binaries[i] = (char *)malloc( sizeof(char)*binary_sizes[i]+100 );
                }
                else
                {
                    binaries[i] = NULL;
                }
            }
            
            clGetProgramInfo( program[a], CL_PROGRAM_BINARIES, sizeof(char *)*nDevices, binaries, NULL );

            if (a==0) 
            {
                printf("\n");
                hlog("Binary size: %d\n",binary_sizes[0]);
                hlog("Doing BFI_INT magic...%s\n","");
            }
            bfi_int_magic(binaries[0],binary_sizes[0]);
            program[a] = clCreateProgramWithBinary(context[a], 1, &device[a/ocl_threads], &binary_sizes[0], (const unsigned char **)&binaries[0], NULL, &err );
            if (err!=CL_SUCCESS) elog("Cannot compile binary!\n%s","");
            err = clBuildProgram(program[a], 1, &device[a/ocl_threads], NULL, NULL, NULL );
            if (err!=CL_SUCCESS) elog("Cannot build binary!\n%s","");
        }
        else
        {
            if (hash_list->next) err = clBuildProgram(program[a], 1, &device[a/ocl_threads], "-DOLD_ATI", NULL, NULL );
            else err = clBuildProgram(program[a], 1, &device[a/ocl_threads], "-DSINGLE_MODE -DOLD_ATI", NULL, NULL );
            if (err != CL_SUCCESS)
            {
                elog("clBuildProgram error (%d)\n",err);
                return hash_err;
            }
        }
    }
    
    pthread_mutex_init(&biglock, NULL);

    for (a=0;a<devicesnum*ocl_threads;a++)
    {
	worker_thread_keys[a]=a;
	pthread_create(&crack_threads[a], NULL, ocl_markov_sha512_thread, &worker_thread_keys[a]);
    }
    
    for (a=0;a<devicesnum*ocl_threads;a++) pthread_join(crack_threads[a], NULL);
    printf("\n");
    hlog("Done!\n%s","");
    attack_over=2;

}


/* Main thread - hybrid */
hash_stat ocl_hybrid_sha512(void)
{
    int a,b;
    int err;
    int worker_thread_keys[32];
    FILE *kfile;
    char *source_str;
    size_t source_size;
    char source_path[255];


    /* setup initial OpenCL vars */
    err = clGetPlatformIDs(1, &platform, NULL);
    if (err != CL_SUCCESS)
    {
	elog("clGetPlatformIDs error (%d)\n",err);
	return hash_err;
    }

    err = clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, 16, device, &devicesnum);
    if (err != CL_SUCCESS)
    {
        elog("clGetDeviceIDs error (%d)\n",err);
        return hash_err;
    }

    context[0] = clCreateContext(NULL, devicesnum, device, NULL, NULL, &err);
    if (err != CL_SUCCESS)
    {
	elog("clCreateContext error (%d)\n",err);
	return hash_err;
    }
    
    if (ocl_dev_nvidia == 1)
    {
        sprintf(source_path,"%s%s",KERNEL_PATH,"nvidia_sha512.cl");
        hlog("Compiling OpenCL kernel source (%s)","nvidia_sha512.cl");
    }
    else
    {
        sprintf(source_path,"%s%s",KERNEL_PATH,"amd_sha512.cl");
        hlog("Compiling OpenCL kernel source (%s)","amd_sha512.cl");
    }
    kfile = fopen(source_path, "r");
    if (!kfile) 
    {
        elog("Cannot load kernel: %s\n",source_path);
        return hash_err;
    }
    source_str = (char*)malloc(MAX_SOURCE_SIZE);
    source_size = fread(source_str, 1, MAX_SOURCE_SIZE, kfile);
    fclose(kfile);

    program[0] = clCreateProgramWithSource(context[0], 1, (const char **)&source_str,  (const size_t *)&source_size, &err );
    if (err != CL_SUCCESS)
    {
        elog("clCreateProgramWithSource error (%d)\n",err);
        return hash_err;
    }

    if (hash_list->next) err = clBuildProgram(program[0], 1, device, "-cl-strict-aliasing", NULL, NULL );
    else err = clBuildProgram(program[0], 1, device, "-cl-strict-aliasing -DSINGLE_MODE", NULL, NULL );
    
    if (err != CL_SUCCESS)
    {
        elog("clBuildProgram error (%d)\n",err);
        return hash_err;
    }

    pthread_mutex_init(&biglock, NULL);

    for (a=0;a<devicesnum*ocl_threads;a++)
    {
        worker_thread_keys[a]=a;
        pthread_create(&crack_threads[a], NULL, ocl_hybrid_sha512_thread, &worker_thread_keys[a]);
    }

    for (a=0;a<devicesnum*ocl_threads;a++) pthread_join(crack_threads[a], NULL);
    printf("\n");
    hlog("Done!\n%s","");
    attack_over=2;
}





#endif
