/* hashinterface.h
 *
 * hashkill - a hash cracking tool
 * Copyright (C) 2010 Milen Rangelov <gat3way@gat3way.eu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#ifndef HASHINTERFACE_H
#define HASHINTERFACE_H


/* Do not change those unless you know what you are doing */
#define HASHKILL_VERSION "0.2.4"			// hashkill version
#define HASHKILL_MAXTHREADS (unsigned int)64		// maximum threads
#define HASHKILL_MAXQUEUESIZE 128 			// maximum queue size
#define HASHFILE_MAX_LINE_LENGTH 128			// maximum line length
#define HASHFILE_MAX_PLAIN_LENGTH 16
#define VECTORSIZE 128

#define THREAD_LENPROVIDED 1000
#define THREAD_SALTPROVIDED 20000

#define KERNEL_PATH "/usr/share/hashkill/kernels/"
#define MAX_SOURCE_SIZE 350000


/* thread queues */
typedef struct hash_lifo_node_s
{
    char *plaintext[VECTORSIZE];
    int lens[VECTORSIZE];
} hash_lifo[HASHKILL_MAXQUEUESIZE];

hash_lifo hash_queue[HASHKILL_MAXTHREADS];


/* hashes linked lists */
struct hash_list_s
{
    char *username;
    char *hash;
    char *salt;
    char *salt2;
    struct hash_list_s *prev;
    struct hash_list_s *next;
} *hash_list, *cracked_list, *hash_list_end, *cracked_list_end;


/* Attack method enumeration */
typedef enum attack_method_e
{
    attack_method_dictionary,
    attack_method_simple_bruteforce,
    attack_method_hybrid,
    attack_method_markov
} attack_method_t;


/* Hash indexes */
struct hash_index_t
{
    struct hash_list_s *nodes[200];
    int count;
} hash_index[256][256];



attack_method_t attack_method;
int hash_ret_len;
int single_hash;

/* Global variables needed by modules */
char hashlist_file[255];		// Hashlist file
char hash_dictionary_file[255];		// Dictionary file
int bruteforce_start;			// Min len of bruteforce candidates
int bruteforce_end;			// Max len of bruteforce candidates
char bruteforce_charset[255];		// Bruteforce charset
char bruteforce_prefix[64];		// Used in CPU attacks only
char bruteforce_suffix[64];		// Not used anymore
int hashes_count;			// The number of hashes in the list
int session_restore_flag;		// Did we restore a session?
char markov_statfile[255];		// Markov statfile
int markov_max_len;			// Max len of markov-generated candidates
int markov_threshold;			// Markov threshold
int markov0[88];			// Markov order 0 probs
int markov1[88][88];			// Markov order 1 probs
char *markov_charset;			// Markov charset
int hash_is_raw;			// Whether we keep it in binary or ascii form
int vectorsize;				// Vector (well not quite) size on CPU
int fast_markov;			// Fast markov mode (GPU only)
int cpuonly;				// CPU-only attack
int attack_over;			// Attack over yet?
int cpu_optimize_single;
int ocl_gpu_group;
int ctrl_c_pressed;

/* List manipulation routines */
hash_stat add_hash_list(char *username, char *hash, char *salt, char *salt2);
hash_stat add_cracked_list(char *username, char *hash, char *salt, char *salt2);
hash_stat del_hash_list(struct hash_list_s *node);
hash_stat del_cracked_list(struct hash_list_s *node);


/* Some linked lists stuff */
int get_cracked_num(void);
void print_hash_list(void);
int get_hashes_num(void);
void print_cracked_list(void);
void print_cracked_list_to_file(char *filename);
void print_uncracked_list_to_file(char *filename);


/* Proto callback functions that are registered to the plugin library */
void hash_proto_add_username(const char *username);
void hash_proto_add_hash(const char *hash, int len);
void hash_proto_add_salt(const char *salt);
void hash_proto_add_salt2(const char *salt2);
void hash_proto_md5(char *plaintext[VECTORSIZE], char *hashmd5[VECTORSIZE], int len, int threadid);
void hash_proto_md5_unicode(char *plaintext[VECTORSIZE], char *hashmd5[VECTORSIZE], int len[VECTORSIZE]);
void hash_proto_md5_unicode_slow(char *plaintext[VECTORSIZE], char *hashmd5[VECTORSIZE], int len[VECTORSIZE]);
void hash_proto_md5_slow(char *plaintext[VECTORSIZE], char *hashmd5[VECTORSIZE], int len, int threadid);
void hash_proto_md4(char *plaintext[VECTORSIZE], char *hashmd5[VECTORSIZE], int len[VECTORSIZE]);
void hash_proto_ripemd160(const char *plaintext[VECTORSIZE], char *hashmd5[VECTORSIZE], int len[VECTORSIZE]);
void hash_proto_md5_hex(const char *hash[VECTORSIZE],  char *hashhex[VECTORSIZE]);
void hash_proto_sha1(char *plaintext[VECTORSIZE], char *hashmd5[VECTORSIZE], int len, int threadid);
void hash_proto_sha1_unicode(char *plaintext[VECTORSIZE], char *hashmd5[VECTORSIZE], int len[VECTORSIZE]);
void hash_proto_sha1_slow(char *plaintext[VECTORSIZE], char *hashmd5[VECTORSIZE], int len[VECTORSIZE]);
void hash_proto_sha1_hex(const char *hash[VECTORSIZE], char *hashhex[VECTORSIZE]);
void hash_proto_sha256_unicode(const char *plaintext[VECTORSIZE], char *hashmd5[VECTORSIZE], int len[VECTORSIZE]);
void hash_proto_sha256_hex(const char *hash[VECTORSIZE], char *hashhex[VECTORSIZE]);
void hash_proto_sha512_unicode(const char *plaintext[VECTORSIZE], char *hashmd5[VECTORSIZE], int len[VECTORSIZE]);
void hash_proto_sha512_hex(const char *hash[VECTORSIZE], char *hashhex[VECTORSIZE]);
void hash_proto_fcrypt(const char *password[VECTORSIZE], const char *salt, char *ret[VECTORSIZE]);
void hash_proto_new_biomem(FILE *file);
void hash_proto_PEM_readfile(const char *passphrase, int *RSAret);
void hash_proto_pbkdf2(const char *pass, unsigned char *salt, int saltlen, int iter, int keylen, unsigned char *out);
void hash_proto_hmac_sha1_file(void *key, int keylen, char *filename, long offset, long size, unsigned char *output, int outputlen);
void hash_proto_pbkdf512(const char *pass, unsigned char *salt, int saltlen, int iter, int keylen, unsigned char *out);
void hash_proto_aes_encrypt(const unsigned char *key, int keysize, const unsigned char *in, int len, unsigned char *vec, unsigned char *out, int mode);
void hash_proto_aes_decrypt(const unsigned char *key, int keysize, const unsigned char *in, int len, unsigned char *vec, unsigned char *out, int mode);
void hash_proto_des_ecb_encrypt(const unsigned char *key, int keysize, const unsigned char *in[VECTORSIZE], int len, unsigned char *out[VECTORSIZE], int mode);
void hash_proto_des_ecb_decrypt(const unsigned char *key, int keysize, const unsigned char *in, int len, unsigned char *out, int mode);
void hash_proto_des_cbc_encrypt(const unsigned char *key[VECTORSIZE], int keysize, const unsigned char *in[VECTORSIZE], int len[VECTORSIZE], unsigned char *out[VECTORSIZE], unsigned char *iv[VECTORSIZE], int mode);
void hash_proto_lm(const unsigned char *in[VECTORSIZE], unsigned char *out[VECTORSIZE]);


/* cleanup routines */
void cleanup_lists(void);

/* Markov attack routines */
void markov_attack_init(void);
void markov_print_statfiles(void);
hash_stat markov_load_statfile(char *statname);

/* hash indexes routines */
hash_stat create_hash_indexes(void);


#endif
