/*
 * ocl_sl3.c
 *
 * hashkill - a hash cracking tool
 * Copyright (C) 2010 Milen Rangelov <gat3way@gat3way.eu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CL_CL_H

#include <CL/cl.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include "err.h"
#include "ocl-threads.h"
#include "plugins.h"
#include "hashinterface.h"
#include "sessions.h"
#include "ocl_support.h"

char sl3bruteforce_charset[10];

cl_platform_id platform;
cl_device_id device[16];
pthread_mutex_t biglock;

cl_context context[64];
int devicesnum;
pthread_t crack_threads[32];
cl_program program[64];


/* Some reversal macros/functions */
#define H0 0x67452301
#define H1 0xEFCDAB89
#define H2 0x98BADCFE
#define H3 0x10325476
#define H4 0xC3D2E1F0
#define ROTL(p,q) ((p) << (q)) | ((p) >> (32-(q)));
#define ROTR(p,q) ((p) >> (q)) | ((p) << (32-(q)));
#define REV(p) { tmp=(p);tmp1=ROTL(tmp,8);tmp2=ROTL(tmp,24); (p)=(tmp1 & 0x00FF00FF)|(tmp2 & 0xFF00FF00); }



static void ocl_bruteforce_init_restore(int self, int len);
static void ocl_markov_init_restore(int self);

#define ALGO_THRESHOLD 148

static void bfi_int_magic(unsigned char *kernel, int len)
{
    unsigned int dword0,dword1,dword2,j,i,k;
    
    for (k=4;k<12;k++)
    {
        j=0;
        for (i=4+k;i<(len-8);i+=8)
        {
            dword0 = *(unsigned int *)(kernel + i);
            dword1 = *(unsigned int *)(kernel + i-4);
            if  (
                ((((dword1)>>5)&31)==13) && // INST
                ((((dword1)>>4)&1)==0) &&   // NEG
                ((((dword1)>>23)&1)==0) &&    // CLAMP
                ((((dword1)>>20)&1)==0) &&    // DSTREL
                ((((dword0)>>4)&1)==0) &&    // SRC1NEG
                ((((dword0)>>17)&1)==0)     // SRC2NEG
            ) j++;
        }
        if (j>=ALGO_THRESHOLD)
        for (i=4+k;i<(len-8);i+=8)
        {
            dword0 = *(unsigned int *)(kernel + i);
            dword1 = *(unsigned int *)(kernel + i-4);
            if  (
                ((((dword1)>>5)&31)==13) && // INST
                ((((dword1)>>4)&1)==0) &&   // NEG
                ((((dword1)>>23)&1)==0) &&    // CLAMP
                ((((dword1)>>20)&1)==0) &&    // DSTREL
                ((((dword0)>>4)&1)==0) &&    // SRC1NEG
                ((((dword0)>>17)&1)==0)     // SRC2NEG
            )
            {
                *(unsigned int *)(kernel+i-4) = (dword1 & 0xFFFFFFFF) ^ (11 << 5);
            }
        }
    }

}




hash_stat ocl_bruteforce_sl3_thread(void *arg)
{
    int err;
    cl_command_queue queue;
    cl_mem buffer;
    size_t global_work_size;
    cl_uint4 image;
    int self;
    cl_kernel kernel;
    unsigned char hex1[20];
    char plain[64];
    cl_mem bitmaps_buf;
    cl_uint *bitmaps;
    int a,b,c,d,e,f,g,h;
    int a1,a2,a3,a4,a5,a6,a7,a8,a9,a10;
    char *ptr = malloc(10*10*10*10*10*10*20*ocl_vector);
    int charset_size = 10;
    cl_mem found_ind_buf;
    cl_uint *found_ind=malloc(10*10*10*10*10*10*sizeof(cl_uint));
    int found;
    cl_mem found_buf;
    struct  hash_list_s  *mylist, *addlist;
    unsigned int csize=1;
    cl_mem table_buf;
    unsigned int *table;
    cl_uint8 cbase;
    cl_uint4 singlehash;
    cl_uint4 salt;
    int session_restore_lflag;
    unsigned int tmp,tmp1,tmp2;
    size_t nvidia_local_work_size[3]={128,1,1};
    size_t *local_work_size;

    if (ocl_dev_nvidia) local_work_size = nvidia_local_work_size;
    else local_work_size = NULL;

    pthread_mutex_lock(&biglock);
    memcpy(&self,arg,sizeof(int));
    session_restore_lflag = session_restore_flag;
    kernel = clCreateKernel(program[self], "sl3", &err );
    if (err != CL_SUCCESS)
    {
	elog("clCreateKernel error (%d)\n",err);
	return hash_err;
    }
    queue = clCreateCommandQueue(context[self], device[self/ocl_threads], 0, &err );

    if (err != CL_SUCCESS)
    {
        elog("clCreateCommandQueue error (%d)\n",err);
	return hash_err;
    }
    buffer = clCreateBuffer(context[self], CL_MEM_WRITE_ONLY, 1000000*20*ocl_vector, NULL, &err );
    if (err != CL_SUCCESS)
    {
	elog("clCreateBuffer error (%d)\n",err);
	return hash_err;
    }
    found_buf = clCreateBuffer(context[self], CL_MEM_WRITE_ONLY, 4, NULL, &err );
    if (err != CL_SUCCESS)
    {
	elog("clCreateBuffer error (%d)\n",err);
	return hash_err;
    }
    if (err != CL_SUCCESS)
    {
	elog("clCreateBuffer error (%d)\n",err);
	return hash_err;
    }
    table = malloc(sizeof(uint)*10*10*10*10*10*10);

    for (a1=0;a1<10;a1++)
    for (a2=0;a2<10;a2++)
    for (a3=0;a3<10;a3++)
    for (a4=0;a4<10;a4++)
    for (a5=0;a5<10;a5++)
    for (a6=0;a6<10;a6++)
    {
	table[a1*100000+a2*10000+a3*1000+a4*100+a5*10+a6] = ((((a1)<<4)|(a2))<<16) | ((((a3)<<4)|(a4))<<8) | (((a5)<<4|(a6)));
    }

    salt.x=salt.y=salt.z=salt.w=0;
    salt.x=(hash_list->salt[0]&255)|((hash_list->salt[1]&255) << 8)|((hash_list->salt[2]&255)<<16)|((hash_list->salt[3]&255)<<24);
    salt.y=(hash_list->salt[4]&255)|((hash_list->salt[5]&255) << 8)|((hash_list->salt[6]&255)<<16);
    REV(salt.x);
    REV(salt.y);

    // SHA1 reversal
    unsigned int E;
    mylist = hash_list;
    memcpy(hex1,mylist->hash,4);
    memcpy(&singlehash.x, hex1, 4);
    REV(singlehash.x);
    singlehash.x-=H0;
    memcpy(hex1,mylist->hash+4,4);
    memcpy(&singlehash.y, hex1, 4);
    REV(singlehash.y);
    singlehash.y-=H1;
    memcpy(hex1,mylist->hash+8,4);
    memcpy(&singlehash.z, hex1, 4);
    REV(singlehash.z);
    singlehash.z-=H2;
    memcpy(hex1,mylist->hash+12,4);
    memcpy(&singlehash.w, hex1, 4);
    REV(singlehash.w);
    singlehash.w-=H3;
    memcpy(&E,mylist->hash+16,4);
    REV(E);
    E-=H4;
    singlehash.y = ROTR(E,30);
    clSetKernelArg(kernel, 8, sizeof(cl_uint4), (void*) &singlehash);

    table_buf = clCreateBuffer(context[self], CL_MEM_READ_ONLY|CL_MEM_COPY_HOST_PTR, sizeof(uint)*10*10*10*10*10*10, table, &err );

    cbase.s0 = sl3bruteforce_charset[0];
    cbase.s1 = sl3bruteforce_charset[1];
    cbase.s2 = sl3bruteforce_charset[2];
    cbase.s3 = sl3bruteforce_charset[3];
    cbase.s4 = sl3bruteforce_charset[4];
    cbase.s5 = sl3bruteforce_charset[5];
    cbase.s6 = sl3bruteforce_charset[6];
    cbase.s7 = sl3bruteforce_charset[7];


    err=clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
    if (err != CL_SUCCESS)
    {
	elog("clSetKernelArg error (%d)\n",err);
	return hash_err;
    }

    err=clSetKernelArg(kernel, 0, sizeof(cl_mem), (void*) &buffer);
    if (err != CL_SUCCESS)
    {
	elog("clSetKernelArg error (%d)\n",err);
	return hash_err;
    }

    err=clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);
    if (err != CL_SUCCESS)
    {
	elog("clSetKernelArg error (%d)\n",err);
	return hash_err;
    }

    err=clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);
    if (err != CL_SUCCESS)
    {
	elog("clSetKernelArg error (%d)\n",err);
	return hash_err;
    }

    err=clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
    if (err != CL_SUCCESS)
    {
	elog("clSetKernelArg error (%d)\n",err);
	return hash_err;
    }

    bzero(found_ind,sizeof(uint)*10*10*10*10*10*10);
    found_ind_buf = clCreateBuffer(context[self], CL_MEM_WRITE_ONLY, 10*10*10*10*10*10*sizeof(cl_uint), NULL, &err );
    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    if (err != CL_SUCCESS)
    {
	elog("clCreateBuffer error (%d)\n",err);
	return hash_err;
    }
    clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, 10*10*10*10*10*10*sizeof(cl_uint), found_ind, 0, NULL, NULL);

    clEnqueueWriteBuffer(queue, found_buf, CL_FALSE, 0, 4, &found, 0, NULL, NULL);
    err=clSetKernelArg(kernel, 4, sizeof(cl_mem), (void*) &found_ind_buf);
    if (err != CL_SUCCESS)
    {
	elog("clSetKernelArg error (%d)\n",err);
	return hash_err;
    }

    err=clSetKernelArg(kernel, 5, sizeof(cl_uint4), (void*) &salt);

    if (err != CL_SUCCESS)
    {
	elog("clSetKernelArg error (%d)\n",err);
	return hash_err;
    }
    err=clSetKernelArg(kernel, 6, sizeof(cl_mem), (void*) &found_buf);
    if (err != CL_SUCCESS)
    {
	elog("clSetKernelArg error (%d)\n",err);
	return hash_err;
    }

    err=clSetKernelArg(kernel, 7, sizeof(cl_mem), (void*) &table_buf);
    if (err != CL_SUCCESS)
    {
	elog("clSetKernelArg error (%d)\n",err);
	return hash_err;
    }

    global_work_size = 10*10*10*10*10*10;

    while ((global_work_size %64)!=0) global_work_size++;
    if (err != CL_SUCCESS)
    {
	elog("clEnqueueBuffer error (%d)\n",err);
	return hash_err;
    }

    image.x=image.y=image.z=image.w=0;

    pthread_mutex_unlock(&biglock);

    // Bruteforce, len=15(12+3)
    csize=9;
    clSetKernelArg(kernel, 2, sizeof(uint), (void*) &csize);
    if (session_restore_lflag==0) ocl_bruteforce_init_restore(self, 8);

    for (a1=thread_restore[self].cpos[0];a1<10;a1++)
    for (a2=thread_restore[self].cpos[1];a2<10;a2++)
    for (a3=thread_restore[self].cpos[2];a3<10;a3++)
    for (a4=thread_restore[self].cpos[3];a4<10;a4++)
    for (a5=thread_restore[self].cpos[4];a5<10;a5++)
    for (a6=thread_restore[self].cpos[5];a6<10;a6++)
    for (a7=thread_restore[self].cpos[6];a7<10;a7++)
    for (a8=thread_restore[self].cpos[7];a8<10;a8++)
    for (a9=thread_restore[self].cpos[8];a9<10;a9+=(ocl_vector*ocl_threads*devicesnum))
    {
	cbase.s0 = a9;
	cbase.s1 = a9+1;
	cbase.s2 = a9+2;
	cbase.s3 = a9+3;

	clSetKernelArg(kernel, 3, sizeof(cl_uint8), (void*) &cbase);
	
	thread_save[self].cpos[0]=a1;
	thread_save[self].cpos[1]=a2;
	thread_save[self].cpos[2]=a3;
	thread_save[self].cpos[3]=a4;
	thread_save[self].cpos[4]=a5;
	thread_save[self].cpos[5]=a6;
	thread_save[self].cpos[6]=a7;
	thread_save[self].cpos[7]=a8;
	thread_save[self].cpos[8]=a9;
	thread_save[self].cstr[0]=a1;
	thread_save[self].cstr[1]=a2;
	thread_save[self].cstr[2]=a3;
	thread_save[self].cstr[3]=a4;
	thread_save[self].cstr[4]=a5;
	thread_save[self].cstr[5]=a6;
	thread_save[self].cstr[6]=a7;
	thread_save[self].cstr[7]=a8;
	thread_save[self].cstr[8]=a9;
	thread_save[self].cstr[9]=0;
	thread_save[self].cstart=8;
	
	
	if (attack_over!=0) return hash_ok;
	image.x=(a1)|(a2<<8)|(a3<<16)|(a4<<24);
	REV(image.x);
	image.y=(a5)|(a6<<8)|(a7<<16)|(a8<<24);
	REV(image.y);
	image.z=0;image.w=0;
	clSetKernelArg(kernel, 1, sizeof(cl_uint4), (void*) &image);

	if (found==1)
	{
    	    found = 0;
    	    clEnqueueWriteBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
    	}
        err=clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global_work_size, local_work_size, 0, NULL, NULL);

	invocations++;
	if (err != CL_SUCCESS)
	{
	    elog("clEnqueueNDRangeKernel error (%d)\n",err);
	    return hash_err;
	}
	//clEnqueueReadBuffer(queue, found_buf, CL_TRUE, 0, 4, &found, 0, NULL, NULL);
	int *fnd;
	fnd = clEnqueueMapBuffer(queue, found_buf, CL_TRUE,CL_MAP_READ, 0, 4, 0, 0, NULL, &err);
	found = *fnd;

        if (session_restore_lflag==1) ocl_bruteforce_init_restore(self, 8);
	session_restore_lflag = 0;
	if (found==1) 
	{
	    err = clEnqueueReadBuffer(queue, found_ind_buf, CL_TRUE, 0, 10*10*10*10*10*10*sizeof(uint), found_ind, 0, NULL, NULL);
	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    e=-1;
	    for (a=0;a<10;a++)
	    for (b=0;b<10;b++)
	    for (c=0;c<10;c++)
	    for (d=0;d<10;d++)
	    for (e=0;e<10;e++)
	    for (f=0;f<10;f++)
	    if (found_ind[a*100000+b*10000+c*1000+d*100+e*10+f]==1)
	    {
    	    h=(a*100000+b*10000+c*1000+d*100+e*10+f)*ocl_vector;
    	    if (err!=CL_SUCCESS) elog("clEnqueueReadBuffer error (%d)\n",err);
	    for (g=0;g<ocl_vector;g++)
	    {
    		h=(a*100000+b*10000+c*1000+d*100+e*10+f)*ocl_vector+g;
    		err = clEnqueueReadBuffer(queue, buffer, CL_TRUE, h*20, 20*ocl_vector, ptr+h*20, 0, NULL, NULL);
    		if ( (hash_index[ptr[h*20]&255][ptr[h*20+1]&255].count<200) && (hash_index[ptr[h*20]&255][ptr[h*20+1]&255].count>0))
    		{
    		    int i;
    		    for (i=0;i<hash_index[ptr[h*20]&255][ptr[h*20+1]&255].count;i++)
    		    if (memcmp(hash_index[ptr[h*20]&255][ptr[h*20+1]&255].nodes[i]->hash, (char *)ptr+(h)*20, hash_ret_len) == 0)
    		    {
                        mylist = hash_index[ptr[h*20]&255][ptr[h*20+1]&255].nodes[i];
                        int flag = 0;
			sprintf(plain,"(C+RND):%01x%01x%01x%01x%01x%01x%01x%01x%01x%01x%01x|%01x%01x%01x%01x",a5,a6,a7,a8,a9+g,a,b,c,d,e,f,a1,a2,a3,a4);
                        pthread_mutex_lock(&crackedmutex);
                        addlist = cracked_list;
                        while (addlist)
                        {
                            if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0))
                                    flag = 1;
                            addlist = addlist->next;
                        }
                        pthread_mutex_unlock(&crackedmutex);
                        if (flag == 0)
                        {
                            add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain);
                            attack_over=2;
                            return hash_ok;
                        }
    		    }
    		}
	    }
    	    }
    	clEnqueueUnmapMemObject(queue,found_buf,(void *)&found,0,NULL,NULL);
    	bzero(found_ind,10*10*10*10*10*10*sizeof(cl_uint));
    	clEnqueueWriteBuffer(queue, found_ind_buf, CL_FALSE, 0, 10*10*10*10*10*10*sizeof(cl_uint), found_ind, 0, NULL, NULL);
	}
    }
    return hash_ok;
}









void ocl_bruteforce_init_restore(int self, int len)
{
    int b;

    for (b=0;b<16;b++) thread_restore[self].cpos[b] = 0;
    thread_restore[self].cpos[len] = self*ocl_vector;
    bzero(thread_restore[self].cstr,16);
    thread_restore[self].cstart = 0;
}



hash_stat ocl_bruteforce_sl3(void)
{
    int a,b;
    int err;
    int worker_thread_keys[32];
    FILE *kfile;
    char *source_str;
    size_t source_size;
    char source_path[255];

    /* we are not restoring a session? */
    if (session_restore_flag==0)
    {
	for (a=0;a<64;a++) 
	{
	    ocl_bruteforce_init_restore(a, 0);
	}
    }

    bruteforce_start=bruteforce_end=15;
    sl3bruteforce_charset[0]=0;
    memcpy(sl3bruteforce_charset+1,"\x01\x02\x03\x04\x05\x06\x07\x08\x09",9);

    b=10;
    attack_overall_count = (uint64_t)(1000000000/ocl_vector);

    /* setup initial OpenCL vars */
    err = clGetPlatformIDs(1, &platform, NULL);
    if (err != CL_SUCCESS)
    {
	elog("clGetPlatformIDs error (%d)\n",err);
	return hash_err;
    }

    err = clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, 16, device, &devicesnum);
    if (err != CL_SUCCESS)
    {
        elog("clGetDeviceIDs error (%d)\n",err);
        return hash_err;
    }


    for (a=0;a<devicesnum*ocl_threads;a++)
    {
	context[a] = clCreateContext(NULL, 1, &device[a/ocl_threads], NULL, NULL, &err);
	if (err != CL_SUCCESS)
	{
	    elog("clCreateContext error (%d)\n",err);
	    return hash_err;
	}
    }
    
    if (ocl_dev_nvidia == 1)
    {
        sprintf(source_path,"%s%s",KERNEL_PATH,"nvidia_sl3.cl");
        hlog("Compiling OpenCL kernel source (%s)","nvidia_sl3.cl");
    }
    else
    {
        sprintf(source_path,"%s%s",KERNEL_PATH,"amd_sl3.cl");
        hlog("Compiling OpenCL kernel source (%s)","amd_sl3.cl");
    }
    kfile = fopen(source_path, "r");
    if (!kfile) 
    {
        elog("Cannot load kernel: %s\n",source_path);
        return hash_err;
    }
    source_str = (char*)malloc(MAX_SOURCE_SIZE);
    bzero(source_str,MAX_SOURCE_SIZE-1);
    source_size = fread(source_str, 1, MAX_SOURCE_SIZE, kfile);
    fclose(kfile);
    
    for (a=0;a<devicesnum*ocl_threads;a++)
    {
	program[a] = clCreateProgramWithSource(context[a], 1, (const char **)&source_str,  (const size_t *)&source_size, &err );
	if (err != CL_SUCCESS)
	{
    	    elog("clCreateProgramWithSource error (%d)\n",err);
    	    return hash_err;
	}

	err = clBuildProgram(program[a], 1, &device[a/ocl_threads], "-cl-strict-aliasing", NULL, NULL );

	if (err != CL_SUCCESS)
	{
    	    elog("clBuildProgram error (%d)\n",err);
    	    return hash_err;
	}
        if ((ocl_have_old_ati==0)&&(ocl_dev_nvidia==0))
        {
            size_t nDevices=1;
            size_t * binary_sizes;
            char ** binaries;
            int i;

            if (hash_list->next) err = clBuildProgram(program[a], 1, &device[a/ocl_threads], NULL, NULL, NULL );
            else err = clBuildProgram(program[a], 1, &device[a/ocl_threads], "-DSINGLE_MODE ", NULL, NULL );

            if (err != CL_SUCCESS)
            {
                elog("clBuildProgram error (%d)\n",err);
                return hash_err;
            }

            err = clGetProgramInfo( program[a], CL_PROGRAM_NUM_DEVICES, sizeof(nDevices), &nDevices, NULL );
            binary_sizes = (size_t *)malloc( sizeof(size_t)*nDevices );
            err = clGetProgramInfo( program[a], CL_PROGRAM_BINARY_SIZES, sizeof(size_t)*nDevices, binary_sizes, NULL );

            binaries = (char **)malloc( sizeof(char *)*nDevices );
            for( i = 0; i < nDevices; i++ )
            {
                if( binary_sizes[i] != 0 )
                {
                    binaries[i] = (char *)malloc( sizeof(char)*binary_sizes[i]+100 );
                }
                else
                {
                    binaries[i] = NULL;
                }
            }
            
            clGetProgramInfo( program[a], CL_PROGRAM_BINARIES, sizeof(char *)*nDevices, binaries, NULL );

            if (a==0) 
            {
                printf("\n");
                hlog("Binary size: %d\n",binary_sizes[0]);
                hlog("Doing BFI_INT magic...%s\n","");
            }
            bfi_int_magic(binaries[0],binary_sizes[0]);
            program[a] = clCreateProgramWithBinary(context[a], 1, &device[a/ocl_threads], &binary_sizes[0], (const unsigned char **)&binaries[0], NULL, &err );
            if (err!=CL_SUCCESS) elog("Cannot compile binary!\n%s","");
            err = clBuildProgram(program[a], 1, &device[a/ocl_threads], NULL, NULL, NULL );
            if (err!=CL_SUCCESS) elog("Cannot build binary!\n%s","");
        }
        else if (ocl_have_sm21==1)
        {
            if (hash_list->next) err = clBuildProgram(program[a], 1, &device[a/ocl_threads], "-DSM21", NULL, NULL );
            else err = clBuildProgram(program[a], 1, &device[a/ocl_threads], "-DSINGLE_MODE -DSM21", NULL, NULL );
            if (err != CL_SUCCESS)
            {
                elog("clBuildProgram error (%d)\n",err);
                return hash_err;
            }
        }
        else
        {
            if (hash_list->next) err = clBuildProgram(program[a], 1, &device[a/ocl_threads], "-DOLD_ATI", NULL, NULL );
            else err = clBuildProgram(program[a], 1, &device[a/ocl_threads], "-DSINGLE_MODE -DOLD_ATI", NULL, NULL );
            if (err != CL_SUCCESS)
            {
                elog("clBuildProgram error (%d)\n",err);
                return hash_err;
            }
        }
    }


    pthread_mutex_init(&biglock, NULL);

    for (a=0;a<ocl_threads*devicesnum;a++)
    {
        worker_thread_keys[a]=a;
        pthread_create(&crack_threads[a], NULL, ocl_bruteforce_sl3_thread, &worker_thread_keys[a]);
    }    
        
    for (a=0;a<ocl_threads*devicesnum;a++) pthread_join(crack_threads[a], NULL);
    printf("\n");
    hlog("Done!\n%s","");
    attack_over=2;
}

#endif
